#include "adder.h"

#include "collections.h"
#include "matchers.h"
#include "pointers.h"
#include "predicate.h"
#include "sorter.h"
#include "strings.h"

#include <vector>
#include <string>


namespace adder
{

void  Adder::addFoodsNoAutoOrder
      (collections::Collections& collections,
       const std::vector<std::string>& collNames,
       const collections::FoodConstPtrs& newFoods) const
{
  for (StringVec::const_iterator collName =
       collNames.begin(); collName != collNames.end(); ++collName)
  {
    for (collections::FoodConstPtrs::const_iterator newFood =
         newFoods.begin(); newFood != newFoods.end(); ++newFood)
    {
      collections[*collName].push_back(*(*newFood));
    }
  }
}

collections::FoodConstPtrs
Adder::getMatchingFoods(const std::string& date,
                        const std::string& meal,
                        const collections::FoodList& foodList) const
{
  /** Regular expression used to determine whether a food's order
   * trait is already set to a value that can be incremented
   */
  const std::string ORDER_RE = "^[0-9]{4}$";

  pointers::Shared<matchers::Matcher> dateMatcher
      (new matchers::Exact(date, false));
  pointers::Shared<matchers::Matcher> mealMatcher
      (new matchers::Exact(meal, false));
  pointers::Shared<matchers::Matcher> orderMatcher
      (new matchers::RE(ORDER_RE, false));

  predicate::Regexps regexps;
  regexps.date.push_back(dateMatcher);
  regexps.meal.push_back(mealMatcher);
  regexps.order.push_back(orderMatcher);

  predicate::Predicate predicate(regexps);
  collections::FoodConstPtrs result;
  for (collections::FoodList::const_iterator food =
       foodList.begin(); food != foodList.end(); ++food)
  {
    if (predicate(*food))
      result.push_back(&*food);
  }
  return result;
}

void Adder::sortFoods(collections::FoodConstPtrs& foods) const
{
  sorter::Key key(sorter::T_ORDER, sorter::D_ASCENDING);
  std::vector<sorter::Key> keys;
  keys.push_back(key);
  sorter::Sorter sorter(keys, sorter::CompareTraits());
  std::sort(foods.begin(), foods.end(), sorter);
}

std::string
Adder::getNewOrder(const collections::FoodConstPtrs& matchingFoods) const
{
  std::string result;

  std::string oldMaxOrder = 
      matchingFoods.back()->getOrder();
  oldMaxOrder = strings::ltrim(oldMaxOrder, "0");
  if (! oldMaxOrder.empty())
      oldMaxOrder.erase(oldMaxOrder.size() - 1, 1);
  int oldMaxOrderInt = 0;
  if (! oldMaxOrder.empty())
      strings::deStringify(oldMaxOrder, oldMaxOrderInt); 

  int newOrderInt = oldMaxOrderInt + 1;
  strings::stringify(newOrderInt, result);
  result.append("0");
  unsigned int padOnLeft;
  if (result.size() > 3)
      padOnLeft = 0;
  else
      padOnLeft = 4 - result.size();
  result.insert(result.begin(), padOnLeft, '0');

  return result;
}

void 
Adder::addFoodsAutoOrder(collections::FoodList& foodList,
                         const collections::FoodConstPtrs& newFoods) const
{
  for (collections::FoodConstPtrs::const_iterator newFood =
          newFoods.begin(); newFood != newFoods.end(); ++newFood)
  {
    collections::FoodConstPtrs matchingFoods =
        getMatchingFoods((*newFood)->getDate(), (*newFood)->getMeal(),
                         foodList);

    std::string newOrder = "0010";
    if (! matchingFoods.empty())
    {
      sortFoods(matchingFoods);
      newOrder = getNewOrder(matchingFoods);
    }

    food::Food incremented(**newFood);
    incremented.setOrder(newOrder);
    foodList.push_back(incremented);
  }
}

Adder::Adder(const StringVec& addTo,
             bool autoOrder)
    : addTo_(addTo), autoOrder_(autoOrder) {}

void 
Adder::process(collections::Collections& collections,
               collections::Unsaved& unsaved,
               const collections::FoodConstPtrs& newFoods) const
{
  if (newFoods.empty())
    return;
  if (autoOrder_)
    for (StringVec::const_iterator collName =
            addTo_.begin(); collName != addTo_.end(); ++collName)
    {
      addFoodsAutoOrder(collections[*collName], newFoods);
    }
  else
    addFoodsNoAutoOrder(collections, addTo_, newFoods);

  for (std::vector<std::string>::const_iterator collName =
          addTo_.begin(); collName != addTo_.end(); ++collName)
  {
      unsaved[*collName] = true;
  }
}


} // namespace adder
