#ifndef ADDER_H
#define ADDER_H

#include "collections.h"
#include "food.h"

#include <map>
#include <string>
#include <vector>

/** Contains Adder, a class to add foods to a collection.
 */
namespace adder {

typedef std::vector<std::string> StringVec;
typedef std::vector<food::Food> FoodVec;
typedef std::map<std::string, FoodVec > FoodVecMap;

/** Class to add foods to a collection.
 */
class Adder
{
public:
    /** ctor.
     *
     * \param addTo Vector of strings where each string is the name
     * of a collection to which foods will be added.
     *
     * \param autoOrder Whether to automatically increment the order
     * trait when adding foods.
     */
    Adder(const StringVec& addTo, bool autoOrder);

    /** Add new foods to a database.
     *
     * \param collections Collections to which to add the
     * new foods.
     *
     * \param unsaved Map indicating which collections in database
     * have not been saved yet; process will update this
     * appropriately.
     *
     * \param newFoods Foods to add.
     */
    void process(collections::Collections& collections,
                 collections::Unsaved& unsaved,
                 const collections::FoodConstPtrs& newFoods) const;

private:
  /** Which collections to which to add new foods.
   */
  StringVec addTo_;

  /** Whether to automatically increment the order trait.
   */
  bool autoOrder_;

  /** Add foods to a database without incrementing the Order trait.
   *
   * \param database Database to which to add the new foods.
   *
   * \param addTo Collections to which to add the new foods. If these
   * collections are not already in database, they will be created.
   *
   * \param newFoods Vector of pointers to the foods to be added.
   */
  void addFoodsNoAutoOrder
       (collections::Collections& collections,
        const StringVec& collNames,
        const collections::FoodConstPtrs& newFoods) const;

  /** Get foods that match given date and meal traits,
   * and that already have an appropriately set Order trait.
   * After sorting these foods in ascending order, they may be used to
   * find the next Order trait.
   *
   * \param date Date value to use in search
   * \param meal Meal value to use in search
   * \param collection Collection to search for matches
   *
   * \return Vector of foods whose date and meal traits match those of
   * the paramters, and whose Order trait could have been
   * automatically set
   */
  collections::FoodConstPtrs
  getMatchingFoods(const std::string& date,
                   const std::string& meal,
                   const collections::FoodList& foodList) const;

  /** Sort foods in ascending order by their Order trait.
   *
   * \param foods Sort these food pointers, in place.
   */
  void sortFoods(collections::FoodConstPtrs& foods) const;

  /** Obtain a new Order trait value to use for autoOrder.
   *
   * \param matchingFoods Vector of foods which has been sorted using
   * sortFoods.
   *
   * \return value for new Order trait.
   */
  std::string getNewOrder
              (const collections::FoodConstPtrs& matchingFoods) const;

  /** Add a food to a collection, automatically incrementing the Order
   * trait of new foods.
   *
   * \param collection Collection to which to add foods.
   *
   * \param newFoods vector of pointers to foods to add to the
   * collection.
   */
  void addFoodsAutoOrder(collections::FoodList& collection,
                         const collections::FoodConstPtrs& newFoods) const;

};

} // namespace adder

#endif
