#include "changer.h"

#include "collections.h"
#include "db.h"
#include "food.h"
#include "matchers.h"
#include "numbers.h"
#include "stl.h"
#include "strings.h"

#include <algorithm>
#include <list>
#include <map>
#include <string>
#include <utility>
#include <vector>

namespace changer
{

Changer::Renamer::Renamer(const std::string& argToRename,
                          const std::string& argNewName,
                          bool ignoreCase, bool exactMatch)
  : toRename(matchers::getMatcher(argToRename, ignoreCase, exactMatch)),
    pattern(argToRename),
    newName(argNewName) {}

Changer::ChangeTag::ChangeTag()
  : doChange(false) {}

Changer::ChangeNumber::ChangeNumber(const std::string& argName,
                                    const std::string& argNumber)
  : name(argName), number(numbers::Number(argNumber)) {}

Changer::TraitIfSet::TraitIfSet
    (const std::vector<std::pair<std::string, 
                                 std::string> >& traitStrings)
  : name(std::make_pair(false, "")),
    group(std::make_pair(false, "")),
    date(std::make_pair(false, "")),
    meal(std::make_pair(false, "")),
    comment(std::make_pair(false, "")),
    order(std::make_pair(false, "")),
    refDesc(std::make_pair(false, "")),
    instructions(std::make_pair(false, "")),
    qty(std::make_pair(false, "")),
    pctRefuse(std::make_pair(false, "")),
    yield(std::make_pair(false, "")),
    unit(std::make_pair(false, ""))
{
  std::map<std::string, std::pair<bool, std::string>* > traitMap
      = getTraitMap();

  std::vector<std::string> traitNames = stl::getKeys(traitMap);
  for (std::vector<std::pair<std::string, 
                             std::string> >::const_iterator traitString
          = traitStrings.begin(); traitString != traitStrings.end();
          ++traitString)
  {
    std::vector<std::string> matches =
        strings::matchAbbrev(traitNames, traitString->first);
    if (matches.size() != 1)
        throw BadTraitAbbreviation(matches.size(),
                                   traitString->first);
    const std::string& traitName = matches[0];
    std::pair<bool, std::string>* resultPtr =
        traitMap.find(traitName)->second;
    resultPtr->first = true;
    resultPtr->second = traitString->second;
  }
}

std::map<std::string, std::pair<bool, std::string>* >
Changer::TraitIfSet::getTraitMap()
{
  using std::make_pair;
  using namespace food;
  std::map<std::string, std::pair<bool, std::string>* > result;

  result.insert(make_pair(NAME, &this->name));
  result.insert(make_pair(GROUP, &this->group));
  result.insert(make_pair(DATE, &this->date));
  result.insert(make_pair(MEAL, &this->meal));
  result.insert(make_pair(COMMENT, &this->comment));
  result.insert(make_pair(ORDER, &this->order));
  result.insert(make_pair(REFDESC, &this->refDesc));
  result.insert(make_pair(INSTRUCTIONS, &this->instructions));
  result.insert(make_pair(QUANTITY, &this->qty));
  result.insert(make_pair(PCTREFUSE, &this->pctRefuse));
  result.insert(make_pair(YIELD, &this->yield));
  result.insert(make_pair(UNIT, &this->unit));
  return result;
}

void Changer::initTraits(const StringPairVec& newTraits,
                         bool ignoreCase, bool exactMatch)
{
  TraitIfSet traitIfSet(newTraits);
  if (traitIfSet.name.first)
  {
    name_.doChange = true;
    name_.newValue = traitIfSet.name.second;
  }

  if (traitIfSet.group.first)
  {
    group_.doChange = true;
    group_.newValue = traitIfSet.group.second;
  }

  if (traitIfSet.date.first)
  {
    date_.doChange = true;
    date_.newValue = traitIfSet.date.second;
  }

  if (traitIfSet.meal.first)
  {
    meal_.doChange = true;
    meal_.newValue = traitIfSet.meal.second;
  }

  if (traitIfSet.order.first)
  {
    order_.doChange = true;
    order_.newValue = traitIfSet.order.second;
  }

  if (traitIfSet.comment.first)
  {
    comment_.doChange = true;
    comment_.newValue = traitIfSet.comment.second;
  }


  if (traitIfSet.instructions.first)
  {
    instructions_.doChange = true;
    instructions_.newValue = traitIfSet.instructions.second;
  }

  if (traitIfSet.refDesc.first)
  {
    refDesc_.doChange = true;
    refDesc_.newValue = traitIfSet.refDesc.second;
  }

  if (traitIfSet.unit.first)
  {
    unit_ = matchers::getMatcher(traitIfSet.unit.second,
                                 ignoreCase, exactMatch);
    unitPattern_ = traitIfSet.unit.second;
  }

  if (traitIfSet.qty.first)
  {
    changeQty_ = true;
    newQty_ = numbers::StrNumber(traitIfSet.qty.second);
  }

  if (traitIfSet.pctRefuse.first)
  {
    changePctRefuse_ = true;
    newPctRefuse_ = numbers::Percent(traitIfSet.pctRefuse.second);
  }

  if (traitIfSet.yield.first)
  {
    changeYield_ = true;
    newYield_ = numbers::Number(traitIfSet.yield.second);
  }
}

void Changer::initCByNutMember(const StringPairVec& stringPairs,
                               bool ignoreCase, bool exactMatch)
{
  if (stringPairs.empty())
    return;
  const std::string& pattern      = stringPairs.back().first;
  const std::string& numberString = stringPairs.back().second;

  byNutMatcher_ = matchers::getMatcher(pattern, ignoreCase, exactMatch);
  byNutAmt_ = numbers::Number(numberString);
  byNutPattern_ = pattern;
}

void Changer::initChangeNumberVecMember(ChangeNumberVec& member,
                                        const StringPairVec& newNuts)
{
  for (StringPairVec::const_iterator nameAndAmt = newNuts.begin();
       nameAndAmt != newNuts.end(); ++nameAndAmt)
  {
    member.push_back(ChangeNumber(nameAndAmt->first,
                                  nameAndAmt->second));
  }
}

void Changer::initRenamerVecMember(RenamerVec& member,
                                   const StringPairVec& strings,
                                   bool ignoreCase, bool exactMatch)
{
  for (StringPairVec::const_iterator patAndName = strings.begin();
       patAndName != strings.end(); ++patAndName)
  {
    member.push_back(Renamer(patAndName->first, patAndName->second,
                             ignoreCase, exactMatch));
  }
}

void Changer::initMatcherVecMember(MatcherVec& member,
                                   const std::vector<std::string>& patterns,
                                   bool ignoreCase, bool exactMatch)
{
  for (std::vector<std::string>::const_iterator pattern =
       patterns.begin(); pattern != patterns.end(); ++pattern)
  {
    member.push_back(matchers::getMatcher(*pattern,
                                          ignoreCase, exactMatch));
  }
}

void Changer::initIngredientsMember
      (const db::Db& db,
       const std::vector<std::string>& collNames)
{
  std::vector<std::string> dbCollNames = db.getCollectionNames();
  for (std::vector<std::string>::const_iterator abbrev =
       collNames.begin(); abbrev != collNames.end(); ++abbrev)
  {
    std::vector<std::string> matchingCollNames =
      strings::matchAbbrev(dbCollNames, *abbrev);
    if (matchingCollNames.size() != 1)
      throw IngredientCollectionNotFoundError(matchingCollNames, *abbrev);
    std::list<food::Food> ingredients =
      db.getFoodList(matchingCollNames[0]);
    
    for (std::list<food::Food>::const_iterator ingredient =
         ingredients.begin();
         ingredient != ingredients.end(); ++ingredient)
    {
      ingredients_.push_back(*ingredient);
    }
  }
}

Changer::Changer(const StringPairVec& newTraits,
                 const StringPairVec& cByNut,
                 const bool& useRefuse,
                 const StringPairVec& newNuts,
                 const StringPairVec& argRenameNuts,
                 const std::vector<std::string>& argDeleteNuts,
                 const StringPairVec& newArbUnits,
                 const StringPairVec& argRenameArbUnits,
                 const std::vector<std::string>& argDeleteArbUnits,
                 const db::Db& db,
                 const std::vector<std::string>& ingrCollNames,
                 const std::vector<std::string>& argDeleteIngredients,
                 const bool& ignoreCase,
                 const bool& exactMatch)
  : changeQty_(false), changePctRefuse_(false),
    changeYield_(false), useRefuse_(false)
{
  initTraits(newTraits, ignoreCase, exactMatch);
  initCByNutMember(cByNut, ignoreCase, exactMatch);
  useRefuse_ = useRefuse;
  initChangeNumberVecMember(newNuts_, newNuts);
  initRenamerVecMember(renameNuts_, argRenameNuts, ignoreCase, exactMatch);
  initMatcherVecMember(deleteNuts_, argDeleteNuts, ignoreCase, exactMatch);
  initChangeNumberVecMember(arbUnits_, newArbUnits);
  initRenamerVecMember(renameArbUnits_, argRenameArbUnits, 
                       ignoreCase, exactMatch);
  initMatcherVecMember(deleteArbUnits_, argDeleteArbUnits,
                       ignoreCase, exactMatch);
  initIngredientsMember(db, ingrCollNames);
  initMatcherVecMember(deleteIngredients_, argDeleteIngredients,
                       ignoreCase, exactMatch);
} 

food::Food Changer::operator() (food::Food food) const
{
  addArbUnits(food);
  renameArbUnits(food);
  changeTraits(food);
  deleteArbUnits(food);
  deleteIngredients(food);
  addIngredients(food);
  doRefuse(food);
  addNuts(food);
  renameNuts(food);
  deleteNuts(food);
  doCByNut(food);
  return food;
}

void Changer::addArbUnits(food::Food& food) const
{
  for (ChangeNumberVec::const_iterator changeNumber = arbUnits_.begin();
       changeNumber != arbUnits_.end(); ++changeNumber)
  {
    food.addArbUnit(changeNumber->name, changeNumber->number);
  }
}

void Changer::renameArbUnits(food::Food& food) const
{
  std::string currUnitName = food.getCurrentUnitName();
  std::vector<std::string> absUnitNames = food.getAbsUnitNames();
  std::sort(absUnitNames.begin(), absUnitNames.end());
  bool currUnitIsAbs =
    std::binary_search(absUnitNames.begin(), absUnitNames.end(),
                       currUnitName);
  for (RenamerVec::const_iterator renamer = renameArbUnits_.begin();
       renamer != renameArbUnits_.end(); ++renamer)
  {
    std::vector<std::string> arbUnitNames = food.getArbUnitNames();
    std::vector<std::string> toRename =
      matchers::selectStrings(*renamer->toRename, arbUnitNames);
    if (toRename.size() > 1)
      throw RenameUnitError(food, renamer->pattern, toRename);
    if (toRename.empty())
      continue;
    if ((! currUnitIsAbs) && (currUnitName == toRename[0]))
      throw RenameArbUnitError(food, toRename[0]);
    food.renameArbUnit(toRename[0], renamer->newName);
  }
}

void Changer::changeTraits(food::Food& food) const
{
  if (unit_.get() != 0)
    changeUnit(food);
  if (name_.doChange)
    food.setName(name_.newValue);
  if (group_.doChange)
    food.setGroup(group_.newValue);
  if (date_.doChange)
    food.setDate(date_.newValue);
  if (meal_.doChange)
    food.setMeal(meal_.newValue);
  if (order_.doChange)
    food.setOrder(order_.newValue);
  if (comment_.doChange)
    food.setComment(comment_.newValue);
  if (instructions_.doChange)
    food.setInstructions(instructions_.newValue);
  if (refDesc_.doChange)
    food.setRefDesc(refDesc_.newValue);

  if (changeQty_)
    food.setQty(newQty_);
  if (changePctRefuse_)
    food.setPctRefuse(newPctRefuse_);
  if (changeYield_)
    food.setYield(newYield_);
}

void Changer::changeUnit(food::Food& food) const
{
    std::vector<std::string> arbUnitNames = food.getArbUnitNames();
    std::vector<std::string> absUnitNames = food.getAbsUnitNames();
    std::vector<std::string> allUnitNames = arbUnitNames;
    allUnitNames.insert(allUnitNames.end(), absUnitNames.begin(),
                        absUnitNames.end());
    std::sort(allUnitNames.begin(), allUnitNames.end());
    allUnitNames.erase(std::unique(allUnitNames.begin(), allUnitNames.end()),
                       allUnitNames.end());
    std::vector<std::string> matchingUnits =
      matchers::selectStrings(*unit_, allUnitNames);
    if (matchingUnits.size() != 1)
      throw ChangeUnitError(food, unitPattern_, matchingUnits);
    const std::string& unitName = matchingUnits[0];
    if (std::find(absUnitNames.begin(), absUnitNames.end(), unitName) !=
        absUnitNames.end())
    {
      food.setCurrentUnitAbsolute(unitName);
    }
    else
    {
      food.setCurrentUnitArbitrary(unitName);
    }
}

void Changer::deleteArbUnits(food::Food& food) const
{
  for (MatcherVec::const_iterator matcher = deleteArbUnits_.begin();
       matcher != deleteArbUnits_.end(); ++matcher)
  {
    std::vector<std::string> arbUnitNames = food.getArbUnitNames();
    std::vector<std::string> unitsToDelete =
      matchers::selectStrings(**matcher, arbUnitNames);
    std::string currUnit = food.getCurrentUnitName();
    if (std::find(unitsToDelete.begin(), unitsToDelete.end(), currUnit) !=
      unitsToDelete.end())
      throw DeleteArbUnitError(food, currUnit);

    for (std::vector<std::string>::const_iterator unitToDelete =
         unitsToDelete.begin(); unitToDelete != unitsToDelete.end();
         ++unitToDelete)
    {
      food.deleteArbUnit(*unitToDelete);
    }
  }
}

void Changer::deleteIngredients(food::Food& food) const
{
  std::vector<food::Food> ingredientsVec = food.getIngredients();
  std::list<food::Food> ingredientList(ingredientsVec.begin(),
                                       ingredientsVec.end());
  for (MatcherVec::const_iterator matcher = deleteIngredients_.begin();
       matcher != deleteIngredients_.end(); ++matcher)
  {
    std::vector<std::list<food::Food>::iterator> ingredientsToDelete;
    for (std::list<food::Food>::iterator ingredient = ingredientList.begin();
         ingredient != ingredientList.end(); ++ingredient)
    {
      std::string ingredientName = ingredient->getName();
      if (! (**matcher)(ingredientName))
        ingredientsToDelete.push_back(ingredient);
    }
    for (std::vector<std::list<food::Food>::iterator>::const_iterator
         ingredientToDelete = ingredientsToDelete.begin();
         ingredientToDelete != ingredientsToDelete.end();
         ++ingredientToDelete)
    {
      ingredientList.erase(*ingredientToDelete);
    }
  }
  ingredientsVec.clear();
  ingredientsVec.insert(ingredientsVec.end(), ingredientList.begin(),
                        ingredientList.end());
  food.setIngredients(ingredientsVec);
}

void Changer::addIngredients(food::Food& food) const
{
  std::vector<food::Food> ingredients = food.getIngredients();
  ingredients.insert(ingredients.end(), ingredients_.begin(),
                     ingredients_.end());
  food.setIngredients(ingredients);
}

void Changer::doRefuse(food::Food& food) const
{
  if (! useRefuse_)
    return;
  numbers::StrNumber newQty = food::getQtyMinusRefuse(food);
  food.setQty(newQty);
}

void Changer::addNuts(food::Food& food) const
{
  if (newNuts_.empty())
    return;
  numbers::Number grams = food::getGrams(food);
  if (grams.getValue() < 1)
    throw AddNutError(food);
  for (ChangeNumberVec::const_iterator newNut = newNuts_.begin();
       newNut != newNuts_.end(); ++newNut)
  {
    food.addNut(newNut->name, newNut->number);
  }
}

void Changer::renameNuts(food::Food& food) const
{
  for (RenamerVec::const_iterator renamer = renameNuts_.begin();
       renamer != renameNuts_.end(); ++renamer)
  {
    std::vector<std::string> nutNames = food.getNutNames();
    std::vector<std::string> nutsToRename =
      matchers::selectStrings(*renamer->toRename, nutNames);
    if (nutsToRename.size() > 1)
      throw RenameNutError(food, renamer->pattern, nutsToRename);
    if (! nutsToRename.empty())
      food.renameNut(nutsToRename[0], renamer->newName);
  }
}

void Changer::deleteNuts(food::Food& food) const
{
  for (MatcherVec::const_iterator matcher = deleteNuts_.begin();
       matcher != deleteNuts_.end(); ++matcher)
  {
    std::vector<std::string> nutNames = food.getNutNames();
    std::vector<std::string> nutsToDelete =
      matchers::selectStrings(**matcher, nutNames);
    for (std::vector<std::string>::const_iterator nutToDelete =
         nutsToDelete.begin(); nutToDelete != nutsToDelete.end();
         ++nutToDelete)
    {
      food.deleteNut(*nutToDelete);
    }
  }
}

void Changer::doCByNut(food::Food& food) const
{
  if (byNutMatcher_.get() == 0)
    return;
  std::vector<std::string> nutNames = food.getNutNames();
  std::vector<std::string> matchingNuts =
    matchers::selectStrings(*byNutMatcher_, nutNames);
  if (matchingNuts.size() != 1)
    throw ByNutMatchError(food, byNutPattern_, matchingNuts);
  numbers::StrNumber newQty;
  bool byNutReturnValue =
    food.getQtyByNut(matchingNuts[0], byNutAmt_, newQty);
  if (byNutReturnValue == false)
    throw ByNutZeroNutError(food, matchingNuts[0]);
  food.setQty(newQty);
}

ByNutMatchError::ByNutMatchError(const food::Food& food,
                const std::string& pattern,
                const std::vector<std::string>& matchingNames)
  : foodName_(food.getName()),
    pattern_(pattern), matchingNames_(matchingNames) {}

std::string ByNutMatchError::getBriefMessage() const
{
  std::string result = "get quantity by nutrient: ";
  if (matchingNames_.empty())
    result.append("no ");
  else
    result.append("multiple ");
  result.append("matching nutrients found for pattern ");
  result.append(pattern_);
  result.append(" for food named ");
  result.append(foodName_);
  return result;
}

ByNutZeroNutError::ByNutZeroNutError(const food::Food& food,
                                     const std::string& nutName)
  : foodName_(food.getName()), nutName_(nutName) {}

std::string ByNutZeroNutError::getBriefMessage() const
{
  std::string result;
  result.append("get quantity by nutrient failed because nutrient named ");
  result.append(nutName_);
  result.append(" for food named ");
  result.append(foodName_);
  result.append("has a value of zero");
  return result;
};

RenameNutError::RenameNutError(const food::Food& food,
                               const std::string& oldNutPattern,
                               const std::vector<std::string>& matches)
    : foodName_(food.getName()), oldNutPattern_(oldNutPattern),
      matches_(matches) {}

std::string RenameNutError::getBriefMessage() const
{
    std::string result="could not rename nutrient with pattern ";
    result.append(oldNutPattern_);
    result.append(" for food named ");
    result.append(foodName_);
    result.append("; more than one matching nutrient name. Matches:");
    for (std::vector<std::string>::const_iterator match = matches_.begin();
            match != matches_.end(); ++match)
    {
        result.append("\n");
        result.append(*match);
    }
    return result;
}

AddNutError::AddNutError(const food::Food& food)
  : foodName_(food.getName()) {}

std::string AddNutError::getBriefMessage() const
{
  std::string result = "could not add nutrient to food named ";
  result.append(foodName_);
  result.append("; food's weight is less than one gram.");
  return result;
}

DeleteArbUnitError::DeleteArbUnitError(const food::Food& food,
                                       const std::string& unitName)
  : foodName_(food.getName()), unitName_(unitName) {}

std::string DeleteArbUnitError::getBriefMessage() const
{
  std::string result = "could not delete unit named ";
  result.append(unitName_);
  result.append(" from food named ");
  result.append(foodName_);
  result.append("; it is the food's current unit");
  return result;
}

RenameArbUnitError::RenameArbUnitError(const food::Food& food,
                                       const std::string& unitName)
  : foodName_(food.getName()), unitName_(unitName) {}

std::string RenameArbUnitError::getBriefMessage() const
{
  std::string result = "could not rename unit named ";
  result.append(unitName_);
  result.append(" in food named ");
  result.append(foodName_);
  result.append("; it is the food's current unit");
  return result;
}

ChangeUnitError::ChangeUnitError(const food::Food& food,
                                 const std::string& pattern,
                                 const std::vector<std::string> matches)
  : foodName_(food.getName()), pattern_(pattern), matches_(matches) {}

std::string ChangeUnitError::getBriefMessage() const
{
  std::string result = "could not change unit on food named ";
  result.append(foodName_);
  result.append(" because pattern ");
  result.append(pattern_);
  result.append("matches ");
  if (matches_.empty())
    result.append("no ");
  else
    result.append("more than one ");
  result.append("available unit.");
  return result;
}

RenameUnitError::RenameUnitError(const food::Food& food,
                                 const std::string& oldUnitPattern,
                                 const std::vector<std::string>& matches)
  : foodName_(food.getName()), oldUnitPattern_(oldUnitPattern),
    matches_(matches) {}

std::string RenameUnitError::getBriefMessage() const
{
  std::string result = "could not rename unit on food named ";
  result.append(foodName_);
  result.append(" because pattern ");
  result.append(oldUnitPattern_);
  result.append(" matches ");
  if (matches_.empty())
    result.append("no ");
  else
    result.append("more than one ");
  result.append("of the food's nutrients.");
  return result;
}

BadTraitAbbreviation::BadTraitAbbreviation(unsigned int nMatches,
                              const std::string& abbrev)
    : nMatches_(nMatches), abbrev_(abbrev) {}

std::string BadTraitAbbreviation::getBriefMessage() const
{
  std::string result = "invalid trait abbreviation: ";
  result.append(abbrev_);
  if (nMatches_ == 0)
      result.append(": no");
  else
      result.append(": multiple");
  result.append(" matches found");
  return result;
}

IngredientCollectionNotFoundError::IngredientCollectionNotFoundError
  (const std::vector<std::string>& matchingCollNames,
   const std::string& abbreviation)
  : matchingCollNames_(matchingCollNames), abbreviation_(abbreviation) {}

std::string IngredientCollectionNotFoundError::getBriefMessage() const
{
  std::string result = "collection name given for ingredients "
                       "does not match exactly one collection";
  return result;
}



} // namespace changer
