#ifndef CHANGER_H
#define CHANGER_H

#include "food.h"

#include "collections.h"
#include "db_fwd.h"
#include "errors.h"
#include "matchers.h"
#include "pointers.h"

#include <list>
#include <map>
#include <string>
#include <utility>
#include <vector>

namespace changer {

typedef std::vector<std::pair<std::string, std::string> > StringPairVec;

/** Change a food's traits, arbitrary units, nutrients, and
 * ingredients.
 */
class Changer
{
public:
  /** ctor. Most of the parameters are vectors of pairs of
   * strings so that command-line input may be passed
   * directly to the constructor. If the input is invalid,
   * this will throw an appropriate subclass of
   * errors::NonFatal.
   */
  Changer(const StringPairVec& newTraits,
          const StringPairVec& cByNut,
          const bool& useRefuse,
          const StringPairVec& newNuts,
          const StringPairVec& renameNuts,
          const std::vector<std::string>& deleteNuts,
          const StringPairVec& newArbUnits,
          const StringPairVec& renameArbUnits,
          const std::vector<std::string>& deleteArbUnits,
          const db::Db& db,
          const std::vector<std::string>& ingrCollNames,
          const std::vector<std::string>& deleteIngredients,
          const bool& ignoreCase,
          const bool& exactMatch);

  // default copy ctor, assignment operator, dtor OK

  /** Change a food to the new values that were given in the ctor.
   * \param food Food to change.
   *
   * \throw errors::UnitMatch Matcher for unit did not match
   * exactly one of the food's available units.
   *
   * \throw errors::ZeroFoodGrams Cannot add any nuts to this food
   * because the total grams of the food is zero.
   *
   * \throw errors::DeleteUnit Attempted to delete the current
   * unit of the food.
   *
   * \throw errors::NutMatch change quantity by nutrient
   * requested, but the pattern given did not match exactly one
   * nutrient of the food.
   */
  food::Food operator() (food::Food food) const;

private:
  struct Renamer
  {
    Renamer(const std::string& argToRename,
            const std::string& argNewName,
            bool ignoreCase, bool exactMatch);
    pointers::Shared<matchers::Matcher> toRename;
    std::string pattern; 
    std::string newName;
  };

  typedef std::vector<Renamer> RenamerVec;
  typedef std::vector<pointers::Shared<matchers::Matcher> > MatcherVec;

  struct ChangeTag
  {
    ChangeTag();
    bool doChange;
    std::string newValue;
  };

  struct ChangeNumber
  {
    ChangeNumber(const std::string& argName,
                 const std::string& argNumber);
    std::string name;
    numbers::Number number;
  };

  typedef std::vector<ChangeNumber> ChangeNumberVec;

  /** Convert values passed from command line for --change option
   * to a struct.
   *
   * \param traitStrings a vector of pairs, where the first value
   * of each pair is the name of a trait. This can be the full
   * name (such as "date") or an abbreviation (such as "da" or
   * "d".) The second value is the desired new value for the
   * trait.
   *
   * creates TraitIfSet, a struct which contains a pair for each
   * trait. The first value is a bool where True indicates that
   * the user wishes for this trait to be changed; False indicates
   * that the user does not want this trait to be changed. The
   * second value is a string which holds the desired new value
   * for the trait, so long as the first value of the pair is
   * True.
   *
   * \throw BadTraitAbbreviation One of the trait
   * abbreviations was not recognized.
   */
  struct TraitIfSet
  {
    TraitIfSet(const std::vector<std::pair<std::string, 
                                           std::string> >& traitStrings);
    std::pair<bool, std::string>     name;
    std::pair<bool, std::string>     group;
    std::pair<bool, std::string>     date;
    std::pair<bool, std::string>     meal;
    std::pair<bool, std::string>     comment;
    std::pair<bool, std::string>     order;
    std::pair<bool, std::string>     refDesc;
    std::pair<bool, std::string>     instructions;
    std::pair<bool, std::string>     qty;
    std::pair<bool, std::string>     pctRefuse;
    std::pair<bool, std::string>     yield;
    std::pair<bool, std::string>     unit;

    /** Get a map to lookup pairs in a TraitIfSet map.
     *
     * Say for example we have a series of pairs, such as "name
     * apples" and "date Wednesday". In order to get these pairs
     * into a struct, we need a way to take the trait names ("name"
     * and "date") and look up the corresponding members in a
     * TraitIfSet struct (TraitIfSet.name and TraitIfSet.date). Then
     * we need to be able to place the corresponding values
     * ("apples" and "Wednesday") into the struct, and then change
     * the boolean in TraitIfSet to indicate that the user wants the
     * trait to be changed.
     *
     * getTraitMap creates a map where the keys are the trait names
     * (including "name" and "date"). The values are pointers to
     * members of a TraitIfSet instance. Thus, for the "name Apples"
     * pair, we can use the map that this function returns to look
     * up a pointer. The pointer stored at key "name" will point to
     * target.Name. Then it is simple to use that pointer to
     * perform pointer->first = true, and
     * pointer->second = "apples". 
     */
    std::map<std::string, std::pair<bool, std::string>* >
    getTraitMap();

  };

  void initTraits(const StringPairVec& newTraits,
                  bool ignoreCase, bool exactMatch);
  void initCByNutMember(const StringPairVec& stringPairs,
                        bool ignoreCase, bool exactMatch);
  void initChangeNumberVecMember(ChangeNumberVec& member,
                                 const StringPairVec& newNuts);
  void initRenamerVecMember(RenamerVec& member,
                            const StringPairVec& stringPairs,
                            bool ignoreCase, bool exactMatch);
  void initMatcherVecMember(MatcherVec& member,
                            const std::vector<std::string>& patterns,
                            bool ignoreCase, bool exactMatch);
  void initIngredientsMember(const db::Db& db,
                             const std::vector<std::string>& collNames);

  void addArbUnits(food::Food& food) const;
  void renameArbUnits(food::Food& food) const;
  void changeTraits(food::Food& food) const;
 
  // changeTraits calls changeUnit
  void changeUnit(food::Food& food) const;
  void deleteArbUnits(food::Food& food) const;
  void deleteIngredients(food::Food& food) const;
  void addIngredients(food::Food& food) const;
  void doRefuse(food::Food& food) const;
  void addNuts(food::Food& food) const;
  void renameNuts(food::Food& food) const;
  void deleteNuts(food::Food& food) const;
  void doCByNut(food::Food& food) const;

  //// TRAITS
  // Tags (8)
  ChangeTag                             name_;
  ChangeTag                             group_;
  ChangeTag                             date_;
  ChangeTag                             meal_;
  ChangeTag                             order_;
  ChangeTag                             comment_;
  ChangeTag                             instructions_;
  ChangeTag                             refDesc_;

  // Unit (1)
  pointers::Shared<matchers::Matcher>   unit_;
  std::string                           unitPattern_;

  // Quantity (1)
  bool                                  changeQty_;
  numbers::StrNumber                    newQty_;

  // PctRefuse, Yield (2)
  bool                                  changePctRefuse_;
  numbers::Percent                      newPctRefuse_;
  bool                                  changeYield_;
  numbers::Number                       newYield_;
  //// END TRAITS

  pointers::Shared<matchers::Matcher>   byNutMatcher_;
  std::string                           byNutPattern_;
  numbers::Number                       byNutAmt_;

  bool                                  useRefuse_;
  ChangeNumberVec                       newNuts_;
  RenamerVec                            renameNuts_;
  MatcherVec                            deleteNuts_;
  ChangeNumberVec                       arbUnits_;
  RenamerVec                            renameArbUnits_;
  MatcherVec                            deleteArbUnits_;
  std::vector<food::Food>               ingredients_;
  MatcherVec                            deleteIngredients_;
};

/** Could not add nutrient because food's current grams is
 * not at least 1.
 */
class AddNutError : public errors::NonFatal
{
public:
  AddNutError(const food::Food& food);
  std::string getBriefMessage() const;

private:
  std::string foodName_;
};

/** An incorrect trait abbreviation was used when specifying 
 * new traits for a food.
 */
class BadTraitAbbreviation : public errors::NonFatal
{
public:
  BadTraitAbbreviation(unsigned int nMatches,
                       const std::string& abbrev);
  std::string getBriefMessage() const;

private:
  unsigned int nMatches_;
  std::string abbrev_;
};

/** Could not rename nutrient because pattern matched more than one
 * nutrient.
 */
class RenameNutError : public errors::NonFatal
{
public:
  RenameNutError(const food::Food& food,
                 const std::string& oldNutPattern,
                 const std::vector<std::string>& matches);
  std::string getBriefMessage() const;

private:
  std::string foodName_;
  std::string oldNutPattern_;
  std::vector<std::string> matches_;
};

/** Could not rename arbitrary unit because pattern matched more
 * than one nutrient.
 */
class RenameUnitError : public errors::NonFatal
{
public:
  RenameUnitError(const food::Food& food,
                 const std::string& oldUnitPattern,
                 const std::vector<std::string>& matches);
  std::string getBriefMessage() const;

private:
  std::string foodName_;
  std::string oldUnitPattern_;
  std::vector<std::string> matches_;
};

/** Could not change unit for food because pattern did not
 * match exactly one unit.
 */
class ChangeUnitError : public errors::NonFatal
{
public:
  ChangeUnitError(const food::Food& food,
                  const std::string& pattern,
                  const std::vector<std::string> matches);
  std::string getBriefMessage() const;

private:
  std::string foodName_;
  std::string pattern_;
  std::vector<std::string> matches_;
};

/** Could not delete arbitrary unit because it is the food's
 * current unit.
 */
class DeleteArbUnitError : public errors::NonFatal
{
public:
  DeleteArbUnitError(const food::Food& food,
                     const std::string& unitName);
  std::string getBriefMessage() const;

private:
  std::string foodName_;
  std::string unitName_;
};

/** Could not rename arbitrary unit because it is the food's
 * current unit.
 */
class RenameArbUnitError : public errors::NonFatal
{
public:
  RenameArbUnitError(const food::Food& food,
                     const std::string& unitName);
  std::string getBriefMessage() const;

private:
  std::string foodName_;
  std::string unitName_;
};


/** Could not perform change by nutrient amount because
 * pattern did not match exactly one nutrient name.
 */
class ByNutMatchError : public errors::NonFatal
{
public:
  ByNutMatchError(const food::Food& food,
                  const std::string& pattern,
                  const std::vector<std::string>& matchingNames);
  std::string getBriefMessage() const;

private:
  std::string foodName_;
  std::string pattern_;
  std::vector<std::string> matchingNames_;
};

/** Could not perform change by nutrient amount because
 * food has zero quantity of that nutrient.
 */
class ByNutZeroNutError : public errors::NonFatal
{
public:
  ByNutZeroNutError(const food::Food& food,
                    const std::string& nutName);
  std::string getBriefMessage() const;

private:
  std::string foodName_;
  std::string nutName_;
};

/** Could not initialize changer because an ingredient
 * collection name given does not match exactly one
 * collection.
 */
class IngredientCollectionNotFoundError : public errors::NonFatal
{
public:
  IngredientCollectionNotFoundError
    (const std::vector<std::string>& matchingCollNames,
     const std::string& abbreviation);
  std::string getBriefMessage() const;

private:
  std::vector<std::string> matchingCollNames_;
  std::string abbreviation_;
};

} // namespace changer


#endif
