#include "collections.h"

#include "stl.h"
#include "strings.h"

#include <algorithm>
#include <vector>
#include <string>

namespace collections
{

namespace
{
  
StringVec getCollNamesFromAbbrevs(const StringVec& abbreviations,
                                  StringVec collNames)
{
  StringVec result;
  std::sort(collNames.begin(), collNames.end());
  for (StringVec::const_iterator abbrev = abbreviations.begin();
       abbrev != abbreviations.end(); ++abbrev)
  {
    if (std::binary_search(collNames.begin(), collNames.end(), *abbrev))
      result.push_back(*abbrev);
    else
    {
      StringVec matches = strings::matchAbbrev(collNames, *abbrev);
      if (matches.size() != 1)
        throw BadAbbreviationError(*abbrev, matches);
      result.push_back(matches[0]);
    }
  }
  return result;
}

} // unnamed namespace

CollectionConstPtrs
getCollectionPtrs(const Collections& collections,
                  const std::vector<std::string>& abbreviations)
{
  CollectionConstPtrs result;
  StringVec collNames = getCollNamesFromAbbrevs(abbreviations,
                                                stl::getKeys(collections));
  for (StringVec::const_iterator collName = collNames.begin();
       collName != collNames.end(); ++collName)
  {
    Collections::const_iterator collection = collections.find(*collName);
    result.push_back(&*collection);
  }
  return result;
}

CollectionPtrs
getCollectionPtrs(Collections& collections,
                  const std::vector<std::string>& abbreviations)
{
  CollectionPtrs result;
  StringVec collNames = getCollNamesFromAbbrevs(abbreviations,
                                                stl::getKeys(collections));
  for (StringVec::const_iterator collName = collNames.begin();
       collName != collNames.end(); ++collName)
  {
    Collections::iterator collection = collections.find(*collName);
    result.push_back(&*collection);
  }
  return result;
}

FoodPtrs getFoodPtrs(Collections& collections)
{
  FoodPtrs result;
  for (Collections::iterator nameAndList = collections.begin();
       nameAndList != collections.end(); ++nameAndList)
  {
    FoodList& foodList = nameAndList->second;
    for (FoodList::iterator food = foodList.begin();
         food != foodList.end(); ++food)
    {
      result.push_back(&*food);
    }
  }
  return result;
}

FoodConstPtrs getFoodPtrs(const Collections& collections)
{
  FoodConstPtrs result;
  for (Collections::const_iterator nameAndList = collections.begin();
       nameAndList != collections.end(); ++nameAndList)
  {
    const FoodList& foodList = nameAndList->second;
    for (FoodList::const_iterator food = foodList.begin();
         food != foodList.end(); ++food)
    {
      result.push_back(&*food);
    }
  }
  return result;
}

FoodConstPtrs
getFoodPtrs(const FoodVecMap& foodVecMap)
{
  FoodConstPtrs result;
  for (FoodVecMap::const_iterator nameAndFoodVec = foodVecMap.begin();
       nameAndFoodVec != foodVecMap.end(); ++nameAndFoodVec)
  {
    const FoodVec& foodVec = nameAndFoodVec->second;
    for (FoodVec::const_iterator food = foodVec.begin();
         food != foodVec.end(); ++food)
    {
      result.push_back(&*food);
    }
  }
  return result;
}

BadAbbreviationError::BadAbbreviationError
                      (const std::string& badAbbrev,
                       const std::vector<std::string>& matches)
  : badAbbrev_(badAbbrev), matches_(matches) {}

std::string BadAbbreviationError::getBriefMessage() const
{
    std::string result = badAbbrev_;
    result.append(": bad collection name abbreviation. ");
    if (matches_.empty())
        result.append("No matches.");
    else
    {
        result.append("Multiple matches:");
        for (std::vector<std::string>::const_iterator match =
                matches_.begin(); match != matches_.end();
                ++match)
        {
            result.append("\n");
            result.append(*match);
        }
    }
    return result;
}

} // namespace collections
