#include "comm.h"

#include "pointers.h"
#include "strings.h"

#include <iostream>
#include <string>
#include <sstream>
#include <vector>

#include <errno.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <pwd.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

namespace comm
{

namespace
{

std::string getSocketFilename()
{
  char* envFilename = getenv("PANTRY_SOCKET");
  if (envFilename != NULL)
    return envFilename;

  const std::string FILENAME = ".pantrySocket";

  std::string home;
  char* homeEnv = getenv("HOME");
  if (homeEnv != NULL)
    home = std::string(homeEnv);
  else
  {
    passwd* userinfo = getpwuid(getuid());
    if (userinfo == NULL)
        throw GetHomeDirError(errno);
    home = std::string(userinfo->pw_dir);
  }
  return strings::pathJoin(home, FILENAME);
}

void cleanup(int sig)
{
  unlink(getSocketFilename().c_str());
  signal(sig, SIG_DFL);
  raise(sig);
}

std::string readString(std::istream& is)
{
  unsigned long int len = 0;
  is >> len;
  is.ignore();
  std::string result;
  result.reserve(len);
  for (unsigned long int i=0; i < len; ++i)
  {
    char c;
    is.get(c);
    result.append(1, c);
  }
  is.ignore();
  return result;
}

void saveString(std::ostream& os,
                const std::string& toSave)
{
  os << toSave.size() << ' ' << toSave << ' ';
}

} // unnamed namespace

Request::Request(const std::string& clientCurrDir,
                 const std::vector<std::string>& args)
  : clientCurrDir_(clientCurrDir), args_(args) {}

Request::Request(std::istream& is)
{
  clientCurrDir_ = readString(is);

  unsigned int nArgs = 0;
  is >> nArgs;
  args_.reserve(nArgs);
  for (unsigned long int i=0; i < nArgs; ++i)
  {
    args_.push_back(readString(is));
  }
}

void Request::serialize(std::ostream& os) const
{
  saveString(os, clientCurrDir_);
  os << args_.size() << ' ';
  for (StringVec::const_iterator string = args_.begin();
       string != args_.end(); ++string)
  {
    saveString(os, *string);
  }
}

std::string Request::getClientCurrDir() const
{
  return clientCurrDir_;
}

std::vector<std::string> Request::getClientArgs() const
{
  return args_;
}

Response::Response(int returnCode,
                   bool quit,
                   const std::string& output,
                   const std::string& errors)
  : quit_(quit), returnCode_(returnCode),
    output_(output), errors_(errors) {}

Response::Response(std::istream& is)
{
  int quitNo;
  is >> quitNo;
  if (quitNo == 1)
    quit_ = true;
  else
    quit_ = false;
  is >> returnCode_;
  output_ = readString(is);
  errors_ = readString(is);
}

void Response::serialize(std::ostream& os) const
{
  signal(SIGPIPE, SIG_IGN);
  if (quit_)
    os << 1 << ' ';
  else
    os << 0 << ' ';
  os << returnCode_ << ' ';
  saveString(os, output_);
  saveString(os, errors_);
  // don't throw if anything goes wrong, as this
  // runs in a destructor
}
  
int Response::getReturnCode() const
{
  return returnCode_;
}

std::string Response::getOutput() const
{
  return output_;
}

std::string Response::getErrors() const
{
  return errors_;
}

bool Response::getQuit() const
{
  return quit_;
}

ClientSocket::ClientSocket(const Request& outbound)
{
  fd::Fdbuf fdbuf(getSocketFilename());
  std::iostream stream(&fdbuf);
  stream.tie(&stream);
  outbound.serialize(stream);
  if (! stream.good())
    throw SendRequestEncodeError();

  response_.reset(new Response(stream));
  if (! stream.good())
    throw ReceiveResponseDecodeError();
  fdbuf.close();
}

Response ClientSocket::getResponse() const
{
  return *response_;
}

Bottle::Bottle(NamedSocket& namedSocket)
  : fdbuf_(namedSocket.listener_), stream_(&fdbuf_),
    request_(stream_)
{
  if (! stream_.good())
    throw GetRequestDecodeError();
}

Bottle::~Bottle()
{
  response_->serialize(stream_);
  try
  {
    fdbuf_.close();
  }
  catch (errors::Pantry& e)
  {
    std::cerr << "pantryd: warning: " << e.getBriefMessage() << std::endl;
  }
  if (! stream_.good())
    std::cerr << "pantryd: warning: error when encoding "
                 "message to client\n";
}

Request Bottle::getRequest() const
{
  return request_;
}

void Bottle::setResponse(std::auto_ptr<Response> response)
{
  response_ = response;
}

NamedSocket::NamedSocket()
  : listener_(getSocketFilename())
{
  signal(SIGHUP, &cleanup);
  signal(SIGINT, &cleanup);
  signal(SIGTERM, &cleanup);
}

NamedSocket::~NamedSocket()
{
  unlink(getSocketFilename().c_str());
}

GetHomeDirError::GetHomeDirError(int errorNumber)
    : errorNumber_(errorNumber) {}

std::string GetHomeDirError::getBriefMessage() const
{
    std::string result = "could not determine home directory";
    result.append(": ");
    result.append(strerror(errorNumber_));
    return result;
}

std::string SendRequestEncodeError::getBriefMessage() const
{
    std::string result = "error while encoding request to server";
    return result;
}

std::string ReceiveResponseDecodeError::getBriefMessage() const
{
    std::string result = "error while decoding response";
    return result;
}

std::string GetRequestDecodeError::getBriefMessage() const
{
    std::string result = "error while decoding request";
    return result;
}

std::string SendResponseEncodeError::getBriefMessage() const
{
    std::string result = "error while encoding response";
    return result;
}

} // namespace comm
