#ifndef COMM_H
#define COMM_H

#include "fd.h"
#include "errors.h"
#include "pointers.h"

#include <istream>
#include <ostream>
#include <memory>
#include <streambuf>
#include <string>
#include <vector>

namespace comm {

typedef std::vector<std::string> StringVec;

class Request
{
public:
  Request(const std::string& clientCurrDir,
          const StringVec& args);
  Request(std::istream& is);
  void serialize(std::ostream& os) const;
  std::string getClientCurrDir() const;
  StringVec getClientArgs() const;

private:
  std::string              clientCurrDir_;
  StringVec args_;
};

class Response
{
public:
  Response(int returnCode,
           bool quit,
           const std::string& output,
           const std::string& errors);
  Response(std::istream& is);
  void serialize(std::ostream& os) const;

  int getReturnCode() const;
  std::string getOutput() const;
  std::string getErrors() const;
  bool getQuit() const;

private:
  bool quit_;
  int returnCode_;
  std::string output_;
  std::string errors_;
};

class ClientSocket
{
public:
  ClientSocket(const Request& outbound);
  Response getResponse() const;

private:
  pointers::Shared<const Response> response_;
};

class NamedSocket;

/** This contains an fdbuf, so it is noncopyable
 */
class Bottle
{
public:
  Bottle(NamedSocket& namedSocket);
  ~Bottle();
  Request getRequest() const;
  void setResponse(std::auto_ptr<Response> response);

private:
  fd::Fdbuf fdbuf_;
  // istream: client to server; ostream: server to client
  std::iostream stream_;
  Request request_;
  std::auto_ptr<Response> response_;
};

class NamedSocket
{
public:
  NamedSocket();
  ~NamedSocket();

  friend class Bottle;

private:
  fd::Listener listener_;
};

/** Could not determine the location of the home directory.
 */
class GetHomeDirError : public errors::Fatal
{
public:
    GetHomeDirError(int errorNumber);
    std::string getBriefMessage() const;

private:
    int errorNumber_;
};

class SendRequestEncodeError : public errors::Fatal
{
public:
    std::string getBriefMessage() const;

private:
    int errorNumber_;
};

class ReceiveResponseDecodeError : public errors::Fatal
{
public:
    std::string getBriefMessage() const;
};

class GetRequestDecodeError : public errors::Fatal
{
public:
    std::string getBriefMessage() const;
};

class SendResponseEncodeError : public errors::Fatal
{
public:
    std::string getBriefMessage() const;
};

} // namespace comm

#endif
