#include "db.h"

#include "adder.h"
#include "absUnitPool.h"
#include "changer.h"
#include "food.h"
#include "reporter.h"
#include "searcher.h"
#include "stl.h"

#include <algorithm>
#include <iomanip>
#include <list>
#include <map>
#include <sstream>
#include <string>
#include <vector>

namespace db
{

Db::Db()
  : strPool_(new strPool::StrPool())
{
  absUnitPool::AbsUnit unit;
  unit.name = "g";
  unit.value = numbers::Number(1.0);
  absUnitPool_.reset(new absUnitPool::AbsUnitPool(unit));
  unit.name = "oz";
  unit.value = numbers::Number(28.35);
  absUnitPool_->addAbsUnit(unit);
  unit.name = "lb";
  unit.value = numbers::Number(453.6);
  absUnitPool_->addAbsUnit(unit);
}

Db::FoodVecMap Db::doChanges(const searcher::SelectionsVecs& selections,
                             const changer::Changer& changer)
{
  FoodVecMap result;
  for (searcher::SelectionsVecs::const_iterator collNameAndSelections =
       selections.begin(); collNameAndSelections != selections.end();
       ++collNameAndSelections)
  {
    const std::string& collName = collNameAndSelections->first;
    const searcher::SelectionsVec& selectionsVec =
      collNameAndSelections->second;
    FoodVec& foods = result[collName];
    for (searcher::SelectionsVec::const_iterator food =
         selectionsVec.begin();
         food != selectionsVec.end();
         ++food)
    {
      foods.push_back(changer(**food));
    }
  }
  return result;
}

std::string Db::process(const searcher::Searcher& searcher,
                        const changer::Changer& changer,
                        bool editOriginals,
                        bool deleteOriginals,
                        const adder::Adder& adder,
                        const reporter::Reporter& reporter)
{
  searcher::SelectionsVecs selections = searcher.process(collections_);
  FoodVecMap changedFoods = doChanges(selections, changer);
  //////////////////////////////////////////////////////////
  // THROW BOUNDARY - statements below this line must
  // not throw Pantry exceptions
  //////////////////////////////////////////////////////////
  doEditsOrDeletes(selections, changedFoods, 
                   editOriginals, deleteOriginals);
  adder.process(collections_, changed_,
                collections::getFoodPtrs(changedFoods));
  return reporter.print(changedFoods);
}

std::string Db::process(const changer::Changer& changer,
                        const adder::Adder& adder,
                        const reporter::Reporter& reporter)
{
  food::Food food(strPool_, absUnitPool_);
  food::Food changed(changer(food));
  //////////////////////////////////////////////////////////
  // THROW BOUNDARY - statements below this line must
  // not throw Pantry exceptions
  //////////////////////////////////////////////////////////
  adder.process(collections_, changed_,
                collections::FoodConstPtrs(1, &changed));
  return reporter.print(changed);
}

void Db::clearChanges()
{
  std::vector<std::string> collNames = stl::getKeys(collections_);
  std::sort(collNames.begin(), collNames.end());

  typedef std::map<std::string, bool>::iterator unsavedIt;
  std::vector<unsavedIt> itsToErase;
  for (unsavedIt nameAndBool = changed_.begin();
       nameAndBool != changed_.end(); ++nameAndBool)
  {
    const std::string& name = nameAndBool->first;
    if (std::binary_search(collNames.begin(), collNames.end(),
                           name))
    {
      nameAndBool->second = false;
    }
    else
      itsToErase.push_back(nameAndBool);
  }
  
  for (std::vector<unsavedIt>::const_iterator itToErase =
       itsToErase.begin(); itToErase != itsToErase.end();
       ++itToErase)
  {
    changed_.erase(*itToErase);
  }
}

void Db::swap(Db& rhs)
{
  strPool_.swap(rhs.strPool_);
  absUnitPool_.swap(rhs.absUnitPool_);
  collections_.swap(rhs.collections_);
  changed_.swap(rhs.changed_);
}

std::vector<std::string> Db::getCollectionNames() const
{
  return stl::getKeys(collections_);
}

std::list<food::Food> Db::getFoodList(const std::string& collName) const
{
  return stl::checkedGet(collections_, collName);
}

void Db::doEditsOrDeletes(const searcher::SelectionsVecs& selections,
                          const FoodVecMap& changedFoods,
                          bool doEdits,
                          bool doDeletes)
{
  if (! doEdits && ! doDeletes)
    return;
  for (searcher::SelectionsVecs::const_iterator collNameAndSelVec =
       selections.begin(); collNameAndSelVec != selections.end();
       ++collNameAndSelVec)
  {
    const std::string& collName = collNameAndSelVec->first;
    const FoodVec& changedFoodsVec = stl::checkedGet(changedFoods, collName);
    searcher::SelectionsVec selectionsVec = collNameAndSelVec->second;
    bool& changedBool = stl::checkedGet(changed_, collName);

    collections::FoodList& foodList =
      stl::checkedGet(collections_, collName);
    for (searcher::SelectionsVec::size_type i=0; i < selectionsVec.size();
         ++i)
    {
      std::list<food::Food>::iterator foodToEditIt =
        stl::checkedAt(selectionsVec, i);

      food::Food& foodToEdit = *foodToEditIt;
      const food::Food& changedFood = stl::checkedAt(changedFoodsVec, i);
      changedBool = true;
      if (doEdits)
        foodToEdit = changedFood;
      if (doDeletes)
        foodList.erase(foodToEditIt);
    }
    if (foodList.empty())
      collections_.erase(collName);
  }
}

std::string Db::getStatus() const
{
  const unsigned int NUMBER_WIDTH = 6;
  std::ostringstream os;
  os << "Collections: ";
  if (changed_.empty())
      os << "(none) \n";
  else
  {
    os << '\n';
    for (std::map<std::string, bool>::const_iterator changedPair =
         changed_.begin();
         changedPair != changed_.end(); ++changedPair)
    {
      os << std::right << std::setw(NUMBER_WIDTH);
      collections::Collections::const_iterator collection =
          collections_.find(changedPair->first);
      unsigned int nFoods = 0;
      if (collection != collections_.end())
          nFoods = collection->second.size();
      
      os << nFoods << ' ' << std::left << changedPair->first << ' ';
      if (changedPair->second)
          os << "(changed)";
      os << '\n';
    }
  }
  return os.str();
}

void Db::merge(const Db& rhs)
{
  for (collections::Collections::const_iterator collNameAndList =
       rhs.collections_.begin();
       collNameAndList != rhs.collections_.end(); ++collNameAndList)
  {
    std::list<food::Food>& destList = collections_[collNameAndList->first];
    const std::list<food::Food>& fromList = collNameAndList->second;
    for (std::list<food::Food>::const_iterator food = fromList.begin();
         food != fromList.end(); ++food)
    {
      destList.push_back(*food);
    }
    changed_[collNameAndList->first] = true;
  }
}

} // namespace db
