#ifndef DB_H
#define DB_H

#include "absUnitPool.h"
#include "adder.h"
#include "changer_fwd.h"
#include "collections.h"
#include "food.h"
#include "pointers.h"
#include "reporter.h"
#include "searcher.h"
#include "serializer_fwd.h"
#include "strPool.h"

#include <iosfwd>
#include <list>
#include <map>
#include <string>
#include <vector>

namespace db
{

class Db
{
public:

  /** Initialize a blank database.
   */
  Db();

  /** Search current collections for foods, change them,
   * edit them and delete them (if called for), add them to
   * collections, and get a report.
   *
   * \return A string holding the report.
   *
   * Offers basic guarantee.
   */
  std::string process(const searcher::Searcher& searcher,
                      const changer::Changer& changer,
                      bool editOriginals,
                      bool deleteOriginals,
                      const adder::Adder& adder,
                      const reporter::Reporter& reporter);

  /** Create a new food, change it, add it to collections,
   * and get a report.
   *
   * \return A string holding the report.
   *
   * Offers basic guarantee.
   */
  std::string process(const changer::Changer& changer,
                      const adder::Adder& adder,
                      const reporter::Reporter& reporter);

  /** Get a string useful for a report showing collections
   * which are unchanged.
   *
   * Offers strong guarantee.
   */
  std::string getStatus() const;

  /** Reset the map indicating which collections have been
   * changed, so that every collection in the database has
   * no changed foods indicated. Also removes collections
   * from the map that are no longer in the database.
   *
   * Never throws.
   */
  void clearChanges();

  /** Get a vector of the names of the collections inside
   * this db.
   */
  std::vector<std::string> getCollectionNames() const;

  /** Get a collection that is inside this db.
   *
   * Offers strong guarantee.
   */
  std::list<food::Food> getFoodList(const std::string& collName) const;

  /** Swap contents with another database.
   *
   * Never throws.
   */
  void swap(Db& rhs);

  /** Merge this database with another database.
   *
   * Not exception safe.
   */
  void merge(const Db& rhs);

  friend class serializer::Loader;
  friend class serializer::Saver;

private:
  pointers::Shared<strPool::StrPool> strPool_;
  pointers::Shared<absUnitPool::AbsUnitPool> absUnitPool_;

  /** This map does not contain empty food lists.
   */
  collections::Collections collections_;

  /** The changed map always holds the name of every
   * collection in the database. If every food in a
   * collection is erased, but has not yet been written to
   * disk, that collection name will be in here as well.
   */
  std::map<std::string, bool> changed_;

  typedef std::vector<food::Food> FoodVec;
  typedef std::map<std::string, FoodVec > FoodVecMap;

  FoodVecMap doChanges(const searcher::SelectionsVecs& selections,
                       const changer::Changer& changer);

  void doEditsOrDeletes(const searcher::SelectionsVecs& selections,
                        const FoodVecMap& changed,
                        bool doEdits,
                        bool doDeletes);
};

} // namespace db

#endif
