#include "dbholder.h"

#include "comm.h"
#include "meta.h"
#include "inputs.h"
#include "parser.h"
#include "serializer.h"
#include "strings.h"

#include <memory>
#include <sstream>
#include <string>
#include <iomanip>

namespace
{

} // unnamed namespace

namespace dbHolder
{

DbHolder::DbHolder()
  : hasCurrFilename_(false) {}

std::auto_ptr<comm::Response>
DbHolder::processRequest(const comm::Request& request)
{
  std::auto_ptr<comm::Response> response;
  try
  {
    response = process(request);
  }
  catch (errors::NonFatal& e)
  {
    std::string eMsg = "pantry: error: ";
    eMsg.append(e.getBriefMessage());
    eMsg.append("\n");
    response.reset(new comm::Response(1, false, "", eMsg));
  }
  return response;
}

std::auto_ptr<comm::Response>
DbHolder::process(const comm::Request& request)
{
  parser::Result parsed(inputs::getParsed(request.getClientArgs()));
  inputs::Inputs inputs(parsed, db_);
  bool newHasCurrFilename = hasCurrFilename_;
  std::string newCurrFilename = currFilename_;

  if (inputs.doOpen)
  {
    open(db_, request.getClientCurrDir(), parsed, 
         inputs, newHasCurrFilename,
         newCurrFilename);
  }
  if (inputs.doRead)
    read(db_, request.getClientCurrDir(), parsed, inputs);
  if (inputs.doClose)
    close(db_, newHasCurrFilename);
  if (inputs.doSave)
    save(db_);
  if (inputs.doSaveAs)
    saveAs(db_, inputs.saveAsFilename, request.getClientCurrDir(),
           newHasCurrFilename, newCurrFilename);
  
  std::string output;
  if (inputs.doCreate)
    output = db_.process(inputs.changer, inputs.adder, inputs.reporter);
  else
    output = db_.process(inputs.searcher, inputs.changer,
                            inputs.doEdit, inputs.doDelete,
                            inputs.adder, inputs.reporter);
  output = addMeta(output, inputs, db_);

  std::auto_ptr<comm::Response>
    result(new comm::Response(0, inputs.doQuit, output, ""));

  hasCurrFilename_ = newHasCurrFilename;
  currFilename_.swap(newCurrFilename);
  return result;
}

void DbHolder::open(db::Db& db,
                    const std::string& clientCurrDir,
                    const parser::Result& parsed,
                    inputs::Inputs& inputs,
                    bool& newHasCurrFilename,
                    std::string& newCurrFilename)
{
  std::string filename = strings::getAbsoluteFilename(clientCurrDir,
                                                inputs.openFilename);
  db::Db newDb = serializer::load(filename);
  inputs = inputs::Inputs(parsed, newDb);
  db.swap(newDb);
  newHasCurrFilename = true;
  newCurrFilename = filename;
}

void DbHolder::read(db::Db& db,
                    const std::string& clientCurrDir,
                    const parser::Result& parsed,
                    inputs::Inputs& inputs)
{
  std::string filename =
    strings::getAbsoluteFilename(clientCurrDir, inputs.readFilename);
  db::Db readDb = serializer::load(filename);
  db.merge(readDb);
  inputs = inputs::Inputs(parsed, db);
}

void DbHolder::close(db::Db& db, bool& newHasCurrFilename)
{
  db = db::Db();
  newHasCurrFilename = false;
}

void DbHolder::save(db::Db& db)
{
  if (! hasCurrFilename_)
    throw NoSaveFilenameError();
  serializer::save(db, currFilename_);
  db.clearChanges();
}

void DbHolder::saveAs(db::Db& db, const std::string& saveFilename,
                      const std::string& clientCurrDir,
                      bool& newHasCurrFilename,
                      std::string& newCurrFilename)
{
  std::string filename =
    strings::getAbsoluteFilename(clientCurrDir, saveFilename);
  serializer::save(db, filename);
  newHasCurrFilename = true;
  newCurrFilename = filename;
}

std::string DbHolder::addMeta(const std::string& output,
                              const inputs::Inputs& inputs,
                              const db::Db& db)
{
  std::string result = output;
  if (inputs.doHelp)
    result.append(meta::HELP);
  if (inputs.doVersion)
    result.append(meta::VER);
  if (inputs.doCopyright)
  {
    result.append(meta::VER);
    result.append(meta::COPYRIGHT);
  }
  if (inputs.doStatus)
  {
    result.append("Current filename: ");
    if (hasCurrFilename_)
    {
      result.append(currFilename_);
      result.append("\n");
    }
    else
      result.append("(none)\n");
    result.append(db.getStatus());
  }
  return result;
}

std::string NoSaveFilenameError::getBriefMessage() const
{
  return "cannot save database without a filename";
}


} // namespace dbHolder
