#ifndef DBHOLDER_H
#define DBHOLDER_H

#include "comm.h"
#include "db.h"
#include "errors.h"
#include "inputs.h"
#include "parser.h"

#include <memory>
#include <string>

namespace dbHolder
{


class DbHolder
{
public:

  DbHolder();

  /** Process a request. Makes necessary changes to
   * databases, and return a response. Catches all
   * errors::nonFatal exceptions.
   *
   * This returns an auto_ptr (rather than a simple
   * automatic object) for exception safety reasons. If this
   * returned a comm::Response, it would be possible for the
   * return statement itself to throw, because copying a
   * comm::Response could conceivably throw. (This almost
   * certainly wouldn't happen, for a variety of reasons,
   * but it would be possible.) By instead returning a
   * pointer, the only thing copied in the return is a
   * pointer, an operation that cannot throw. This can't be
   * accomplished by using a parameter, because
   * comm::Response does not have a default constructor.
   */
  std::auto_ptr<comm::Response>
  processRequest(const comm::Request& request);

private:
  db::Db db_;
  bool hasCurrFilename_;
  std::string currFilename_;
  
  std::auto_ptr<comm::Response> process(const comm::Request& request);

  void open(db::Db& db,
            const std::string& clientCurrDir,
            const parser::Result& parsed,
            inputs::Inputs& inputs,
            bool& newHasCurrFilename,
            std::string& newCurrFilename);

  void read(db::Db& db,
            const std::string& clientCurrDir,
            const parser::Result& parsed,
            inputs::Inputs& inputs);

  void close(db::Db& db, bool& newHasCurrFilename);

  void save(db::Db& db);

  void saveAs(db::Db& db, const std::string& saveFilename,
              const std::string& clientCurrDir,
              bool& newHasCurrFilename,
              std::string& newCurrFilename);

  std::string addMeta(const std::string& output,
                      const inputs::Inputs& inputs,
                      const db::Db& db);
};

class NoSaveFilenameError : public errors::NonFatal
{
public:
  std::string getBriefMessage() const;
};


} // namespace dbHolder

#endif
