#ifndef ERRORS_H
#define ERRORS_H

#include <string>

/** Pantry exception classes.
 *
 * Here is an explanation for the rationale for exception
 * safety in Pantry. First of all, Pantry is not exception
 * safe in the sense that it makes no guarantees such as the
 * strong guarantee, basic guarantee, etc. Making these
 * sorts of guarantees would require scrutinizing every
 * function that dynamically allocates memory; because most
 * uses of the C++ standard library containers dynamically
 * allocate memory, this means practically every function in
 * Pantry would cause exception safety issues. For this
 * small system, the complexity that such exception safety
 * would cause outweighs the benefits. This is especially
 * true when one considers that if Pantry cannot dynamically
 * allocate memory, there's little it can do but fail.
 *
 * Here are the three main classes of Pantry exceptions:
 *
 * errors::NonFatal is an error that should NOT terminate
 * the server. Instead, these errors should report an error
 * to the user and leave the server running. Examples
 * include erroneous user command-line input, or
 * file-not-found errors.
 *
 * errors::Fatal is an error that terminates the server.
 * Examples include extraordinary i/o errors, such a failure
 * of the Unix socket(2) call.
 *
 * errors::Logic are errors that should NEVER happen. They
 * indicate a flaw in the program logic. assert could have
 * been used for this, but I see no benefit in removing
 * these assertions from "production" code when the
 * performance benefit is practically zero.
 *
 * Pantry's code is designed so that any errors::NonFatal
 * instances will be thrown BEFORE any changes are made to
 * the database that is in memory. This way, if there are
 * any non-fatal errors, the database in memory will be in a
 * predictable state--that is, it will be unchanged.
 * errors::NonFatal will be the errors that a Pantry user
 * encounters the vast majority of the time. For this to
 * work, all functions that throw errors::NonFatal must be
 * called BEFORE any changes are committed to the database
 * in memory. db.cc has two "throw boundaries" below which
 * no Pantry exceptions are thrown.
 *
 * All errors::Fatal, errors::Logic, and exceptions not
 * derived from errors::Pantry will simply cause the server
 * to terminate. Realistically the server could not do
 * anything to recover from these errors anyway.
 *
 */
namespace errors {

/** Root class for all Pantry exceptions.
 */
struct Pantry
{
    /** Returns a short string that can be used for an error
     * message.
     */
    virtual std::string getBriefMessage() const = 0;

    virtual ~Pantry();
};

struct NonFatal : public Pantry {};
struct Fatal : public Pantry {};

/** Something that should never happen happened anyway. For example:
 * a value in a std::map was not found, even though it should always
 * be found. 
 */
struct Logic
{
    std::string getBriefMessage() const;
};

} // namespace errors

#endif
