#include "fd.h"

#include <string>

#include <errno.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/un.h>
#include <string.h>
#include <unistd.h>

namespace fd
{

namespace
{

sockaddr_un getAddress(const std::string& filename)
{
  sockaddr_un result;
  memset(&result, 0, sizeof(struct sockaddr_un));
  result.sun_family = AF_UNIX;
  if (filename.size() > sizeof(result.sun_path) - 1)
      throw SocketNameTooLongError(filename);
  strncpy(result.sun_path, filename.c_str(),
          sizeof(result.sun_path) - 1);
  return result;
}

} // unnamed namespace

Fdbuf::Fdbuf(const std::string& filename, Flag flag)
  : closed_(false), filename_(filename)
{
  int openFlags;
  if (flag == Read)
    openFlags = O_RDONLY;
  else
    openFlags = O_WRONLY | O_CREAT | O_TRUNC;
  fd_ = open(filename.c_str(), openFlags,
             S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
  if (fd_ == -1)
    throw FileOpenError(flag, filename, errno);
  setg(inSeq_, inSeq_, inSeq_);
  setp(outSeq_, outSeq_ + (BUFSIZE - 1));
}

Fdbuf::Fdbuf(const std::string& pathToSocket)
  : closed_(false), filename_(pathToSocket)
{
  fd_ = socket(AF_UNIX, SOCK_STREAM, 0);
  if (fd_ == -1)
    throw SocketCreationError(errno);

  sockaddr_un remote = getAddress(pathToSocket);
  int len = strlen(remote.sun_path) + sizeof(remote.sun_family);
  int result = connect(fd_, 
                       reinterpret_cast<sockaddr*>(&remote), len);
  if (result == -1)
  {
    int errorCode = errno;
    ::close(fd_);
    throw SocketConnectionError(pathToSocket, errorCode);
  }
  setg(inSeq_, inSeq_, inSeq_);
  setp(outSeq_, outSeq_ + (BUFSIZE - 1));
}

Fdbuf::Fdbuf(Listener& listener)
  : closed_(false), filename_(listener.socketPath_)
{
  if (listener.closed_)
    throw errors::Logic();
  fd_ = accept(listener.fd_, NULL, NULL);
  if (fd_ == -1)
    throw ConnectToClientError(filename_, errno);
  setg(inSeq_, inSeq_, inSeq_);
  setp(outSeq_, outSeq_ + (BUFSIZE - 1));
}

void Fdbuf::close()
{
  pubsync();
  int closeResult = ::close(fd_);
  if (closeResult == -1)
    throw CloseError(filename_, errno);
  closed_ = true;
}

int Fdbuf::underflow()
{
  ssize_t readResult = read(fd_, inSeq_, BUFSIZE);
  if (readResult == 0)
    return traits_type::eof();
  if (readResult == -1)
    throw InputError(errno, filename_);
  setg(inSeq_, inSeq_, inSeq_ + readResult);
  return *inSeq_;
}

Fdbuf::~Fdbuf()
{
  if (closed_)
    return;
  try
  {
    close();
  }
  catch (...) {}
}

int Fdbuf::overflow(int c)
{
  ssize_t sum = 0;
  char* end = pptr();
  if (c != traits_type::eof())
  {
    ++end;
    *pptr() = traits_type::to_char_type(c);
  }

  while (sum < end - pbase())
  {
    ssize_t n = write(fd_, pbase() + sum, end - (pbase() + sum));
    if (n == -1)
      throw OutputError(errno, filename_);
    sum += n;
  }

  setp(outSeq_, outSeq_ + (BUFSIZE - 1));
  return traits_type::not_eof(c);
}

int Fdbuf::sync()
{
  // no need to check return value of overflow, because if
  // it returns at all it will return not_eof
  overflow();
  return 0;
}

Listener::Listener(const std::string& socketPath)
  : socketPath_(socketPath), closed_(false)
{
  fd_ = socket(AF_UNIX, PF_UNIX, 0);
  if (fd_ == -1)
    throw SocketCreationError(errno);
  sockaddr_un address = getAddress(socketPath);
  int bindResult = bind(fd_, (struct sockaddr *) &address,
                        sizeof(struct sockaddr_un));
  if (bindResult == -1)
  {
    ::close(fd_);
    throw BindError(socketPath_, errno);
  }

  int listenResult = listen(fd_, 5);
  if (listenResult)
  {
    int errorNumber = errno;
    ::close(fd_);
    throw ListenError(socketPath, errorNumber);
  }
}

void Listener::close()
{
  int closeResult = ::close(fd_);
  if (closeResult == -1)
    throw CloseError(socketPath_, errno);
}

Listener::~Listener()
{
  if (closed_)
    return;
  try
  {
    close();
  }
  catch (...) {}
}

InputError::InputError(int errorNumber, const std::string& filename)
  : errorNumber_(errorNumber), filename_(filename)
{
}

std::string InputError::getBriefMessage() const
{
  std::string result = "error while getting input from filename ";
  result.append(filename_);
  result.append(": " );
  result.append(strerror(errorNumber_));
  return result;
}

OutputError::OutputError(int errorNumber, const std::string& filename)
  : errorNumber_(errorNumber), filename_(filename)
{
}

std::string OutputError::getBriefMessage() const
{
  std::string result = "error while putting output to filename ";
  result.append(filename_);
  result.append(": " );
  result.append(strerror(errorNumber_));
  return result;
}

FileOpenError::FileOpenError(Flag flag, const std::string& filename,
              int errorNumber)
  : flag_(flag), filename_(filename), errorNumber_(errorNumber)
{
}

std::string FileOpenError::getBriefMessage() const
{
  std::string result = "error when opening file named ";
  result.append(filename_);
  result.append(" for ");
  if (flag_ == Read)
    result.append("reading: ");
  else
    result.append("writing: ");
  result.append(strerror(errorNumber_));
  return result;
}

SocketCreationError::SocketCreationError(int errorNumber)
  : errorNumber_(errorNumber)
{
}

std::string SocketCreationError::getBriefMessage() const
{
  std::string result = "could not create socket: ";
  result.append(strerror(errorNumber_));
  return result;
}

SocketConnectionError::SocketConnectionError
                       (const std::string& pathToSocket,
                        int errorNumber)
  : pathToSocket_(pathToSocket), errorNumber_(errorNumber)
{
}

std::string SocketConnectionError::getBriefMessage() const
{
  std::string result = "could not connect to socket named ";
  result.append(pathToSocket_);
  result.append(": ");
  result.append(strerror(errorNumber_));
  if (errorNumber_ == ECONNREFUSED || errorNumber_ == ENOENT)
    result.append(". Have you started the pantryd server?");
  return result;
}

ConnectToClientError::ConnectToClientError
                      (const std::string& pathToSocket,
                       int errorNumber)
  : pathToSocket_(pathToSocket), errorNumber_(errorNumber)
{
}

std::string ConnectToClientError::getBriefMessage() const
{
  std::string result = "could not connect to client socket at path ";
  result.append(pathToSocket_);
  result.append(": ");
  result.append(strerror(errorNumber_));
  return result;
}

CloseError::CloseError(const std::string& filename, int errorNumber)
  : filename_(filename), errorNumber_(errorNumber)
{
}

std::string CloseError::getBriefMessage() const
{
  std::string result = "could not close file named ";
  result.append(filename_);
  result.append(": ");
  result.append(strerror(errorNumber_));
  return result;
}

BindError::BindError(const std::string& pathToSocket, int errorNumber)
  : pathToSocket_(pathToSocket), errorNumber_(errorNumber)
{
}

std::string BindError::getBriefMessage() const
{
  std::string result = "could not bind socket to file named ";
  result.append(pathToSocket_);
  result.append(": ");
  result.append(strerror(errorNumber_));
  if (errorNumber_ == EADDRINUSE)
  {
    result.append(". Check to see if pantryd server is already "
                  "running; if not, run 'rm ");
    result.append(pathToSocket_);
    result.append("' and try again.");
  }
  return result;
}

ListenError::ListenError(const std::string& pathToSocket, int errorNumber)
  : pathToSocket_(pathToSocket), errorNumber_(errorNumber)
{
}

std::string ListenError::getBriefMessage() const
{
  std::string result = "could not listen on socket named ";
  result.append(pathToSocket_);
  result.append(": ");
  result.append(strerror(errorNumber_));
  return result;
}

SocketNameTooLongError::SocketNameTooLongError
                        (const std::string& pathToSocket)
  : pathToSocket_(pathToSocket)
{
}

std::string SocketNameTooLongError::getBriefMessage() const
{
  std::string result = "socket name ";
  result.append(pathToSocket_);
  result.append(" is too long");
  return result;
}


} // namespace fd
