#ifndef FD_H
#define FD_H

#include "errors.h"

#include <streambuf>
#include <string>

// This is entirely arbitrary. I thought it was small, but
// apparently the GNU C Library defaults to 8192 on my
// machine
#define BUFSIZE 64000

namespace fd
{

enum Flag
{
  Read,
  Write
};

class Fdbuf;

class Listener
{
public:
  Listener(const std::string& socketPath);
  void close();
  ~Listener();

  friend class Fdbuf;

private:
  int fd_;
  std::string socketPath_;
  bool closed_;

  // Do not copy
  Listener(const Listener&);
  Listener& operator=(const Listener&);
};

class Fdbuf : public std::streambuf
{
public:
  /** Create Fdbuf to read from or write to file descriptor.
   *
   * \param filename Filename to read from or write to.
   *
   * \param flag Read mode or write mode. (You can't do
   * both.)
   */
  Fdbuf(const std::string& filename, Flag flag);

  /** Creates fd that connects to the server listening at
   * pathToSocket. This Fdbuf will be in both read mode and
   * write mode.
   */
  Fdbuf(const std::string& pathToSocket);

  /** Accepts a connection from a socket that is listening.
   * This Fdbuf will be in both read mode and write mode.
   */
  Fdbuf(Listener& listener);

  /** Closes the underlying file descriptor. First calls
   * pubsync. After calling close, attempting further io
   * with this Fdbuf will fail.
   */
  void close();

  /** Closes the underlying file descriptor, if it has not
   * already been closed. Any exceptions that arise are
   * suppressed, so if you want to do something about them,
   * call close manually first.
   */
  ~Fdbuf();

protected:
  int sync();
  int underflow();
  int overflow(int c = traits_type::eof());

private:
  int fd_;
  bool closed_;
  char inSeq_[BUFSIZE];
  char outSeq_[BUFSIZE];
  std::string filename_;

  // Do not copy
  Fdbuf(const Fdbuf&);
  Fdbuf& operator=(const Fdbuf&);
};

class InputError : public errors::NonFatal
{
public:
  InputError(int errorNumber, const std::string& filename);
  std::string getBriefMessage() const;

private:
  int errorNumber_;
  std::string filename_;
};

class OutputError : public errors::NonFatal
{
public:
  OutputError(int errorNumber, const std::string& filename);
  std::string getBriefMessage() const;

private:
  int errorNumber_;
  std::string filename_;
};

class FileOpenError : public errors::NonFatal
{
public:
  FileOpenError(Flag flag, const std::string& filename,
                int errorNumber);
  std::string getBriefMessage() const;

private:
  Flag flag_;
  std::string filename_;
  int errorNumber_;
};

class SocketCreationError : public errors::Fatal
{
public:
  SocketCreationError(int errorNumber);
  std::string getBriefMessage() const;

private:
  int errorNumber_;
};

class SocketConnectionError : public errors::Fatal
{
public:
  SocketConnectionError(const std::string& pathToSocket,
                        int errorNumber);
  std::string getBriefMessage() const;

private:
  std::string pathToSocket_;
  int errorNumber_;
};

class ConnectToClientError : public errors::Fatal
{
public:
  ConnectToClientError(const std::string& pathToSocket,
                        int errorNumber);
  std::string getBriefMessage() const;

private:
  std::string pathToSocket_;
  int errorNumber_;
};

class CloseError : public errors::Fatal
{
public:
  CloseError(const std::string& filename, int errorNumber);
  std::string getBriefMessage() const;

private:
  std::string filename_;
  int errorNumber_;
};

class BindError : public errors::Fatal
{
public:
  BindError(const std::string& pathToSocket, int errorNumber);
  std::string getBriefMessage() const;

private:
  std::string pathToSocket_;
  int errorNumber_;
};

class ListenError : public errors::Fatal
{
public:
  ListenError(const std::string& pathToSocket, int errorNumber);
  std::string getBriefMessage() const;

private:
  std::string pathToSocket_;
  int errorNumber_;
};

class SocketNameTooLongError : public errors::Fatal
{
public:
  SocketNameTooLongError(const std::string& pathToSocket);
  std::string getBriefMessage() const;

private:
  std::string pathToSocket_;
};


} // namespace fd

#endif
