#include "food.h"

#include "absUnitPool.h"
#include "errors.h"
#include "numbers.h"
#include "pointers.h"
#include "strPool.h"

#include <string>
#include <utility>
#include <vector>

namespace food
{

void Food::renameNumberMapItem(const std::string& oldName,
                               const std::string& newName,
                               pointers::Shared<StrPtrNumMap>& mapPtr,
                               strPool::StrPool& strPool)
{
  strPool::StrPtr oldNamePtr = strPool.getStrPtr(oldName);
  if (oldNamePtr.get() == 0)
    throw errors::Logic();
  if (! mapPtr.unique())
    mapPtr.reset(new StrPtrNumMap(*mapPtr));

  StrPtrNumMap::iterator oldStrPtrAndNum = mapPtr->find(oldNamePtr);
  if (oldStrPtrAndNum == mapPtr->end())
    throw errors::Logic();
  numbers::Number& amt = oldStrPtrAndNum->second;

  strPool::StrPtr newNamePtr = strPool.addToPool(newName);

  mapPtr->insert(std::make_pair(newNamePtr, amt));
  mapPtr->erase(oldStrPtrAndNum);
}

void Food::calcIngrNuts() const
{
  ingrNuts_->clear();
  for (std::vector<food::Food>::const_iterator ingredient =
       ingredients_->begin();
       ingredient != ingredients_->end();
       ++ingredient)
  {
    std::vector<std::string> nutNames = ingredient->getNutNames();
    for (std::vector<std::string>::const_iterator nutName =
         nutNames.begin();
         nutName != nutNames.end();
         ++nutName)
    {
      numbers::Number& totalNutValue = (*ingrNuts_)[*nutName];
      totalNutValue = 
        numbers::Number(totalNutValue.getValue() +
                        ingredient->getNut(*nutName).getValue());
    }
  }
}

Food::Food()
{}

Food::Food(const pointers::Shared<strPool::StrPool>& strPool,
           const pointers::Shared<absUnitPool::AbsUnitPool>& absUnitPool)
  : strPool_        (strPool),
    absUnitPool_    (absUnitPool),
    nuts_           (new std::map<strPool::StrPtr, numbers::Number>),
    arbUnits_       (new std::map<strPool::StrPtr, numbers::Number>),
    ingredients_    (new std::vector<Food>),
    name_           (strPool->addToPool("")),
    group_          (strPool->getStrPtr("")),
    date_           (strPool->getStrPtr("")),
    meal_           (strPool->getStrPtr("")),
    order_          (strPool->getStrPtr("")),
    comment_        (strPool->getStrPtr("")),
    instructions_   (strPool->getStrPtr("")),
    refDesc_        (strPool->getStrPtr("")),
    unitIsAbsolute_ (true),
    absUnit_        (absUnitPool->getDefaultUnit()),
    qty_            (new numbers::StrNumber("0")),
    pctRefuse_      (new numbers::Percent(0)),
    yield_          (new numbers::Number(0)),
    recalcIngrNuts_ (new bool(false)),
    ingrNuts_       (new StrNumMap())
{}

void Food::addNut(const std::string& name, const numbers::Number& amount)
{
  if (! nuts_.unique())
    nuts_.reset(new StrPtrNumMap(*nuts_));

  strPool::StrPtr namePtr = strPool_->addToPool(name);
  numbers::Number foodGrams(getGrams(*this));
  if (foodGrams.getValue() < 1)
    throw errors::Logic();

  numbers::Number ratio(amount.getValue() / foodGrams.getValue());
  std::pair<StrPtrNumMap::iterator, bool> insertResult =
    nuts_->insert(std::make_pair(namePtr, ratio));
  if (! insertResult.second)
    insertResult.first->second = ratio;
}

std::vector<std::string> Food::getNutNames() const
{
  if (*recalcIngrNuts_)
  {
    calcIngrNuts();
    *recalcIngrNuts_ = false;
  }
  std::vector<std::string> result;
  for (StrPtrNumMap::const_iterator namePtrAndNumber = nuts_->begin();
       namePtrAndNumber != nuts_->end();
       ++namePtrAndNumber)
  {
    result.push_back(*(namePtrAndNumber->first));
  }

  for (StrNumMap::const_iterator nameAndNumber = ingrNuts_->begin();
       nameAndNumber != ingrNuts_->end(); ++nameAndNumber)
  {
    result.push_back(nameAndNumber->first);
  }
  return result;
}

void Food::deleteNut(const std::string& name)
{
  strPool::StrPtr namePtr = strPool_->getStrPtr(name);
  if (namePtr.get() == 0)
    return;
  if (! nuts_.unique())
    nuts_.reset(new StrPtrNumMap(*nuts_));
  nuts_->erase(namePtr);
}

numbers::Number Food::getNut(const std::string& name) const
{
  strPool::StrPtr namePtr = strPool_->getStrPtr(name);
  if (namePtr.get() == 0)
    return getNutIngr(name);
  StrPtrNumMap::const_iterator strPtrAndNum = nuts_->find(namePtr);
  if (strPtrAndNum == nuts_->end())
    return getNutIngr(name);
  
  return numbers::Number(strPtrAndNum->second.getValue() *
                           getGrams(*this).getValue());
}

numbers::Number Food::getNutIngr(const std::string& name) const
{
  if (*recalcIngrNuts_)
  {
    calcIngrNuts();
    *recalcIngrNuts_ = false;
  }
  StrNumMap::const_iterator nameAndAmt = ingrNuts_->find(name);
  if (nameAndAmt == ingrNuts_->end())
    throw errors::Logic();
  double totalNut = nameAndAmt->second.getValue();
  return numbers::Number(totalNut / getRecipeGrams(*this).getValue() *
                         getGrams(*this).getValue());
}

void Food::renameNut(const std::string& oldName,
                     const std::string& newName)
{
  renameNumberMapItem(oldName, newName,
                      nuts_, *strPool_);
}

void Food::addArbUnit(const std::string& name,
                      const numbers::Number& amount)
{
  if (! arbUnits_.unique())
    arbUnits_.reset(new StrPtrNumMap(*arbUnits_));
  strPool::StrPtr newNamePtr = strPool_->addToPool(name);

  std::pair<StrPtrNumMap::iterator, bool> insertResult = 
    arbUnits_->insert(std::make_pair(newNamePtr, amount));
  if (! insertResult.second)
    insertResult.first->second = amount;
}

std::vector<std::string> Food::getArbUnitNames() const
{
  std::vector<std::string> result;
  for (StrPtrNumMap::const_iterator strPtrAndNum =
       arbUnits_->begin();
       strPtrAndNum != arbUnits_->end();
       ++strPtrAndNum)
  {
    result.push_back(*strPtrAndNum->first);
  }
  return result;
}

std::vector<std::string> Food::getAbsUnitNames() const
{
  return absUnitPool_->getAbsUnitNames();
}

numbers::Number Food::getArbUnitAmount(const std::string& name) const
{
  strPool::StrPtr strPtr = strPool_->getStrPtr(name);
  if (strPtr.get() == 0)
    throw errors::Logic();
  StrPtrNumMap::const_iterator strPtrAndAmt =
    arbUnits_->find(strPtr);
  if (strPtrAndAmt == arbUnits_->end())
    throw errors::Logic();
  return strPtrAndAmt->second;
}

numbers::Number Food::getAbsUnitAmount(const std::string& name) const
{
  numbers::Number result;
  bool getResult = absUnitPool_->getAbsUnitAmount(name, result); 
  if (! getResult)
    throw errors::Logic();
  return result;
}

void Food::deleteArbUnit(const std::string& name)
{
  strPool::StrPtr strPtr = strPool_->getStrPtr(name);
  if (strPtr.get() == 0)
    throw errors::Logic();
  if (! unitIsAbsolute_ &&
      (strPtr.get() == arbUnit_->first.get()))
    throw errors::Logic();
  if (! arbUnits_.unique())
    arbUnits_.reset(new StrPtrNumMap(*arbUnits_));
  arbUnits_->erase(strPtr);
}

void Food::setIngredients(const std::vector<Food>& ingredients)
{
  if (ingredients_.unique())
  {
    *ingredients_ = ingredients;
    *recalcIngrNuts_ = true;
  }
  else
  {
    ingredients_.reset(new std::vector<Food>(ingredients));
    recalcIngrNuts_.reset(new bool(true));
  }
}

std::vector<Food> Food::getIngredients() const
{
  return *ingredients_;
}

void Food::changeTag(const std::string& newValue,
                     strPool::StrPtr& member)
{
  member = strPool_->addToPool(newValue);
}

void Food::setName(const std::string& value)
{
  changeTag(value, name_);
}

void Food::setGroup(const std::string& value)
{
  changeTag(value, group_);
}

void Food::setDate(const std::string& value)
{
  changeTag(value, date_);
}

void Food::setMeal(const std::string& value)
{
  changeTag(value, meal_);
}

void Food::setOrder(const std::string& value)
{
  changeTag(value, order_);
}

void Food::setComment(const std::string& value)
{
  changeTag(value, comment_);
}

void Food::setInstructions(const std::string& value)
{
  changeTag(value, instructions_);
}

void Food::setRefDesc(const std::string& value)
{
  changeTag(value, refDesc_);
}

std::string Food::getName() const
{
  return *name_;
}

std::string Food::getGroup() const
{
  return *group_;
}

std::string Food::getDate() const
{
  return *date_;
}

std::string Food::getMeal() const
{
  return *meal_;
}

std::string Food::getOrder() const
{
  return *order_;
}

std::string Food::getComment() const
{
  return *comment_;
}

std::string Food::getInstructions() const
{
  return *instructions_;
}

std::string Food::getRefDesc() const
{
  return *refDesc_;
}

void Food::renameArbUnit(const std::string& oldName,
                         const std::string& newName)
{
  strPool::StrPtr oldArbUnitName;
  if (! unitIsAbsolute_)
  {
    oldArbUnitName = arbUnit_->first;
    if (oldArbUnitName.get() == strPool_->getStrPtr(newName).get())
      throw errors::Logic();
  }

  renameNumberMapItem(oldName, newName,
                      arbUnits_, *strPool_);

  if (! unitIsAbsolute_)
    arbUnit_ = arbUnits_->find(oldArbUnitName);
}

void Food::setCurrentUnitAbsolute(const std::string& name)
{
  absUnitPool::AbsUnitPtr newAbsUnitPtr = absUnitPool_->getAbsUnit(name);
  if (newAbsUnitPtr.get() == 0)
    throw errors::Logic();
  unitIsAbsolute_ = true;
  absUnit_ = newAbsUnitPtr;
}

void Food::setCurrentUnitArbitrary(const std::string& name)
{
  strPool::StrPtr newUnitNamePtr = strPool_->getStrPtr(name);
  if (newUnitNamePtr.get() == 0)
    throw errors::Logic();
  StrPtrNumMap::const_iterator arbIt =
    arbUnits_->find(newUnitNamePtr);
  if (arbIt == arbUnits_->end())
    throw errors::Logic();
  unitIsAbsolute_ = false;
  arbUnit_ = arbIt;
}

std::string Food::getCurrentUnitName() const
{
  if (unitIsAbsolute_)
    return absUnit_->name;
  return *arbUnit_->first;
}

numbers::Number Food::getCurrentUnitAmount() const
{
  if (unitIsAbsolute_)
    return absUnit_->value;
  return arbUnit_->second;
}

numbers::StrNumber Food::getQty() const
{
  return *qty_;
}

void Food::setQty(const numbers::StrNumber& qty)
{
  qty_.reset(new numbers::StrNumber(qty));
}

numbers::Percent Food::getPctRefuse() const
{
  return *pctRefuse_;
}

void Food::setPctRefuse(const numbers::Percent& percent)
{
  pctRefuse_.reset(new numbers::Percent(percent));
}

numbers::Number Food::getYield() const
{
  return *yield_;
}

void Food::setYield(const numbers::Number& yield)
{
  yield_.reset(new numbers::Number(yield));
}

bool Food::getQtyByNut(const std::string& nutName,
                       const numbers::Number& nutAmount,
                       numbers::StrNumber& result) const
{
  strPool::StrPtr strPtr = strPool_->getStrPtr(nutName);
  if (strPtr.get() == 0)
    return getQtyByNutIngr(nutName, nutAmount, result);
  StrPtrNumMap::const_iterator nutNameAndAmt = nuts_->find(strPtr);
  if (nutNameAndAmt == nuts_->end())
    return getQtyByNutIngr(nutName, nutAmount, result);

  if (nutNameAndAmt->second.isZero())
    return false;
  double nutPerg = nutNameAndAmt->second.getValue();
  double unitValue;
  if (unitIsAbsolute_)
    unitValue = absUnit_->value.getValue();
  else
    unitValue = arbUnit_->second.getValue();
  double resultAmt = nutAmount.getValue() / nutPerg / unitValue;
  result = numbers::StrNumber(resultAmt);
  return true;
}

bool Food::getQtyByNutIngr(const std::string& nutName,
                           const numbers::Number& nutAmount,
                           numbers::StrNumber& result) const
{
  if (*recalcIngrNuts_)
  {
    calcIngrNuts();
    *recalcIngrNuts_ = false;
  }
  StrNumMap::const_iterator nutNameAndAmt = ingrNuts_->find(nutName);
  if (nutNameAndAmt == ingrNuts_->end())
    throw errors::Logic();
  if (nutNameAndAmt->second.isZero())
    return false;
  double totalNutAmt = nutNameAndAmt->second.getValue();
  double recipeGrams = getRecipeGrams(*this).getValue();
  double unitAmt = getCurrentUnitAmount().getValue();
  result = numbers::StrNumber((nutAmount.getValue() * recipeGrams) /
                              (totalNutAmt * unitAmt));
  return true;
}

numbers::Number getGrams(const Food& food)
{
  return numbers::Number(food.getCurrentUnitAmount().getValue() *
                         food.getQty().getValue());
}

numbers::Number getRecipeGrams(const Food& food)
{
  std::vector<food::Food> ingredients = food.getIngredients();
  if (ingredients.empty())
    return numbers::Number(0);
  numbers::Number yield = food.getYield();
  if (! yield.isZero())
    return yield;
  double grams = 0;
  for (std::vector<food::Food>::const_iterator ingredient =
       ingredients.begin(); ingredient != ingredients.end();
       ++ingredient)
  {
    grams += getGrams(*ingredient).getValue();
  }
  return numbers::Number(grams);
}

numbers::StrNumber getQtyMinusRefuse(const Food& food)
{
  double pctRefuse = food.getPctRefuse().getValue() / 100.0;
  double oldQty = food.getQty().getValue();
  return numbers::StrNumber(oldQty - pctRefuse * oldQty);
}

} // namespace food
