#ifndef FOOD_H
#define FOOD_H

#include "absUnitPool.h"
#include "numbers.h"
#include "pointers.h"
#include "strPool.h"
#include "serializer_fwd.h"

#include <map>
#include <string>
#include <vector>

namespace food
{

const std::string NAME              = "name";
const std::string GROUP             = "group";
const std::string DATE              = "date";
const std::string MEAL              = "meal";
const std::string COMMENT           = "comment";
const std::string ORDER             = "order";
const std::string REFDESC           = "refuse-description";
const std::string INSTRUCTIONS      = "instructions";
const std::string QUANTITY          = "quantity";
const std::string PCTREFUSE         = "percent-refuse";
const std::string YIELD             = "yield";
const std::string UNIT              = "unit";

class Food
{
public:

  /** ctor.
   *
   * \param strPool Shared pointer to the pool from which
   * this food will pull all its strings.
   *
   * \param absUnitPool Shared pointer to the pool of
   * absolute units.
   */
  Food(const pointers::Shared<strPool::StrPool>& strPool,
       const pointers::Shared<absUnitPool::AbsUnitPool>& absUnitPool);

  /** Add a nutrient to the food.
   *
   * \param name Name of the nutrient to add. If this
   * nutrient already exists in the food, the existing one
   * is overwritten.
   *
   * \param amount Amount of the nutrient.
   *
   * It is an error to attempt to add a nutrient when the
   * food's grams is less than one. An attempt to do so
   * throws errors::logic.
   */

  void addNut(const std::string& name, const numbers::Number& amount);

  /** Get a vector containing all the names of nutrients in
   * this food. This vector is unsorted.
   */
  std::vector<std::string> getNutNames() const;

  /** Delete nutrient with the given name. Has no effect if
   * this food has no nutrient with the given name.
   */
  void deleteNut(const std::string& name);

  /** Get value for nutrient with given name.
   *
   * \param name Nutrient to get value for.
   *
   * \param amount Result is placed in this variable;
   * variable is unchanged if food has no nutrient with
   * given name.
   *
   * It is an error to get a nutrient if the food has no
   * nutrient by that name. An attempt to do so throws
   * errors::Logic.
   */
  numbers::Number getNut(const std::string& name) const;

  /** Change nutrient name.
   *
   * \param oldName Name of nutrient to rename.
   *
   * \param newName New nutrient name.
   *
   * It is an error to attempt to change a nutrient name if
   * the food has no nutrients with that name; attempting to
   * do so throws errors::Logic.
   */
  void renameNut(const std::string& oldName, const std::string& newName);

  /** Add arbitrary unit.
   *
   * \param name Name of arbitrary unit to add.
   *
   * \param amount Amount, in grams, of this arbitrary unit.
   */
  void addArbUnit(const std::string& name, const numbers::Number& amount);

  /** Return a vector of names of this food's arbitrary
   * units.
   */
  std::vector<std::string> getArbUnitNames() const;

  /** Return a vector of names of this food's absolute
   * units.
   */
  std::vector<std::string> getAbsUnitNames() const;

  /** Get the amount of a particular arbitrary unit.
   * 
   * \param name Name of unit.
   *
   * \param amount Result is returned here.
   *
   * It is an error to attempt to get an arbitrary unit
   * amount if the food has no unit with that name;
   * attempting to do so throws errors::Logic.
   *
   * \return True if unit was found; False otherwise.
   */
  numbers::Number getArbUnitAmount(const std::string& name) const;

  /** Get the amount of a particular absolute unit.
   *
   * \param name Name of unit.
   *
   * \param amount Result is returned here.
   *
   * It is an error to attempt to get an absolute unit
   * amount if the food has no unit with that name;
   * attempting to do so throws errors::Logic.
   */
  numbers::Number getAbsUnitAmount(const std::string& name) const;

  /** Delete arbitrary unit with a given name.
   *
   * \param name Name of arbitrary unit to delete.
   *
   * It is an error to attempt to delete an arbitrary unit
   * if there is no arbitrary unit with that name; doing so
   * throws errors::Logic. It is also an error to attempt to
   * delete an arbitrary unit that is the food's current
   * unit; such an attempt throws errors::Logic.
   */

  void deleteArbUnit(const std::string& name);

  /** Change name of particular arbitrary unit.
   *
   * \param oldName Name of arbitrary unit to rename.
   *
   * \param newName New name for arbitrary unit.
   *
   * It is an error to attempt to rename an arbitrary unit
   * if there is no unit with that name; attempting to do so
   * throws errors::Logic. It is also an error to attempt to
   * rename an arbitrary unit that is the food's current
   * unit; attempting to do so throws errors::Logic.
   */

  void renameArbUnit(const std::string& oldName,
                     const std::string& newName);

  void setIngredients(const std::vector<Food>& ingredients);
  std::vector<Food> getIngredients() const;

  void setName(const std::string& value);
  void setGroup(const std::string& value);
  void setDate(const std::string& value);
  void setMeal(const std::string& value);
  void setOrder(const std::string& value);
  void setComment(const std::string& value);
  void setInstructions(const std::string& value);
  void setRefDesc(const std::string& value);

  std::string getName() const;
  std::string getGroup() const;
  std::string getDate() const;
  std::string getMeal() const;
  std::string getOrder() const;
  std::string getComment() const;
  std::string getInstructions() const;
  std::string getRefDesc() const;

  /** Set the current unit to a particular absolute unit.
   *
   * \param name Absolute unit name.
   *
   * It is an error to attempt to set the current unit to a
   * unit that the food does not have; attempting to do this
   * throws errors:::Logic.
   */
  void setCurrentUnitAbsolute(const std::string& name);

  /** Set the current unit to a particular arbitrary unit.
   *
   * \param name Arbitrary unit name to use.
   *
   * It is an error to attempt to set the current unit to a
   * unit that the food does not have; attempting to do this
   * throws errors:::Logic.
   */
  void setCurrentUnitArbitrary(const std::string& name);

  /** Get the name of the food's current unit.
   */
  std::string getCurrentUnitName() const;

  /** Get the amount, in grams, of the food's current unit.
   */
  numbers::Number getCurrentUnitAmount() const;

  /** Get the food's current quantity.
   */
  numbers::StrNumber getQty() const;

  /** Set the food's current quantity.
   *
   * Offers strong guarantee.
   */
  void setQty(const numbers::StrNumber& qty);

  /** Get the food's current percent refuse.
   *
   * Never throws.
   */
  numbers::Percent getPctRefuse() const;

  /** Set the food's percent refuse.
   *
   * Offers strong guarantee.
   */
  void setPctRefuse(const numbers::Percent& percent);

  /** Get the food's current yield.
   * 
   * Never throws.
   */
  numbers::Number getYield() const;

  /** Set the food's yield.
   *
   * Offers strong guarantee.
   */
  void setYield(const numbers::Number& yield);

  /** For a given quantity of nutrient, find the amount of
   * the food that contains that quantity. 
   *
   * It is an error to pass a nutrient that the food does
   * not have in nutName; doing so throws errors::Logic.
   *
   * The result is returned in the result parameter.
   *
   * \return True if result is successfully returned in the
   * result parameter. False if result is not successfully
   * returned because the value of the nutrient specified in
   * nutName is zero.
   */
  bool getQtyByNut(const std::string& nutName,
                   const numbers::Number& nutAmount,
                   numbers::StrNumber& result) const;

  friend class serializer::Loader;
  friend class serializer::Saver;

private:
  typedef std::map<strPool::StrPtr, numbers::Number> StrPtrNumMap;
  typedef std::map<std::string, numbers::Number> StrNumMap;

  pointers::Shared<strPool::StrPool> strPool_;
  pointers::Shared<absUnitPool::AbsUnitPool> absUnitPool_;

  pointers::Shared<StrPtrNumMap> nuts_;
  pointers::Shared<StrPtrNumMap> arbUnits_;
  pointers::Shared<std::vector<Food> > ingredients_;

  strPool::StrPtr name_;
  strPool::StrPtr group_;
  strPool::StrPtr date_;
  strPool::StrPtr meal_;
  strPool::StrPtr order_;
  strPool::StrPtr comment_;
  strPool::StrPtr instructions_;
  strPool::StrPtr refDesc_;

  bool unitIsAbsolute_;
  absUnitPool::AbsUnitPtr absUnit_;
  StrPtrNumMap::const_iterator arbUnit_;
  
  pointers::Shared<numbers::StrNumber> qty_;
  pointers::Shared<numbers::Percent> pctRefuse_;
  pointers::Shared<numbers::Number> yield_;
  
  mutable pointers::Shared<bool> recalcIngrNuts_;
  mutable pointers::Shared<StrNumMap> ingrNuts_;

  Food();

  void changeTag(const std::string& newValue, strPool::StrPtr& member);


  void renameNumberMapItem(const std::string& oldName,
                           const std::string& newName,
                           pointers::Shared<StrPtrNumMap>& mapPtr,
                           strPool::StrPool& strPool);
  void calcIngrNuts() const;
  bool getQtyByNutIngr(const std::string& nutName,
                       const numbers::Number& nutAmount,
                       numbers::StrNumber& result) const;
  numbers::Number getNutIngr(const std::string& name) const;
};

numbers::Number getGrams(const Food& food);
numbers::StrNumber getQtyMinusRefuse(const Food& food);

/** Return the total grams of a single recipe.
 *
 * If the food has no ingredients, this is always zero. If
 * the food has ingredients, and Yield is non-zero (as
 * defined by Number.nonZero()) then this returns the
 * yield. If the food has ingredients, and Yield is not
 * non-zero, then this returns the grams of the
 * ingredients.
 */
numbers::Number getRecipeGrams(const Food& food);

} // namespace food

#endif
