#include "inputs.h"

#include "adder.h"
#include "changer.h"
#include "errors.h"
#include "parser.h"
#include "reporter.h"
#include "stl.h"

#include <string>
#include <vector>

namespace inputs {

namespace {

const std::string OPEN              = "open";
const std::string READ              = "read";
const std::string CLOSE             = "close";
const std::string SAVE              = "save";
const std::string SAVE_AS           = "save-as";
const std::string QUIT              = "quit";
const std::string STATUS            = "status";
const std::string SEARCH            = "search";
const std::string CHANGE            = "change";
const std::string BY_NUT            = "by-nut";
const std::string REFUSE            = "refuse";
const std::string C_NUT             = "change-nut";
const std::string R_NUT             = "rename-nut";
const std::string DEL_NUTS          = "delete-nuts";
const std::string C_AVAIL_UNIT      = "change-avail-unit";
const std::string R_AVAIL_UNIT      = "rename-avail-unit";
const std::string DEL_AVAIL_UNIT    = "delete-avail-units";
const std::string ADD_INGREDIENTS   = "add-ingredients";
const std::string DEL_INGREDIENTS   = "delete-ingredients";
const std::string IGNORE_CASE       = "ignore-case";
const std::string EXACT_MATCH       = "exact-match";
const std::string EDIT              = "edit";
const std::string DELETE            = "delete";
const std::string CREATE            = "create";
const std::string LIMIT             = "limit";
const std::string PRINT             = "print";
const std::string KEY               = "key";
const std::string GOAL              = "goal";
const std::string LIST              = "list";
const std::string ADD               = "add";
const std::string AUTO_ORDER        = "auto-order";
const std::string HELP              = "help";
// don't call this next one version (in all caps)
// as this will conflict with a macro defined by the GNU build
// system
const std::string VER               = "version";
const std::string COPYRIGHT         = "copyright";

parser::Parser getParser()
{
    parser::Parser result;

    result.addOption(OPEN,                      1);
    result.addOption(READ,                      1);
    result.addOption(CLOSE,                     0);
    result.addOption(SAVE,                      0);
    result.addOption(SAVE_AS,                   1);
    result.addOption(QUIT,                      0);
    result.addOption(STATUS,                    0);

    result.addOption(SEARCH,            's',    2);
    result.addOption(CHANGE,            'c',    2);
    result.addOption(BY_NUT,                    2);
    result.addOption(REFUSE,            'r',    0);
    result.addOption(C_NUT,                     2);
    result.addOption(R_NUT,                     2);
    result.addOption(DEL_NUTS,                  1);
    result.addOption(C_AVAIL_UNIT,              2);
    result.addOption(R_AVAIL_UNIT,              2);
    result.addOption(DEL_AVAIL_UNIT,            1);
    result.addOption(ADD_INGREDIENTS,           1);
    result.addOption(DEL_INGREDIENTS,           1);
    result.addOption(IGNORE_CASE,       'i',    0);
    result.addOption(EXACT_MATCH,       'x',    0);
    result.addOption(EDIT,                      0);
    result.addOption(DELETE,                    0);
    result.addOption(CREATE,                    0);
    result.addOption(LIMIT,                     1);
    result.addOption(PRINT,             'p',    1);
    result.addOption(KEY,               'k',    1);
    result.addOption(GOAL,              'g',    2);
    result.addOption(LIST,              'l',    2);
    result.addOption(ADD,               'a',    1);
    result.addOption(AUTO_ORDER,        'o',    0);
    result.addOption(HELP,              'h',    0);
    result.addOption(VER,               'v',    0);
    result.addOption(COPYRIGHT,                 0);
    return result;
}

bool initBool(const parser::Result& parsed, const std::string& name)
{
  return (! stl::checkedGet(parsed.optArgs, name).empty());
}

std::string initString(const parser::Result& parsed,
                       const std::string& name)
{
  const std::vector<std::vector<std::string> >& vecs =
    stl::checkedGet(parsed.optArgs, name);
  std::string result;
  if (! vecs.empty() && ! vecs.back().empty())
    result = vecs.back().back();
  return result;
}

searcher::Searcher getSearcher(const parser::Result& parsed)
{
  bool ignoreCase = initBool(parsed, IGNORE_CASE);
  bool exactMatch = initBool(parsed, EXACT_MATCH);
  bool doLimit = initBool(parsed, LIMIT);
  std::string limit = initString(parsed, LIMIT);
  return searcher::Searcher
         (parser::vecsToPairs(stl::checkedGet(parsed.optArgs, SEARCH)),
          std::make_pair(doLimit, limit),
          ignoreCase,
          exactMatch,
          parsed.posArgs);
}

changer::Changer
getChanger (const parser::OptArgsMapType& optArgs,
            const db::Db& db)
{
    using namespace parser;
    return changer::Changer(
               vecsToPairs(stl::checkedGet(optArgs, CHANGE)),
               vecsToPairs(stl::checkedGet(optArgs, BY_NUT)),
               isSet(stl::checkedGet(optArgs, REFUSE)),
               vecsToPairs(stl::checkedGet(optArgs, C_NUT)),
               vecsToPairs(stl::checkedGet(optArgs, R_NUT)),
               vecsToVec(stl::checkedGet(optArgs, DEL_NUTS)),
               vecsToPairs(stl::checkedGet(optArgs, C_AVAIL_UNIT)),
               vecsToPairs(stl::checkedGet(optArgs, R_AVAIL_UNIT)),
               vecsToVec(stl::checkedGet(optArgs, DEL_AVAIL_UNIT)),
               db,
               vecsToVec(stl::checkedGet(optArgs, ADD_INGREDIENTS)),
               vecsToVec(stl::checkedGet(optArgs, DEL_INGREDIENTS)),
               isSet(stl::checkedGet(optArgs, IGNORE_CASE)),
               isSet(stl::checkedGet(optArgs, EXACT_MATCH)));
}

adder::Adder getAdder(const parser::OptArgsMapType& optArgs)
{
    return adder::Adder
           (parser::vecsToVec(stl::checkedGet(optArgs, ADD)),
            parser::isSet(stl::checkedGet(optArgs, AUTO_ORDER)));
}

reporter::Reporter getReporter(const parser::OptArgsMapType& optArgs)
{
    return reporter::Reporter
           (parser::vecsToVec(stl::checkedGet(optArgs, PRINT)),
            parser::vecsToPairs(stl::checkedGet(optArgs, GOAL)),
            parser::vecsToVec(stl::checkedGet(optArgs, KEY)),
            parser::vecsToPairs(stl::checkedGet(optArgs, LIST)));
}

} // unnamed namespace

parser::Result getParsed(const std::vector<std::string>& args)
{
    return getParser().parse(args);
}

Inputs::Inputs(const parser::Result& parsed,
               const db::Db& db)
  : doOpen(initBool(parsed, OPEN)),
    openFilename(initString(parsed, OPEN)),
    doRead(initBool(parsed, READ)),
    readFilename(initString(parsed, READ)),
    doClose(initBool(parsed, CLOSE)),
    doSave(initBool(parsed, SAVE)),
    doSaveAs(initBool(parsed, SAVE_AS)),
    saveAsFilename(initString(parsed, SAVE_AS)),
    doQuit(initBool(parsed, QUIT)),
    doStatus(initBool(parsed, STATUS)),
    doCreate(initBool(parsed, CREATE)),
    doHelp(initBool(parsed, HELP)),
    doVersion(initBool(parsed, VER)),
    doCopyright(initBool(parsed, COPYRIGHT)),
    searcher(getSearcher(parsed)),
    changer(getChanger(parsed.optArgs, db)),
    doEdit(initBool(parsed, EDIT)),
    doDelete(initBool(parsed, DELETE)),
    adder(getAdder(parsed.optArgs)),
    reporter(getReporter(parsed.optArgs))
{
  if ((doOpen && openFilename.empty()) ||
      (doRead && readFilename.empty()) ||
      (doSaveAs && saveAsFilename.empty()))
    throw EmptyFilenameError();
}

std::string EmptyFilenameError::getBriefMessage() const
{
  return "cannot perform --save, --save-as, --read, or --open "
         "with an empty filename";
}


} // namespace inputs
