#include "matchers.h"

#include "errors.h"
#include "strings.h"

#include <string>
#include <vector>

#include <regex.h>

namespace matchers {

struct RE::Impl
{
  std::string pattern;
  bool ignoreCase;
  regex_t regex;
};

RE::RE(const std::string& pattern, bool ignoreCase)
  : pimpl_(new Impl)
{
  pimpl_->pattern = pattern;
  pimpl_->ignoreCase = ignoreCase;

  int flags = REG_NOSUB | REG_EXTENDED;
  if (ignoreCase)
    flags |= REG_ICASE;

  int errorCode = regcomp(&(pimpl_->regex), pattern.c_str(), flags);
  if (errorCode)
  {
    size_t length = regerror(errorCode, &pimpl_->regex, 0, 0);
    std::vector<char> buffer(length);
    regerror(errorCode, &pimpl_->regex, &(buffer[0]), length);
    throw BadRePattern(&(buffer[0]), pattern);
  }
}

RE::~RE()
{
  regfree(&(pimpl_->regex));
}

bool RE::operator() (const std::string& x) const
{
  int errorcode = regexec(&(pimpl_->regex), x.c_str(), 0, 0, 0);
  switch (errorcode)
  {
    case REG_NOMATCH:
        return false;
    case REG_ESPACE:
        throw ExecOutOfMemory();
    default:
        return true;
  }
}

MatcherPtr getMatcher(const std::string& pattern,
                      bool ignoreCase, bool exactMatch)
{
  if (exactMatch)
    return MatcherPtr(new Exact(pattern, ignoreCase));
  else
    return MatcherPtr(new RE(pattern, ignoreCase));
}

BadRePattern::BadRePattern(const std::string& eMsg,
                           const std::string& pattern)
    : eMsg_(eMsg), pattern_(pattern) {}

std::string BadRePattern::getBriefMessage() const
{
  std::string result = "bad regular expression: ";
  result.append(pattern_);
  result.append(": ");
  result.append(eMsg_);
  return result;
}

Exact::Exact(const std::string& pattern, bool ignoreCase)
{
  if (ignoreCase)
    pattern_ = strings::toLower(pattern);
  else
    pattern_ = pattern;
  ignoreCase_ = ignoreCase;
}

bool Exact::operator() (const std::string& x) const
{
  if (ignoreCase_)
  {
    return strings::toLower(x) == pattern_;
  }
  else
    return x == pattern_;
}

bool testString(const MatcherVec& regexps,
                const std::string& subject)
{
  for (MatcherVec::const_iterator regexp = regexps.begin();
       regexp != regexps.end(); ++regexp)
  {
    if ((*(*regexp))(subject))
      return true;
  }
  return false;
}

bool doesNotMatch(const MatcherVec& regexps,
                  const std::string& subject)
{
  for (MatcherVec::const_iterator regexp = regexps.begin();
       regexp != regexps.end(); ++regexp)
  {
    if (! (*(*regexp))(subject))
      return true;
  }
  return false;
}

StringVec selectStrings(const MatcherVec& test,
                        const StringVec& strings)
{
  StringVec result;
  for (StringVec::const_iterator string = strings.begin();
       string != strings.end(); ++string)
  {
    if (testString(test, *string))
      result.push_back(*string);
  }
  return result;
}

StringVec selectStrings(const Matcher& matcher,
                        const StringVec& strings)
{
  StringVec result;
  for (StringVec::const_iterator string = strings.begin();
       string != strings.end(); ++string)
  {
    if (matcher(*string))
      result.push_back(*string);
  }
  return result;
}

StringVec deleteStrings(const MatcherVec& test,
                        const StringVec& strings)
{
  StringVec result;
  for (StringVec::const_iterator string = strings.begin();
       string != strings.end(); ++string)
  {
    if (! testString(test, *string))
      result.push_back(*string);
  }
  return result;
}

StringVec deleteStrings(const RE& test,
                        const StringVec& strings)
{
  StringVec result;
  for (StringVec::const_iterator string = strings.begin();
       string != strings.end(); ++string)
  {
    if (! test(*string))
      result.push_back(*string);
  }
  return result;
}

std::string ExecOutOfMemory::getBriefMessage() const
{
    return "ran out of memory while matching regular expression";
}

} // namespace matchers
