#ifndef MATCHERS_H
#define MATCHERS_H

#include <memory>
#include <string>
#include <vector>

#include "errors.h"
#include "pointers.h"

/** Objects that perform pattern matching.
 *
 * This can include pattern matching by regular expression or by a
 * simple lexical match.
 */
namespace matchers {

/** Abstract base class for a matcher.
 */
class Matcher
{
public:
  /** See if a string matches the pattern.
   * \param x String to test.
   * \return True if there is a match, False if not.
   */
  virtual bool operator() (const std::string& x) const = 0;

  /** Virtual dtor
   */
  virtual ~Matcher() {}

};

typedef pointers::Shared<Matcher> MatcherPtr;
typedef std::vector<MatcherPtr> MatcherVec;
typedef std::vector<std::string> StringVec;

/** Regular expression matcher. Uses POSIX.
 */
class RE : public Matcher
{
public:
  /** ctor.
   *
   * \param pattern RE pattern to use.
   * \param ignoreCase True if pattern matching should ignore
   * case; False otherwise.
   *
   * \throw errors::BadRePattern Pattern is not a valid regular
   * expression.
   */
  RE(const std::string& pattern, bool ignoreCase);

  /** See if string matches stored pattern.
   * \param subject String to test.
   * \return True if subject matches stored pattern; False
   * otherwise.
   *
   * \throw errors::ExecOutOfMemory regular expression matching
   * routine ran out of memory while attempting to match.
   */
  bool operator() (const std::string& subject) const;

  ~RE();

private:
  RE(const RE&);
  RE& operator= (const RE&); 
  struct Impl;
  std::auto_ptr<Impl> pimpl_;
};

/** Matches based on string equality. Only guaranteed to work
 * with ASCII.
 */
class Exact : public Matcher
{
public:

  /** ctor.
   *
   * \param pattern Pattern to use when matching.
   * \param ignoreCase True if matching should be case
   * insensitive; False otherwise.
   */
  Exact(const std::string& pattern, bool ignoreCase);

  /** See if subject matches stored pattern.
   *
   * \param subject String to test.
   * \return True if subject matches pattern; False otherwise.
   */
  bool operator() (const std::string& subject) const;

private:
  std::string pattern_;
  bool ignoreCase_;
};

/** Return an appropriate Matcher.
 *
 * \param pattern Pattern to use.
 *
 * \param ignoreCase whether to ignore case.
 *
 * \param exactMatch whether to use exact matching.
 *
 * \return Shared pointer to an appropriate Matcher.
 *
 * Offers strong guarantee.
 */
MatcherPtr getMatcher(const std::string& pattern,
                      bool ignoreCase, bool exactMatch);

/** Get vector of strings that match a vector of Matchers.
 *
 * \param test Vector of Matchers against which to test strings.
 *
 * \param strings Vector of strings.
 *
 * \return Vector of strings containing every string in strings that
 * matched at least one Matcher in test. Thus selectStrings
 * "selects" every string in strings that matches one of the
 * Matchers in test.
 */
StringVec selectStrings(const MatcherVec& test,
                        const StringVec& strings);

/** Get vector of strings that match a single Matcher.
 *
 * \param test Matcher against which to test strings.
 *
 * \param strings Vector of strings.
 *
 * \return Vector of strings containing every string in strings that
 * matched test. Thus selectStrings "selects" every string in
 * strings that matches test.
 */
StringVec selectStrings(const Matcher& test,
                        const StringVec& strings);

/** Get vector of strings that do not match a vector of Matchers.
 *
 * \param test Vector of matchers to use for testing.
 *
 * \param strings Vector of strings to test.
 *
 * \return Vector of strings containing every string in strings that
 * did NOT match ANY of the Matchers in test. Thus, deleteStrings
 * "deletes" every string that matches one of the Matchers in test.
 */
StringVec deleteStrings(const MatcherVec& test,
                        const StringVec& strings);

/** Get vector of strings that do not match a Matcher.
 *
 * \param test Matcher to use for testing.
 *
 * \param strings Vector of strings to test.
 *
 * \return Vector of strings containing every string in strings that
 * did NOT match test. Thus, deleteStrings
 * "deletes" every string that matches test.
 */
StringVec deleteStrings(const Matcher& test,
                        const StringVec& strings);

/** Tests string against vector of Matchers.
 *
 * \param test Vector of Matchers.
 *
 * \param subject String to test.
 *
 * \return True if subject matches at least one Matcher in test,
 * False otherwise.
 */
bool testString(const MatcherVec& test,
                const std::string& subject);

/** Tests string against vector of Matchers.
 *
 * \param test Vector of Matchers.
 *
 * \param subject String to test.
 *
 * \return True if subject does NOT match ANY Matcher in test,
 * False otherwise.
 */
bool doesNotMatch(const MatcherVec& test,
                  const std::string& subject);

/** A regular expression pattern has an error.
 */
class BadRePattern : public errors::NonFatal
{
public:
  /** ctor.
   * \param pattern offending pattern.
   * \param eMsg error message.
   */
  BadRePattern(const std::string& eMsg, const std::string& pattern);
  std::string getBriefMessage() const;

private:
  std::string eMsg_;
  std::string pattern_;
};

/** Ran out of memory while attempting to match a regular
 * expression.
 */
struct ExecOutOfMemory : public errors::NonFatal
{
  std::string getBriefMessage() const;
};

} // namespace matchers

#endif
