#include "numbers.h"
#include "strings.h"

#include <algorithm>
#include <cmath>
#include <iomanip>
#include <sstream>
#include <string>

namespace 
{
  const unsigned int PRECISION = 15;

  /** Compare two floats to see if they are (roughly) equal.
   *
   * For details, see
   * http://www.parashift.com/c++-faq-lite/newbie.html#faq-29.17
   */
  bool isEqual(double x, double y)
  {
    const double epsilon = 1e-5;
    return std::fabs(x - y) <= epsilon * std::fabs(x);
  }

  /** Convert a string to a floating point number.
   *
   * \param toConvert a string, which may be a single number (like
   * "3.14") a fraction (like "3/4") or a mixed number (like "3
   * 1/4").
   *
   * \param result The result is placed here.
   *
   * \return True if successfully converted; False otherwise.
   * Failure may be because toConvert was empty, or because a
   * denominator was zero.
   */
  bool convertToFloat(const std::string& toConvert,
                      double& result)
  {
    std::string trimmed = strings::rtrim(toConvert);
    if (std::count(trimmed.begin(), trimmed.end(), '/') > 1)
        return false;
    std::string numbersStr = strings::replace(trimmed, '/', " ");
    std::istringstream is(numbersStr);

    unsigned int noOfNumbers = 1;
    double no1, no2, no3;
    is >> no1;
    if (is.fail())
        return false;
    is >> no2;
    if (! is.fail())
        ++noOfNumbers;
    is >> no3;
    if (! is.fail())
        ++noOfNumbers;
    if (! is.eof())
        return false;

    if (noOfNumbers == 1)
        result = no1;
    else if (noOfNumbers == 2)
    {
        if (isEqual(no2, 0))
            return false;
        result = no1 / no2;
    }
    else
    {
        if (isEqual(no3, 0))
            return false;
        result = no1 + no2 / no3;
    }
    return true;
  }

  /** Convert a float to a string.
   */
  std::string convertToString (const double& number)
  {
    std::ostringstream os;
    os << std::setprecision(PRECISION) << std::fixed << number;
    return os.str();
  }

} // unnamed namespace

namespace numbers
{

Number::Number()
  : value_(0.0)
{
}

Number::Number(double value)
  : value_(value)
{
  if (value < 0)
    throw NegativeNumber(value);
}

Number::Number(const std::string& value)
{
  bool convertResult = convertToFloat(value, value_);
  if (! convertResult)
    throw Unconvertible(value);
  if (value_ < 0)
    throw NegativeNumber(value_);
}

double Number::getValue() const
{
  return value_;
}

std::string Number::getString() const
{
  return convertToString(value_);
}

bool Number::isZero() const
{
  return isEqual(value_, 0.0);
}

Percent::Percent()
  : value_(0)
{
}

Percent::Percent(int value)
  : value_(value)
{
  if (value < 0 ||  value > 100)
    throw NonPercent(value);
}

Percent::Percent(const std::string& value)
{
  double floatValue;
  bool convertResult = convertToFloat(value, floatValue);
  if (! convertResult)
    throw Unconvertible(value);
  if (floatValue < 0 || floatValue > 100)
    throw NonPercent(floatValue);
  value_ = static_cast<int>(floatValue);
}

int Percent::getValue() const
{
  return value_;
}

std::string Percent::getString() const
{
  return convertToString(value_);
}

StrNumber::StrNumber()
  : string_("0")
{
}

StrNumber::StrNumber(double value)
  : string_(convertToString(value))
{
  if (value < 0)
    throw NegativeNumber(value);
}

StrNumber::StrNumber(const std::string& value)
  : string_(value)
{
  double floatValue;
  if (! convertToFloat(value, floatValue))
    throw Unconvertible(value);
  if (floatValue < 0)
    throw NegativeNumber(floatValue);
}

double StrNumber::getValue() const
{
  double result;
  convertToFloat(string_, result);
  return result;
}

std::string StrNumber::getString() const
{
  return string_;
}

Unconvertible::Unconvertible(const std::string& badString)
    : badString_(badString) {}

std::string Unconvertible::getBriefMessage() const
{
    return "value not convertible to number: " + badString_;
}
NegativeNumber::NegativeNumber(const double& badNumber)
    : badNumber_(badNumber) {}

std::string NegativeNumber::getBriefMessage() const
{
  std::ostringstream ss;
  ss << "number out of range (must be positive): "
     << badNumber_;
  return ss.str();
}

NonPercent::NonPercent(const double& badNumber)
    : badNumber_(badNumber) {}

std::string NonPercent::getBriefMessage() const
{
  std::ostringstream ss;
  ss << "number out of range (must be between 0 and 100): "
     << badNumber_;
  return ss.str();
}

} // namespace numbers
