#include "parser.h"

#include "errors.h"
#include "stl.h"
#include "strings.h"

#include <algorithm>
#include <map>
#include <string>
#include <utility>
#include <stack>
#include <sstream>
#include <string>
#include <utility>
#include <vector>

namespace parser {

void Parser::addOption(const std::string& longOpt, unsigned int nArgs)
{
  if (longOpts_.find(longOpt) != longOpts_.end())
    throw errors::Logic();
  longOpts_.insert(std::make_pair(longOpt, nArgs));
}

void Parser::addOption(const std::string& longOpt,
                       char shortOpt, 
                       unsigned int nArgs)
{
  if (longOpts_.find(longOpt) != longOpts_.end())
    throw errors::Logic();
  if(shortOpts_.find(shortOpt) != shortOpts_.end())
    throw errors::Logic();
  longOpts_.insert(std::make_pair(longOpt, nArgs));
  shortOpts_.insert(std::make_pair(shortOpt, longOpt));
}

// parse helpers
namespace
{
  typedef std::stack<std::string> TokenStack;
  std::string getOptionForAbbrev(const std::string& abbrev,
                                 const std::vector<std::string>& options)
  {
    if (std::find(options.begin(), options.end(), abbrev) != options.end())
        return abbrev;
    std::vector<std::string> matches = strings::matchAbbrev(options, abbrev);
    if (matches.size() != 1)
        throw LongOptionNotValid(matches.size(), abbrev);
    return matches[0];
  }

  void processLongOptArgs
      (const std::string& option,
       TokenStack* tokens,
       OptArgsMapType* resultOptArgs,
       const std::map<std::string, unsigned int>& longOpts,
       bool hadExplicitValue)
  {
    unsigned int nArgs = longOpts.find(option)->second;
    if (tokens->size() < nArgs)
        throw InsufficientOptArgs(option, tokens->size(), nArgs);
    if ((nArgs == 0) && hadExplicitValue)
        throw NoExplicitValue(option);
    
    (*resultOptArgs)[option].push_back(std::vector<std::string>());
    std::vector<std::string>& optArgs = (*resultOptArgs)[option].back();
    for (unsigned i=0; i < nArgs; ++i)
    {
      optArgs.push_back(tokens->top());
      tokens->pop();
    }
  }

  void processLongOpt(TokenStack* tokens, 
                      OptArgsMapType* resultOptArgs,
                      const std::map<std::string, unsigned int>& longOpts)
  {
    std::string token = tokens->top();
    tokens->pop();

    // Delete "--"
    token = token.substr(2);

    bool hadExplicitValue;
    if (token.find("=") != std::string::npos)
    {
      std::vector<std::string> tokenAndArg = strings::explode(token,
              "=", 1);
      tokens->push(tokenAndArg.back());
      token = tokenAndArg.front();
      hadExplicitValue = true;
    }
    else
        hadExplicitValue = false;

    std::string option = getOptionForAbbrev(token, 
                                            stl::getKeys(longOpts));
    processLongOptArgs(option, tokens, 
                       resultOptArgs, longOpts, hadExplicitValue);
  }

  bool processShortOptArgs
       (const std::string& option,
        const std::string& currToken,
        TokenStack* tokens,
        OptArgsMapType* resultOptArgs,
        std::string::const_iterator currChar,
        unsigned int nArgs)
  {
    bool stop = false;
    (*resultOptArgs)[option].push_back(std::vector<std::string>());
    std::vector<std::string>& argTokens = (*resultOptArgs)[option].back();
    if (nArgs == 0)
      return stop;

    // Any characters left in arg? Pretend they're the next
    // arg, and stop consuming characters of arg.
    if (currChar + 1 < currToken.end())
    {
      tokens->push(std::string(currChar + 1, currToken.end()));
      stop = true;
    }

    if (tokens->size() < nArgs)
        throw InsufficientOptArgs(option, tokens->size(), nArgs);

    for (unsigned i=0; i < nArgs; ++i)
    {
      argTokens.push_back(tokens->top());
      tokens->pop();
    }
    return stop;
  }
        
  void processShortOpt(TokenStack* tokens,
                       OptArgsMapType* resultOptArgs,
                       const std::map<std::string,
                                      unsigned int>& longOpts,
                       const std::map<char, std::string>& shortOpts)
  {
    std::string currToken = tokens->top();
    tokens->pop();
    bool stop = false;

    // delete "-"
    currToken.erase(currToken.begin());
    for (std::string::const_iterator currChar = currToken.begin();
            currChar != currToken.end(); ++currChar)
    {
      std::map<char, std::string>::const_iterator shortOptAndLongOpt =
          shortOpts.find(*currChar);
      if (shortOptAndLongOpt == shortOpts.end())
          throw BadShortOption(*currChar);
      std::string longOpt = shortOptAndLongOpt->second;
      unsigned int nArgs = longOpts.find(longOpt)->second;
      stop = processShortOptArgs(longOpt, currToken, tokens,
                                 resultOptArgs, currChar,
                                 nArgs);
      if (stop)
        break;
    }
  }
} // parse helpers

Result Parser::parse(const std::vector<std::string>& args) const
{
  Result result;
  result.programName = args.front();

  for (std::map<std::string, unsigned int>::const_iterator longOptPair =
          longOpts_.begin();
          longOptPair != longOpts_.end();
          ++longOptPair)
  {
      result.optArgs.insert(std::make_pair(longOptPair->first,
                                           OptArgsVecType()));
  }

  TokenStack tokens;
  for (unsigned i=args.size() - 1; i > 0; --i)
    tokens.push(args[i]);

  while (! tokens.empty())
  {
    std::string currToken(tokens.top());
    if (currToken == "--")
    {
        tokens.pop();
        while (! tokens.empty())
        {
          result.posArgs.push_back(tokens.top());
          tokens.pop();
        }
    }
    else if (strings::startswith(currToken, "--"))
        processLongOpt(&tokens, &result.optArgs, longOpts_);

    else if ((strings::startswith(currToken, "-")) && 
             (currToken.length() > 1))
        processShortOpt(&tokens, &result.optArgs, longOpts_, shortOpts_);

    else
    {
      result.posArgs.push_back(currToken);
      tokens.pop();
    }
  }
  return result;
}

bool isSet(const std::vector<std::vector<std::string> >& x)
{
    return (! x.empty());
}

std::vector<std::string> vecsToVec
    (const std::vector<std::vector<std::string> >& x)
{
    std::vector<std::string> result;
    for (std::vector<std::vector<std::string> >::const_iterator innerVec =
            x.begin(); innerVec != x.end(); ++innerVec)
    {
        result.push_back((*innerVec)[0]);
    }
    return result;
}

std::vector<std::pair<std::string, std::string> > vecsToPairs
    (const std::vector<std::vector<std::string> >& x)
{
    std::vector<std::pair<std::string, std::string> > result;
    for (std::vector<std::vector<std::string> >::const_iterator innerVec =
            x.begin(); innerVec != x.end(); ++innerVec)
    {
        result.push_back(std::make_pair((*innerVec)[0], (*innerVec)[1]));
    }
    return result;
}

std::pair<bool, std::string>
getLastString(const OptArgsVecType& optArgs)
{
  std::pair<bool, std::string> result;
  if ((! optArgs.empty()) && (! optArgs.back().empty()))
  {
    result.first = true;
    result.second = optArgs.back().back();
  }
  else
    result.first = false;
  return result;
}

LongOptionNotValid::LongOptionNotValid(unsigned int nMatches,
                               const std::string& abbrev)
    : nMatches_(nMatches), abbrev_(abbrev) {}

std::string LongOptionNotValid::getBriefMessage() const
{
    std::string result = "long option ";
    result.append(abbrev_);
    result.append(" not valid:");
    if (nMatches_ == 0)
        result.append(" no");
    else
        result.append(" multiple");
    result.append(" possible matches");
    return result;
}

InsufficientOptArgs::InsufficientOptArgs(const std::string& option,
                                         unsigned int nArgs,
                                         unsigned int nArgsRequired)
    : option_(option), nArgs_(nArgs), 
      nArgsRequired_(nArgsRequired) {}

std::string InsufficientOptArgs::getBriefMessage() const
{
    std::ostringstream ss;
    ss << "insufficient number of option arguments"
       << " for option --" << option_ << "; "
       << nArgs_ << " given; " << nArgsRequired_ << " required";

    return ss.str();
}

NoExplicitValue::NoExplicitValue(const std::string& option)
    : option_(option) {}

std::string NoExplicitValue::getBriefMessage() const
{
    std::string result = "option --";
    result.append(option_);
    result.append(" does not take an argument");
    return result;
}

BadShortOption::BadShortOption(char option)
    : option_(option) {}

std::string BadShortOption::getBriefMessage() const
{
    std::string result = "-";
    result.append(1, option_);
    result.append(": invalid short option");
    return result;
}

} // namespace parser
