#ifndef PARSER_H
#define PARSER_H

#include <map>
#include <string>
#include <utility>
#include <vector>

#include "errors.h"

namespace parser {

typedef std::vector<std::string>               OptArgsType;
typedef std::vector<OptArgsType>               OptArgsVecType;
typedef std::map<std::string, OptArgsVecType>  OptArgsMapType;
typedef std::vector<std::string> TokensType;

struct Result
{
  std::string    programName;
  OptArgsMapType optArgs;
  TokensType     posArgs;
};

class Parser
{
public:
  // default ctor, copy ctor, assignment, etc OK
  void addOption(const std::string& longOpt, unsigned int nArgs);
  void addOption(const std::string& longOpt, 
                 char shortOpt, 
                 unsigned int nArgs);
  Result parse(const std::vector<std::string>& args) const;

private:
  std::map<std::string, unsigned int> longOpts_;
  std::map<char, std::string>         shortOpts_;
};

std::pair<bool, std::string> getLastString(const OptArgsVecType&);
bool isSet(const OptArgsVecType& x);
std::vector<std::string> vecsToVec (const OptArgsVecType& x);

std::vector<std::pair<std::string, std::string> > 
vecsToPairs (const OptArgsVecType& x);

/** Abbreviation given for long option not valid.
 */
class LongOptionNotValid : public errors::NonFatal
{
public:
  /** ctor.
   * \param nMatches Number of matches for given abbreviation.
   * \param abbrev The invalid abbreviation.
   */
  LongOptionNotValid(unsigned int nMatches,
                     const std::string& abbrev);
  std::string getBriefMessage() const;

private:
  unsigned int nMatches_;
  std::string abbrev_;
};

/** Insufficient number of option arguments given for a particular
 * option.
 */
class InsufficientOptArgs : public errors::NonFatal
{
public:
  /** ctor.
   * \param option Long option name.
   * \param nArgs Number of arguments given.
   * \param nArgsRequired Number of arguments actually required.
   */
  InsufficientOptArgs(const std::string& option,
                           unsigned int nArgs,
                           unsigned int nArgsRequired);
  std::string getBriefMessage() const;

private:
  std::string option_;
  unsigned int nArgs_;
  unsigned int nArgsRequired_;
};

/** User gave an option an explicit value, but it takes no
 * arguments.
 *
 * For example, user --ignore-case=true, but the --ignore-case
 * option does not take any arguments.
 */
class NoExplicitValue : public errors::NonFatal
{
public:
  /** ctor.
   * \param option Long name of option that was incorrectly given
   * an explicit value.
   */
  NoExplicitValue(const std::string& option);

private:
  std::string option_;
  std::string getBriefMessage() const;
};

/** Incorrect short option given.
 */
class BadShortOption : public errors::NonFatal
{
public:
  /** ctor.
   * \param option The offending option character.
   */
  BadShortOption(char option);
  std::string getBriefMessage() const;

private:
  char option_;
};

} // namespace parser

#endif
