#ifndef POINTERS_H
#define POINTERS_H

#include "errors.h"

/** Smart pointers.
 */
namespace pointers
{

/** Reference counted smart pointer.
 */
template <typename T>
class Shared
{
public:

  /** Construct empty shared ptr.
   *
   * Never throws.
   */
  Shared();

  /** Construct new shared pointer.
   * \param p Pointer that this instance will take ownership of.
   * This must be a pointer that was allocated with operator new.
   *
   * If this throws, it will first delete p.
   */
  Shared(T* p);

  /** Copy constructor. If toCopy is empty, constructs an empty instance;
   * otherwise, both the old and new instances will share ownership of the same
   * pointed-to object.
   *
   * Never throws.
   *
   * \param toCopy Instance to copy.
   */
  Shared(const Shared<T>& toCopy);

  /** Assignment operator. After the assignment, the instance on the
   * left-hand side shares ownership with the same pointed-to object
   * as the instance on the right-hand side. The left-hand side
   * instance loses ownership of whatever it pointed to before the
   * assignment.
   *
   * \param rhs Right-hand side instance.
   *
   * Never throws.
   */
  Shared<T>& operator= (const Shared<T>& rhs);

  /** Destructor. Reduces the reference count for the pointed-to
   * object by one. If the reference count thus becomes zero, the
   * pointed-to object is deleted.
   */
  ~Shared();

  bool unique() const;
  
  unsigned long int getCount() const;

  bool operator< (const Shared<T>& rhs) const;

  T& operator*() const;

  T* operator->() const;

  T* get() const;

  void swap(Shared<T>& rhs);

  void reset(T* p);

  void reset();

private:
  unsigned long int* count_;
  T* pointer_;
};


/////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////

template <typename T>
inline
Shared<T>::Shared()
  : count_(0), pointer_(0) {}

template <typename T>
inline
Shared<T>::Shared(T* p)
  : pointer_(p)
{
  try
  {
    count_ = new unsigned long int(1);
  }
  catch (...)
  {
    delete p;
    throw;
  }
}

template <typename T>
inline
Shared<T>::Shared(const Shared<T>& toCopy)
  : count_(toCopy.count_), pointer_(toCopy.pointer_)
{
  if (count_ != 0)
    ++(*count_);
}

template <typename T>
inline
Shared<T>& Shared<T>::operator= (const Shared<T>& rhs)
{
  Shared<T>(rhs).swap(*this);
  return *this;
}

template <typename T>
inline
Shared<T>::~Shared()
{
  if (count_ != 0)
  {
    --(*count_);
    if (*count_ == 0)
    {
      delete count_;
      delete pointer_;
    }
  }
}

template <typename T>
inline
bool Shared<T>::unique() const
{
  return this->getCount() == 1;
}

template <typename T>
inline
unsigned long int Shared<T>::getCount() const
{
  if (count_ == 0)
    return 0;
  return *count_;
}

template <typename T>
inline
bool Shared<T>::operator< (const Shared<T>& rhs) const
{
  return pointer_ < rhs.pointer_;
}

template <typename T>
inline
T& Shared<T>::operator*() const
{
  if (pointer_ == 0)
    throw errors::Logic();
  return *pointer_;
}

template <typename T>
inline
T* Shared<T>::operator->() const
{
  if (pointer_ == 0)
    throw errors::Logic();
  return pointer_;
}

template <typename T>
inline
T* Shared<T>::get() const
{
  return pointer_;
}

template <typename T>
inline
void Shared<T>::swap(Shared<T>& rhs)
{
  T* tempPtr = rhs.pointer_;
  unsigned long int* tempCount = rhs.count_;
  rhs.pointer_ = pointer_;
  rhs.count_ = count_;
  pointer_ = tempPtr;
  count_ = tempCount;
}

template <typename T>
inline
void Shared<T>::reset(T* p)
{
  Shared<T>(p).swap(*this);
}

template <typename T>
inline
void Shared<T>::reset()
{
  Shared<T>().swap(*this);
}

} // namespace pointers

#endif
