#include "predicate.h"

#include "food.h"
#include "stl.h"
#include "strings.h"

#include <map>
#include <string>
#include <utility>
#include <vector>

namespace predicate {

Regexps::Regexps() {}

Regexps::Regexps(const StringPairs& traits,
                 bool ignoreCase, bool exactMatch)
{
  std::map<std::string, matchers::MatcherVec* > traitMap;
  using std::make_pair;
  traitMap.insert(make_pair(food::NAME, &this->name));
  traitMap.insert(make_pair(food::GROUP, &this->group));
  traitMap.insert(make_pair(food::DATE, &this->date));
  traitMap.insert(make_pair(food::MEAL, &this->meal));
  traitMap.insert(make_pair(food::COMMENT, &this->comment));
  traitMap.insert(make_pair(food::ORDER, &this->order));
  traitMap.insert(make_pair(food::REFDESC, &this->refDesc));
  traitMap.insert(make_pair(food::INSTRUCTIONS, &this->instructions));
  traitMap.insert(make_pair(food::QUANTITY, &this->qty));
  traitMap.insert(make_pair(food::PCTREFUSE, &this->pctRefuse));
  traitMap.insert(make_pair(food::YIELD, &this->yield));
  traitMap.insert(make_pair(food::UNIT, &this->unit));

  std::vector<std::string> traitNames = stl::getKeys(traitMap);

  for (StringPairs::const_iterator stringPair =
          traits.begin(); stringPair != traits.end();
          ++stringPair)
  {
    std::vector<std::string> matches =
        strings::matchAbbrev(traitNames, stringPair->first);
    if (matches.size() != 1)
        throw PredicateBadTraitAbbreviation(matches.size(),
                                                 stringPair->first);
    const std::string& traitName = matches[0];
    matchers::MatcherVec* resultPtr =
      stl::checkedGet(traitMap, traitName);
    resultPtr->push_back(matchers::getMatcher(stringPair->second,
                                              ignoreCase, exactMatch));
  }
}

Predicate::Predicate(const Regexps& regexps)
    : regexps_(regexps)
{}

Predicate::StrTraits Predicate::getStrTraits(const food::Food& food) const
{
  StrTraits result;
  result.name = food.getName();
  result.group = food.getGroup();
  result.date = food.getDate();
  result.meal = food.getMeal();
  result.comment = food.getComment();
  result.order = food.getOrder();
  result.instructions = food.getInstructions();
  result.refDesc = food.getRefDesc();
  result.unit = food.getCurrentUnitName();
  result.qty = food.getQty().getString();
  result.pctRefuse = food.getPctRefuse().getString();
  result.yield = food.getYield().getString();
  return result;
}

bool Predicate::operator() (const food::Food& food) const
{
  StrTraits traits = getStrTraits(food);
  if (matchers::doesNotMatch(regexps_.name, traits.name))
    return false;
  if (matchers::doesNotMatch(regexps_.group, traits.group))
    return false;
  if (matchers::doesNotMatch(regexps_.date, traits.date))
    return false;
  if (matchers::doesNotMatch(regexps_.meal, traits.meal))
    return false;
  if (matchers::doesNotMatch(regexps_.comment, traits.comment))
    return false;
  if (matchers::doesNotMatch(regexps_.order, traits.order))
    return false;
  if (matchers::doesNotMatch(regexps_.refDesc, traits.refDesc))
    return false;
  if (matchers::doesNotMatch(regexps_.instructions, traits.instructions))
    return false;
  if (matchers::doesNotMatch(regexps_.qty, traits.qty))
    return false;
  if (matchers::doesNotMatch(regexps_.pctRefuse, traits.pctRefuse))
    return false;
  if (matchers::doesNotMatch(regexps_.yield, traits.yield))
    return false;
  if (matchers::doesNotMatch(regexps_.unit, traits.unit))
    return false;
  return true;
}
PredicateBadTraitAbbreviation::PredicateBadTraitAbbreviation(unsigned int nMatches,
                              const std::string& abbrev)
    : nMatches_(nMatches), abbrev_(abbrev) {}

std::string PredicateBadTraitAbbreviation::getBriefMessage() const
{
  std::string result = "invalid trait abbreviation: ";
  result.append(abbrev_);
  if (nMatches_ == 0)
    result.append(": no");
  else
    result.append(": multiple");
  result.append(" matches found");
  return result;
}

} // namespace predicate
