#include "reporter.h"

#include "food.h"
#include "pointers.h"
#include "sorter.h"
#include "stl.h"
#include "strings.h"
#include "totaler.h"

#include <algorithm>
#include <iomanip>
#include <map>
#include <ostream>
#include <set>
#include <string>
#include <utility>
#include <vector>

namespace
{

const unsigned int WIDTH_NAME      = 40;
const unsigned int WIDTH_AMOUNT    = 7;
const unsigned int WIDTH_PCT_GOAL  = 6;
const unsigned int WIDTH_PCT_TOTAL = 6;

}

namespace reporter
{

void Reporter::Name::printFood(const food::Food& food,
                               const CollectionNameLookup&,
                               std::ostream& os)
{
  os << std::left << std::setfill(' ') << std::setw(0)
     << food.getName() << '\n';
}

void Reporter::Blank::printFood(const food::Food&,
                                const CollectionNameLookup&,
                                std::ostream& os)
{
  os << std::left << std::setfill(' ') << std::setw(0)
     << '\n';
}

void Reporter::Info::printIfNotEmpty(const std::string& label,
                                     const std::string& string,
                                     std::ostream& os) const
{
  os << std::left << std::setfill(' ') << std::setw(0);
  if (! string.empty())
      os << label << ": " << string << '\n';
}

void Reporter::Info::printFood(const food::Food& food,
                               const CollectionNameLookup&,
                               std::ostream& os)
{
  os << std::left << std::setfill(' ') << std::setw(0);
  numbers::StrNumber qty = food.getQty();
  os << qty.getString() << ' ';

  os << food.getCurrentUnitName() << " (" << food::getGrams(food).getValue()
     << "g)\n";

  printIfNotEmpty("Group", food.getGroup(), os);
  printIfNotEmpty("Date", food.getDate(), os);
  printIfNotEmpty("Meal", food.getMeal(), os);
  printIfNotEmpty("Comment", food.getComment(), os);
  printIfNotEmpty("Order", food.getOrder(), os);

  numbers::Percent pctRefuse = food.getPctRefuse();
  if (pctRefuse.getValue() != 0)
  {
    os << "Refuse: " << pctRefuse.getValue() << " percent "
       << food.getRefDesc() << '\n';
  }
}

sorter::Sorter Reporter::Recipe::getSorter()
{
  std::vector<sorter::Key> keys;
  sorter::Key key(sorter::T_ORDER, sorter::D_ASCENDING);
  keys.push_back(key);
  return sorter::Sorter(sorter::Keys(keys), sorter::CompareTraits());
}

Reporter::Recipe::Recipe()
  : sorter_(getSorter()) {}

void Reporter::Recipe::printFood(const food::Food& food,
                                 const CollectionNameLookup&,
                                 std::ostream& os)
{
  os << std::left << std::setfill(' ') << std::setw(0);
  os << "Yield: " << std::fixed << std::setprecision(2);
  if (! food.getYield().isZero())
    os << food.getYield().getValue() << "g\n";
  else
    os << food::getRecipeGrams(food).getValue()
       << "g (estimated)\n";

  printIngredients(food.getIngredients(), os, sorter_);
  std::string instructions = food.getInstructions();
  if (! instructions.empty())
    os << '\n' << instructions << '\n';
}

void Reporter::Recipe::printIngredients
     (const std::vector<food::Food>& ingredients,
      std::ostream& os,
      const sorter::Sorter& sorter)
{
  std::vector<const food::Food*> sorted;
  for (std::vector<food::Food>::const_iterator ingredient =
       ingredients.begin(); ingredient != ingredients.end();
       ++ingredient)
  {
    sorted.push_back(&*ingredient);
  }
  std::sort(sorted.begin(), sorted.end(), sorter);
  for (std::vector<const food::Food*>::const_iterator 
          ingredient = sorted.begin(); 
          ingredient != sorted.end();
          ++ingredient)
  {
    printIngredient(**ingredient, os);
  }
}

void Reporter::Recipe::printIngredient(const food::Food& food,
                                       std::ostream& os)
{
  double grams = food::getGrams(food).getValue();
  double oz = grams / 28.35;
  os << food.getQty().getString() << " "
     << food.getCurrentUnitName() << " "

     << "(" << std::fixed << std::setprecision(0) << grams << "g, " 
     << std::setprecision(2) << oz << "oz) "

     << food.getName();
  std::string comment = food.getComment();
  if (! comment.empty())
      os << " (" << comment << ")";
  os << '\n';
}

void Reporter::Units::printFood(const food::Food& food,
                                const CollectionNameLookup&,
                                std::ostream& os)
{
  os << std::left << std::setfill(' ') << std::setw(0);
  std::vector<std::string> unitNames = food.getArbUnitNames();
  std::sort(unitNames.begin(), unitNames.end());

  os << std::fixed << std::setprecision(0);
  for (std::vector<std::string>::const_iterator unitName =
       unitNames.begin(); unitName != unitNames.end(); ++unitName)
  {
    os << "   " << *unitName << " (" 
                << food.getArbUnitAmount(*unitName).getValue()
                << "g)\n";
  }
}

void Reporter::Paste::printFood(const food::Food& food,
                                const CollectionNameLookup& lookup,
                                std::ostream& os)
{
  os << std::left << std::setfill(' ') << std::setw(0);
  StringVec unitNames = food.getArbUnitNames();
  std::map<std::string, numbers::Number> unitMap;
  for (StringVec::const_iterator unitName = unitNames.begin();
       unitName != unitNames.end(); ++unitName)
  {
    unitMap.insert(std::make_pair(*unitName,
                                  food.getArbUnitAmount(*unitName)));
  }

  unitNames = food.getAbsUnitNames();
  for (StringVec::const_iterator unitName = unitNames.begin();
       unitName != unitNames.end(); ++unitName)
  {
    unitMap.insert(std::make_pair(*unitName,
                                  food.getAbsUnitAmount(*unitName)));
  }

  std::string foodName = food.getName();
  std::string collectionName;
  CollectionNameLookup::const_iterator ptrAndNamePtr = lookup.find(&food);
  if (ptrAndNamePtr != lookup.end())
    collectionName = *ptrAndNamePtr->second;

  for (std::map<std::string, numbers::Number>::const_iterator
       unitNameAndAmt = unitMap.begin();
       unitNameAndAmt != unitMap.end();
       ++unitNameAndAmt)
  {
    printUnit(unitNameAndAmt->first, foodName,
              collectionName, os);
  }
}

void Reporter::Paste::printUnit(const std::string& unitName,
                                const std::string& foodName,
                                const std::string& collectionName,
                                std::ostream& os)
{
  os << std::left      << std::setfill(' ')           << std::setw(0)
     << "pantry -xsn " << quoteString(foodName)
     << " -cu "        << quoteString(unitName)
     << " "            << quoteString(collectionName) << "\n";
}

std::string Reporter::Paste::quoteString(const std::string& x)
{
  std::string result = x;
  result = strings::replace(result, '\"', "\\\"");
  result = strings::replace(result, '$', "\\$");
  result = strings::replace(result, '`', "\\`");
  result.insert(0, "\"");
  result.append("\"");
  return result;
}

namespace
{
  void printNutReportHeader(bool goalColumn, bool totalColumn,
                            std::ostream& os)
  {
    os << std::left                  << std::setfill(' ')
       << std::setw(WIDTH_NAME)      << "Nutrient name"
       << std::setw(WIDTH_AMOUNT)    << "Amount";

    int totalWidth = WIDTH_NAME + WIDTH_AMOUNT;
    if (goalColumn)
    {
        os << std::setw(WIDTH_PCT_GOAL) << "%G";
        totalWidth += WIDTH_PCT_GOAL;
    }

    if (totalColumn)
    {
        os << std::setw(WIDTH_PCT_TOTAL) << "%T" << "\n";
        totalWidth += WIDTH_PCT_TOTAL;
    }
    else
        os << "\n";

    os << std::setw(totalWidth) << std::setfill('-') << "" << "\n";
  }
}

Reporter::Nuts::Nuts
  (const pointers::Shared<const totaler::Totaler>& totaler)
  : totaler_(totaler) {}

Reporter::NutsTotal::NutsTotal
  (const pointers::Shared<const totaler::Totaler>& totaler)
  : totaler_(totaler) {}

Reporter::Goals::Goals
  (const pointers::Shared<const totaler::Totaler>& totaler)
  : totaler_(totaler) {}

Reporter::GoalsTotal::GoalsTotal
  (const pointers::Shared<const totaler::Totaler>& totaler)
  : totaler_(totaler) {}

void Reporter::Nuts::printFood(const food::Food& food,
                               const CollectionNameLookup&,
                               std::ostream& os)
{
  printNutReportHeader(true, true, os);
  std::map<std::string, totaler::FoodAll> items = 
      totaler_->getFoodAll(food);

  os << std::fixed << std::setprecision(0) << std::setfill(' ');
  for (std::map<std::string, totaler::FoodAll>::const_iterator it =
          items.begin(); it != items.end(); ++it)
  {
    os << std::setw(WIDTH_NAME)
       << it->first

       << std::setw(WIDTH_AMOUNT) 
       << it->second.nutAmt.getValue()

       << std::setw(WIDTH_PCT_GOAL);

    if (it->second.goalStatus == totaler::G_GOAL)
        os << it->second.pctGoal;
    else
        os << "";

    os << std::setw(WIDTH_PCT_TOTAL);
    if (it->second.hasPctTotal)
        os << it->second.pctTotal;
    else
        os << "";
    os << '\n';
  }
}

void Reporter::NutsTotal::printFooter(const SortedFoods&,
                                      const CollectionNameLookup&,
                                      std::ostream& os)
{
  os << "Sum of all nutrients:\n";
  printNutReportHeader(true, false, os);

  std::map<std::string, totaler::TotalAll> totals = 
      totaler_->getTotalAll();

  os << std::fixed << std::setprecision(0) << std::setfill(' ');
  for (std::map<std::string, totaler::TotalAll>::const_iterator it =
          totals.begin(); it != totals.end(); ++it)
  {
    os << std::setw(WIDTH_NAME) << it->first

              << std::setw(WIDTH_AMOUNT) 
              << it->second.totalNutAmt.getValue()

              << std::setw(WIDTH_PCT_GOAL);

    if (it->second.goalStatus == totaler::G_GOAL)
        os << it->second.pctGoal;
    else
        os << "";
    os << "\n";
  }
}

void Reporter::Goals::printFood(const food::Food& food,
                                const CollectionNameLookup&,
                                std::ostream& os)
{
  printNutReportHeader(true, true, os);
  std::vector<totaler::FoodGoal> items =
      totaler_->getFoodGoals(food);

  os << std::left << std::setprecision(0) << std::fixed
     << std::setfill(' ');

  for(std::vector<totaler::FoodGoal>::const_iterator it = 
          items.begin(); it != items.end(); ++it)
  {
    os << std::setw(WIDTH_NAME) << it->nutName;
    if (! it->hasNut)
    {
        os << "\n";
        continue;
    }

    os << std::setw(WIDTH_AMOUNT)
       << it->nutAmt.getValue()
       << std::setw(WIDTH_PCT_GOAL);

    if (it->goalStatus == totaler::G_GOAL)
      os << it->pctGoal;
    else
      os << "";

    os << std::setw(WIDTH_PCT_TOTAL);
    if (it->hasPctTotal)
        os << it->pctTotal;
    else
        os << "";
    os << '\n';
  }
}

void Reporter::GoalsTotal::printFooter
     (const SortedFoods&,
      const CollectionNameLookup&,
      std::ostream& os)
{
  os << "Sum of nutrients with a goal:\n";
  printNutReportHeader(true, false, os);
  os << std::left << std::setprecision(0) << std::setfill(' ')
            << std::fixed;
  std::vector<totaler::TotalGoal> totals = totaler_->getTotalGoals();

  for (std::vector<totaler::TotalGoal>::const_iterator it = 
          totals.begin();
          it != totals.end(); ++it)
  {
    os << std::setw(WIDTH_NAME)
       << it->nutName;

    if (! it->hasNut)
    {
        os << '\n';
        continue;
    }
    os << std::setw(WIDTH_AMOUNT)
       << it->nutAmt.getValue();

    if (it->goalStatus == totaler::G_GOAL)
    {
        os << std::setw(WIDTH_PCT_GOAL)
           << it->pctGoal;
    }
    os << '\n';
  }
}
void Reporter::Groups::printFooter(const SortedFoods& foods,
                                   const CollectionNameLookup&,
                                   std::ostream& os)
{
  const int WIDTH_NAME  = 50;
  const int WIDTH_COUNT = 10;

  os << std::left << std::setfill(' ')
     << std::setw(WIDTH_NAME) << "Group name"
     << std::setw(WIDTH_COUNT) << "Food count"
     << "\n";

  os << std::setfill('-') << std::setw(WIDTH_NAME + WIDTH_COUNT)
     << "" << "\n" << std::setfill(' ');
  
  std::map<std::string, unsigned int> groupCount;
  for (SortedFoods::const_iterator food = foods.begin();
          food != foods.end(); ++food)
  {
    std::string groupName = (*food)->getGroup();
    groupCount[groupName] += 1;
  }

  for(std::map<std::string, unsigned int>::const_iterator countIt = 
          groupCount.begin(); countIt != groupCount.end(); ++countIt)
  {
    os << std::setw(WIDTH_NAME)  << countIt->first
       << std::setw(WIDTH_COUNT) << countIt->second
       << "\n";
  }
  os << std::setfill('-') << std::setw(WIDTH_NAME + WIDTH_COUNT)
     << "" << "\n" << std::setfill(' ');

  os << std::setw(WIDTH_NAME) << "  Total"
     << std::setw(WIDTH_COUNT) << foods.size() << "\n";
}

Reporter::Reporter(const StringVec& reports,
                   const StringPairs& goals,
                   const StringVec& sortKeys,
                   const StringPairs& sortOrders)
  : sorter_(sorter::Keys(sortKeys), sorter::CompareTraits(sortOrders)),
    goals_(totaler::getGoals(goals)),
    reportIDs_(reports) {}

Reporter::Reports::Reports(const ReportVec& reports)
  : reports_(reports) {}

std::string
Reporter::Reports::print(const FoodVecMap& collections,
                         const sorter::Sorter& sorter)
{
  std::vector<const food::Food*> sortedFoods;
  std::map<const food::Food*, const std::string*> lookup;
  for (FoodVecMap::const_iterator collNameAndVec =
       collections.begin();
       collNameAndVec != collections.end();
       ++collNameAndVec)
  {
    const std::string* collName = &(collNameAndVec->first);
    const std::vector<food::Food>& foodVec = collNameAndVec->second;
    for (std::vector<food::Food>::const_iterator food = foodVec.begin();
         food != foodVec.end(); ++food)
    {
      lookup.insert(std::make_pair(&(*food), collName));
      sortedFoods.push_back(&(*food));
    }
  }
  std::sort(sortedFoods.begin(), sortedFoods.end(), sorter);

  std::ostringstream os;

  for (ReportVec::iterator report = reports_.begin();
       report != reports_.end(); ++report)
  {
    (*report)->printHeader(sortedFoods, lookup, os);
  }

  for (SortedFoods::const_iterator food = sortedFoods.begin();
       food != sortedFoods.end(); ++food)
  {
    for (ReportVec::iterator report = reports_.begin();
         report != reports_.end(); ++report)
    {
      (*report)->printFood(**food, lookup, os);
    }
  }

  for (ReportVec::iterator report = reports_.begin();
       report != reports_.end(); ++report)
  {
    (*report)->printFooter(sortedFoods, lookup, os);
  }
  return os.str();
}

std::string Reporter::Reports::print(const food::Food& food)
{
  std::vector<const food::Food*> sortedFoods (1, &food);
  std::map<const food::Food*, const std::string*> lookup;
  std::ostringstream os;

  for (ReportVec::iterator report = reports_.begin();
       report != reports_.end(); ++report)
  {
    (*report)->printHeader(sortedFoods, lookup, os);
  }

  for (ReportVec::iterator report = reports_.begin();
       report != reports_.end(); ++report)
  {
    (*report)->printFood(food, lookup, os);
  }

  for (ReportVec::iterator report = reports_.begin();
       report != reports_.end(); ++report)
  {
    (*report)->printFooter(sortedFoods, lookup, os);
  }
  return os.str();
}

std::string
Reporter::print(const FoodVecMap& collections) const
{
  Reports reports = reportIDs_.getReports(collections, goals_);
  return reports.print(collections, sorter_);
}

std::string
Reporter::print(const food::Food& food) const
{
  Reports reports = reportIDs_.getReports(food, goals_);
  return reports.print(food);
}

Reporter::ReportIDs::ReportIDs(const StringVec& strings)
{
  std::map<std::string, ReportID> reportIDMap;
  using std::make_pair;
  reportIDMap.insert(make_pair("name", R_NAME));
  reportIDMap.insert(make_pair("blank", R_BLANK));
  reportIDMap.insert(make_pair("info", R_INFO));
  reportIDMap.insert(make_pair("recipe", R_RECIPE));
  reportIDMap.insert(make_pair("units", R_UNITS));
  reportIDMap.insert(make_pair("paste", R_PASTE));
  reportIDMap.insert(make_pair("groups", R_GROUPS));
  reportIDMap.insert(make_pair("nuts", R_NUTS));
  reportIDMap.insert(make_pair("Nuts", R_NUTS_TOTAL));
  reportIDMap.insert(make_pair("goals", R_GOALS));
  reportIDMap.insert(make_pair("Goals", R_GOALS_TOTAL));

  StringVec reportNames = stl::getKeys(reportIDMap);
  for (StringVec::const_iterator abbrev = strings.begin();
       abbrev != strings.end(); ++abbrev)
  {
    StringVec matches = strings::matchAbbrev(reportNames, *abbrev);
    if (matches.size() != 1)
      throw BadReportName(matches.size(), *abbrev);
    idVec_.push_back(stl::checkedGet(reportIDMap, matches[0]));
  }
}

pointers::Shared<const totaler::Totaler>
Reporter::ReportIDs::getTotaler
                    (const std::vector<const food::Food*>& foods,
                     const totaler::Goals& goals) const
{
  pointers::Shared<const totaler::Totaler> totaler;
  std::set<ReportID> totalerReportIDs;
  totalerReportIDs.insert(R_NUTS);
  totalerReportIDs.insert(R_NUTS_TOTAL);
  totalerReportIDs.insert(R_GOALS);
  totalerReportIDs.insert(R_GOALS_TOTAL);
  for (std::vector<ReportID>::const_iterator reportID =
       idVec_.begin();
       reportID != idVec_.end();
       ++reportID)
  {
    if (totalerReportIDs.find(*reportID) != totalerReportIDs.end())
    {
      totaler.reset(new totaler::Totaler(foods, goals));
      break;
    }
  }
  return totaler;
}

Reporter::ReportVec Reporter::ReportIDs::getReportVec
(const pointers::Shared<const totaler::Totaler>& totaler) const
{
  ReportVec reportVec;
  for (std::vector<ReportID>::const_iterator reportID =
       idVec_.begin();
       reportID != idVec_.end();
       ++reportID)
  {
    if (*reportID == R_NAME)
      reportVec.push_back(ReportPtr(new Name()));
    else if (*reportID == R_BLANK)
      reportVec.push_back(ReportPtr(new Blank()));
    else if (*reportID == R_INFO)
      reportVec.push_back(ReportPtr(new Info()));
    else if (*reportID == R_RECIPE)
      reportVec.push_back(ReportPtr(new Recipe()));
    else if (*reportID == R_UNITS)
      reportVec.push_back(ReportPtr(new Units()));
    else if (*reportID == R_PASTE)
      reportVec.push_back(ReportPtr(new Paste()));
    else if (*reportID == R_GROUPS)
      reportVec.push_back(ReportPtr(new Groups()));
    else if (*reportID == R_NUTS)
      reportVec.push_back(ReportPtr(new Nuts(totaler)));
    else if (*reportID == R_NUTS_TOTAL)
      reportVec.push_back(ReportPtr(new NutsTotal(totaler)));
    else if (*reportID == R_GOALS)
      reportVec.push_back(ReportPtr(new Goals(totaler)));
    else if (*reportID == R_GOALS_TOTAL)
      reportVec.push_back(ReportPtr(new GoalsTotal(totaler)));
  }
  return reportVec;
}

Reporter::Reports Reporter::ReportIDs::getReports
          (const FoodVecMap& collections,
           const totaler::Goals& goals) const
{
  std::vector<const food::Food*> foods =
    collections::getFoodPtrs(collections);

  pointers::Shared<const totaler::Totaler> totaler =
    getTotaler(foods, goals);

  ReportVec reportVec = getReportVec(totaler);
  return Reports(getReportVec(totaler));
}

Reporter::Reports Reporter::ReportIDs::getReports
                  (const food::Food& food,
                   const totaler::Goals& goals) const
{
  std::vector<const food::Food*> foods(1, &food);
  pointers::Shared<const totaler::Totaler> totaler =
    getTotaler(foods, goals);

  return Reports(getReportVec(totaler));
}

BadReportName::BadReportName(unsigned int nMatches,
                             const std::string& reportName)
    : nMatches_(nMatches), reportName_(reportName) {}

std::string BadReportName::getBriefMessage() const
{
    std::string result;
    if (nMatches_ == 0)
        result.append("no");
    else
        result.append("multiple");
    result.append(" matching reports found for given name ");
    result.append(reportName_);
    return result;
}

} // namespace reporter
