#ifndef REPORTER_H
#define REPORTER_H

#include "collections.h"
#include "errors.h"
#include "food.h"
#include "pointers.h"
#include "sorter.h"
#include "totaler.h"

#include <iosfwd>
#include <map>
#include <string>
#include <utility>
#include <vector>

namespace reporter {

typedef std::vector<std::string> StringVec;
typedef std::vector<std::pair<std::string, std::string> > StringPairs;
typedef std::vector<food::Food> FoodVec;
typedef std::map<std::string, std::vector<food::Food> > FoodVecMap;

class Reporter
{
public:

  /** Construct Reporter object.
   *
   * Offers strong guarantee.
   */
  Reporter(const StringVec& reports,
           const StringPairs& goals,
           const StringVec& sortKeys,
           const StringPairs& sortOrders);

  /** Print reports.
   *
   * Offers strong guarantee.
   */
  std::string print(const FoodVecMap& collections) const;
  std::string print(const food::Food& food) const;

private:
  typedef std::vector<const food::Food*> SortedFoods;
  typedef std::map<const food::Food*, const std::string*>
          CollectionNameLookup;

  class Report
  {
  public:
    virtual void printHeader(const SortedFoods&,
                             const CollectionNameLookup&,
                             std::ostream&) {}
    virtual void printFood(const food::Food&,
                           const CollectionNameLookup&,
                           std::ostream&) {}
    virtual void printFooter(const SortedFoods&,
                             const CollectionNameLookup&,
                             std::ostream&) {}
    virtual ~Report() {}
  };

  class Name : public Report
  {
  public:
    void printFood(const food::Food& food, 
                   const CollectionNameLookup& lookup,
                   std::ostream& os);
  };

  class Blank : public Report
  {
  public:
    void printFood(const food::Food& food,
                   const CollectionNameLookup& lookup,
                   std::ostream& os);
  };

  class Info : public Report
  {
  public:
    void printFood(const food::Food& food,
                   const CollectionNameLookup& lookup,
                   std::ostream& os);

  private:
    void printIfNotEmpty(const std::string& label,
                         const std::string& string,
                         std::ostream& os) const;
  };

  class Recipe : public Report
  {
  public:
    Recipe();
    void printFood(const food::Food& food,
                   const CollectionNameLookup& lookup,
                   std::ostream& os);

  private:
    sorter::Sorter sorter_;
    void printIngredient(const food::Food& food, std::ostream& os);
    void printIngredients(const std::vector<food::Food>& ingredients,
                          std::ostream& os,
                          const sorter::Sorter& sorter);
    sorter::Sorter getSorter();
  };

  class Units : public Report
  {
  public:
    void printFood(const food::Food& food,
                   const CollectionNameLookup& lookup,
                   std::ostream& os);
  };

  class Paste : public Report
  {
  public:
    void printFood(const food::Food& food,
                   const CollectionNameLookup& lookup,
                   std::ostream& os);
  private:
    void printUnit(const std::string& unitName,
                   const std::string& foodName,
                   const std::string& collectionName,
                   std::ostream& os);
    std::string quoteString(const std::string&);
  };

  class Nuts : public Report
  {
  public:
    Nuts(const pointers::Shared<const totaler::Totaler>& totaler);
    void printFood(const food::Food& food,
                   const CollectionNameLookup& lookup,
                   std::ostream& os);

  private:
    pointers::Shared<const totaler::Totaler> totaler_;
  };

  class NutsTotal : public Report
  {
  public:
    NutsTotal(const pointers::Shared<const totaler::Totaler>& totaler);
    void printFooter(const SortedFoods& foods,
                     const CollectionNameLookup& lookup,
                     std::ostream& os);

  private:
    pointers::Shared<const totaler::Totaler> totaler_;
  };

  class Goals : public Report
  {
  public:
    Goals(const pointers::Shared<const totaler::Totaler>& totaler);
    void printFood(const food::Food& food,
                   const CollectionNameLookup& lookup,
                   std::ostream& os);

  private:
    pointers::Shared<const totaler::Totaler> totaler_;
  };

  class GoalsTotal : public Report
  {
  public:
    GoalsTotal(const pointers::Shared<const totaler::Totaler>& totaler);
    void printFooter(const SortedFoods& foods,
                     const CollectionNameLookup& lookup,
                     std::ostream& os);

  private:
    pointers::Shared<const totaler::Totaler> totaler_;
  };

  class Groups : public Report
  {
  public:
    void printFooter(const SortedFoods& foods,
                     const CollectionNameLookup& lookup,
                     std::ostream& os);
  };

  typedef pointers::Shared<Report> ReportPtr;
  typedef std::vector<ReportPtr> ReportVec;

  class Reports
  {
  public:
    Reports(const ReportVec& reports);
    std::string print(const FoodVecMap& collections,
                      const sorter::Sorter& sorter);
    std::string print(const food::Food& food);

  private:
    ReportVec reports_;
  };

  class ReportIDs
  {
  public:
    ReportIDs(const StringVec& strings);
    Reports getReports(const FoodVecMap& collections,
                       const totaler::Goals& goals) const;
    Reports getReports(const food::Food& food,
                       const totaler::Goals& goals) const;

  private:
    enum ReportID
    {
      R_NAME,
      R_BLANK,
      R_INFO,
      R_RECIPE,
      R_UNITS,
      R_PASTE,
      R_GROUPS,

      // these reports need totals
      R_NUTS,
      R_NUTS_TOTAL,
      R_GOALS,
      R_GOALS_TOTAL
    };

    std::vector<ReportID> idVec_;
    pointers::Shared<const totaler::Totaler> getTotaler
        (const std::vector<const food::Food*>& foods,
         const totaler::Goals& goals)const;

    ReportVec getReportVec
    (const pointers::Shared<const totaler::Totaler>& totaler) const;
  };

  sorter::Sorter sorter_;
  totaler::Goals goals_;
  ReportIDs reportIDs_;
};

/** String given for name of a report invalid.
 */
class BadReportName : public errors::NonFatal
{
public:
    /** ctor.
     * \param nMatches Number of matches for given report name.
     * \param reportName Invalid string given for report name.
     */
    BadReportName(unsigned int nMatches,
                  const std::string& reportName);
    std::string getBriefMessage() const;

private:
    unsigned int nMatches_;
    std::string reportName_;
};

} // namespace reporter

#endif
