#include "searcher.h"

#include "errors.h"
#include "collections.h"
#include "food.h"
#include "predicate.h"
#include "stl.h"
#include "strings.h"

#include <algorithm>
#include <list>
#include <string>
#include <utility>
#include <vector>

namespace searcher {

namespace {

std::pair<bool, unsigned int>
getLimit(const std::pair<bool, std::string>& stringPair)
{
  std::pair<bool, unsigned int> result;
  if (! stringPair.first)
  {
    result.first = false;
    return result;
  }

  bool converted = strings::deStringify(stringPair.second, result.second);
  if (! converted)
    throw Limit();
  result.first = true;
  return result;
}

SelectionsVec&
fillSelectionsVec(collections::Collections::value_type& searchColl,
               SelectionsVecs& selectionsVecs,
               const predicate::Predicate& predicate)
{
  SelectionsVec& resultVec = selectionsVecs[searchColl.first];
  std::list<food::Food>& searchList = searchColl.second;

  for (std::list<food::Food>::iterator food = 
          searchList.begin(); food != searchList.end();
          ++food)
  {
    if (predicate(*food))
      resultVec.push_back(food);
  }
  return resultVec;
}

bool processLimit(unsigned int& totalFound,
                  SelectionsVec& resultVec,
                  const unsigned int& limit)
{
  if (totalFound + resultVec.size() >= limit)
  {
    unsigned int nToErase = (totalFound + resultVec.size()
            - limit);
    resultVec.erase(resultVec.begin(), resultVec.begin() + nToErase);
    return true;
  }
  else
      totalFound += resultVec.size();
  return false;
}

} // unnamed namespace

Searcher::Searcher(const std::vector<std::pair<std::string,
                                         std::string> >& traits,
             const std::pair<bool, std::string>& limit,
             bool ignoreCase,
             bool exactMatch,
             const std::vector<std::string>& toSearch)
    : predicate_(predicate::Predicate(predicate::Regexps(traits,
                                                         ignoreCase, 
                                                         exactMatch))),
      limit_(getLimit(limit)),
      toSearch_(toSearch) {}

SelectionsVecs
Searcher::process(collections::Collections& database) const
{
  SelectionsVecs result;
  collections::CollectionPtrs collsToSearch =
      collections::getCollectionPtrs(database, toSearch_);
  for (collections::CollectionPtrs::iterator
       collToSearch = collsToSearch.begin();
       collToSearch != collsToSearch.end();
       ++collToSearch)
  {
    unsigned int totalFound = 0;
    SelectionsVec& resultVec =
        fillSelectionsVec(**collToSearch, result, predicate_);
    if (limit_.first)
    {
      bool doneSearching = processLimit(totalFound, 
                                        resultVec, limit_.second);
      if (doneSearching)
        return result;
    }
  }
  return result;
}

CollectionNotFound::CollectionNotFound(const std::string& name)
    : name_(name) {}

std::string CollectionNotFound::getBriefMessage() const
{
    std::string result = name_;
    result.append(": collection not found");
    return result;
}

std::string Limit::getBriefMessage() const
{
    return "invalid value for --limit option";
}

} // namespace searcher
