#include "serializer.h"

#include "collections.h"
#include "fd.h"
#include "food.h"
#include "numbers.h"
#include "stl.h"

#include <istream>
#include <list>
#include <map>
#include <ostream>
#include <sstream>
#include <string>
#include <utility>
#include <vector>

namespace
{
// don't call this version (with all caps); conflicts with GNU build
// system macros
const int FILE_VER = 2;

std::string getFileHeader()
{
    std::ostringstream ss;
    ss << "Pantry foods file version " << FILE_VER;
    return ss.str();
}

} // namespace

// The code in here assumes that map::size is constant time.
// However, the standard does not require this--it says
// "should" be constant time.
namespace serializer
{

void save(const db::Db& database, const std::string& filename)
{
  fd::Fdbuf fdbuf(filename, fd::Write);
  std::ostream os(&fdbuf);
  os.exceptions(std::ostream::eofbit | 
                std::ostream::failbit | 
                std::ostream::badbit);

  Saver saver;
  try
  {
    os << getFileHeader() << '\n';
    saver.save(database, os);
  }
  catch (std::ios_base::failure&)
  {
    throw SaveWriteData(filename);
  }
}

db::Db load(const std::string& filename)
{
  fd::Fdbuf fdbuf(filename, fd::Read);
  std::istream is(&fdbuf);
  is.exceptions(std::ostream::eofbit | 
                std::ostream::failbit | 
                std::ostream::badbit);
  Loader loader;
  try
  {
    std::string header;
    std::getline(is, header);
    if (header != getFileHeader())
      throw UnrecognizedFile(filename);
    return loader.load(is);
  }
  catch (std::ios_base::failure&)
  {
    throw LoadReadFile(filename);
  }
}

void Saver::save(const db::Db& db, std::ostream& os)
{
  const int PRECISION = 12;
  os.precision(PRECISION);

  result_ = &os;
  clear();
  createDBMemberIDs(db.strPool_, db.absUnitPool_);
  createFoodIDs(db.collections_);
  createStringIDs();
  createAbsUnitIDs();
  saveStringData();
  saveAbsUnitData();
  saveNumMapData();
  saveStrNumberData();
  savePctData();
  saveNumberData();
  saveStrPoolIDs();
  saveAbsUnitPoolIDs();
  saveDBMemberIDs(db.strPool_, db.absUnitPool_);
  saveCollectionIDs(db.collections_);
  saveIngredientIDs();
  saveAllFoodIDs();
}

void Saver::clear()
{
  foodIDs_.clear();
  stringIDs_.clear();
  strPoolIDs_.clear();
  absUnitPoolIDs_.clear();
  numMapIDs_.clear();
  ingredientVecIDs_.clear();
  strNumberIDs_.clear();
  pctIDs_.clear();
  numberIDs_.clear();
  absUnitIDs_.clear();
}

void Saver::createDBMemberIDs(const StrPoolPtr& strPool,
                              const AbsUnitPoolPtr& absUnitPool)
{
  strPoolIDs_.insert(std::make_pair(strPool.get(), strPoolIDs_.size()));
  absUnitPoolIDs_.insert(std::make_pair(absUnitPool.get(),
                                        absUnitPoolIDs_.size()));
}

void Saver::createFoodIDs(const collections::Collections& collections)
{
  for (collections::Collections::const_iterator collection =
       collections.begin(); collection != collections.end();
       ++collection)
  {
    const std::list<food::Food>& foods = collection->second;
    for (std::list<food::Food>::const_iterator food = foods.begin();
         food != foods.end(); ++food)
    {
      createFoodID(*food);
    }
  }
}

void Saver::createFoodID(const food::Food& food)
{
  foodIDs_.insert(std::make_pair(&food, foodIDs_.size()));
  createComponentIDs(food);
  for (std::vector<food::Food>::const_iterator ingredient =
       food.ingredients_->begin();
       ingredient != food.ingredients_->end(); ++ingredient)
  {
    createFoodID(*ingredient);
  }
}

void Saver::createComponentIDs(const food::Food& food)
{
  using std::make_pair;
  strPoolIDs_.insert(make_pair(food.strPool_.get(), 
                               strPoolIDs_.size()));
  absUnitPoolIDs_.insert(make_pair(food.absUnitPool_.get(),
                                  absUnitPoolIDs_.size()));

  numMapIDs_.insert(make_pair(food.nuts_.get(),
                              numMapIDs_.size()));
  numMapIDs_.insert(make_pair(food.arbUnits_.get(),
                              numMapIDs_.size()));
  ingredientVecIDs_.insert(make_pair(food.ingredients_.get(),
                                     ingredientVecIDs_.size()));

  strNumberIDs_.insert(make_pair(food.qty_.get(), strNumberIDs_.size()));
  pctIDs_.insert(make_pair(food.pctRefuse_.get(), pctIDs_.size()));
  numberIDs_.insert(make_pair(food.yield_.get(), numberIDs_.size()));
}

void Saver::createStringIDs()
{
  for (std::map<const strPool::StrPool*, ULong>::const_iterator
       poolPtrAndID = strPoolIDs_.begin();
       poolPtrAndID != strPoolIDs_.end();
       ++poolPtrAndID)
  {
    const strPool::StrPool& pool = *poolPtrAndID->first;
    for (std::map<std::string, strPool::StrPtr>::const_iterator
         stringAndStrPtr = pool.stringPtrMap_.begin();
         stringAndStrPtr != pool.stringPtrMap_.end();
         ++stringAndStrPtr)
    {
      const strPool::StrPtr& strPtr = stringAndStrPtr->second;
      // The string pool may contain strings that are
      // "garbage"--that is, strings that are no longer
      // actually used. This code ensures that only strings
      // that are actually in use are saved. However, you
      // must ensure that saveStrPoolIDs behaves
      // accordingly.
      if (! strPtr.unique())
        stringIDs_.insert(std::make_pair(strPtr.get(), stringIDs_.size()));
    }
  }
}

void Saver::createAbsUnitIDs()
{
  for (std::map<const absUnitPool::AbsUnitPool*, ULong>::const_iterator
       poolPtrAndID = absUnitPoolIDs_.begin();
       poolPtrAndID != absUnitPoolIDs_.end();
       ++poolPtrAndID)
  {
    const std::map<std::string, absUnitPool::AbsUnitPtr>& absUnitMap =
      (*poolPtrAndID->first).absUnits_;
    for (std::map<std::string, absUnitPool::AbsUnitPtr>::const_iterator 
         strAndUnitPtr = absUnitMap.begin();
         strAndUnitPtr != absUnitMap.end();
         ++strAndUnitPtr)
    {
      const absUnitPool::AbsUnit* absUnitPtr =
        strAndUnitPtr->second.get();
      absUnitIDs_.insert(std::make_pair(absUnitPtr, absUnitIDs_.size()));
    }
  }
}

void Saver::saveStringData()
{
  *result_ << "Phase 3.1: save string data\n";
  *result_ << stringIDs_.size() << '\n';
  for (std::map<const std::string*, ULong>::const_iterator strPtrAndID =
       stringIDs_.begin();
       strPtrAndID != stringIDs_.end(); ++strPtrAndID)
  {
    const std::string& string = *strPtrAndID->first;
    *result_ << strPtrAndID->second << ' '
            << string.size()       << ' '
            << string              << '\n';
  }
  *result_ << '\n';
}

void Saver::saveAbsUnitData()
{
  *result_ << "Phase 3.2: save abs unit data\n";
  *result_ << absUnitIDs_.size() << '\n';
  for (std::map<const absUnitPool::AbsUnit*, ULong>::const_iterator
       absUnitPtrAndID = absUnitIDs_.begin();
       absUnitPtrAndID != absUnitIDs_.end();
       ++absUnitPtrAndID)
  {
    const absUnitPool::AbsUnit& absUnit = *absUnitPtrAndID->first;
    *result_ << absUnitPtrAndID->second << ' '
            << absUnit.name.size()     << ' '
            << absUnit.name            << ' '
            << absUnit.value.value_    << '\n';
  }
  *result_ << '\n';
}

void Saver::saveNumMapData()
{
  *result_ << "Phase 3.3: save num map data\n";
  *result_ << numMapIDs_.size() << '\n';
  for (std::map<const StrPtrNumMap*, ULong>::const_iterator mapPtrAndID =
       numMapIDs_.begin(); mapPtrAndID != numMapIDs_.end();
       ++mapPtrAndID)
  {
    const StrPtrNumMap& numMap = *mapPtrAndID->first;
    *result_ << mapPtrAndID->second << ' ' << numMap.size() << '\n';
    for (StrPtrNumMap::const_iterator strPtrAndNum = numMap.begin();
         strPtrAndNum != numMap.end(); ++strPtrAndNum)
    {
      const std::string& string = *strPtrAndNum->first;
      *result_ << stl::checkedGet(stringIDs_, &string) << ' '
              << strPtrAndNum->second.value_         << '\n';
    }
  }
  *result_ << '\n';
}

void Saver::saveStrNumberData()
{
  *result_ << "Phase 3.4: save strNumber data\n";
  *result_ << strNumberIDs_.size() << '\n';
  for (std::map<const numbers::StrNumber*, ULong>::const_iterator
       strNumberPtrAndID = strNumberIDs_.begin();
       strNumberPtrAndID != strNumberIDs_.end();
       ++strNumberPtrAndID)
  {
    const numbers::StrNumber& strNumber = *strNumberPtrAndID->first;
    *result_ << strNumberPtrAndID->second << ' '
            << strNumber.string_.size()  << ' '
            << strNumber.string_ << '\n';
  }
  *result_ << '\n';
}

void Saver::savePctData()
{
  *result_ << "Phase 3.5: save percent data\n";
  *result_ << pctIDs_.size() << '\n';
  for (std::map<const numbers::Percent*, ULong>::const_iterator
       pctPtrAndID = pctIDs_.begin();
       pctPtrAndID != pctIDs_.end();
       ++pctPtrAndID)
  {
    *result_ << pctPtrAndID->second << ' ' 
            << (*pctPtrAndID->first).value_
            << '\n';
  }
  *result_ << '\n';
}

void Saver::saveNumberData()
{
  *result_ << "Phase 3.6: save number data\n";
  *result_ << numberIDs_.size() << '\n';
  for (std::map<const numbers::Number*, ULong>::const_iterator
       numberPtrAndID = numberIDs_.begin();
       numberPtrAndID != numberIDs_.end();
       ++numberPtrAndID)
  {
    *result_ << numberPtrAndID->second << ' ' 
            << (*numberPtrAndID->first).value_
            << '\n';
  }
  *result_ << '\n';
}

void Saver::saveStrPoolIDs()
{
  *result_ << "Phase 4.1: save strPool IDs\n";
  *result_ << strPoolIDs_.size() << '\n';
  for (std::map<const strPool::StrPool*, ULong>::const_iterator ptrAndID =
       strPoolIDs_.begin();
       ptrAndID != strPoolIDs_.end(); ++ptrAndID)
  {
    const strPool::StrPool& pool = *ptrAndID->first;
    *result_ << ptrAndID->second << ' ';

    // Ensure that only strings that are in use are saved.
    // The IDs must be put in a vector first so we can count
    // how many we actually need to save before writing the number
    // saved to the stream.
    std::vector<ULong> idsInUse;
    for (std::map<std::string, strPool::StrPtr>::const_iterator
         stringAndStrPtr = pool.stringPtrMap_.begin();
         stringAndStrPtr != pool.stringPtrMap_.end();
         ++stringAndStrPtr)
    {
      const strPool::StrPtr& strPtr = stringAndStrPtr->second;
      const std::string& string = *strPtr;
      if (! strPtr.unique())
        idsInUse.push_back(stl::checkedGet(stringIDs_, &string));
    }
    *result_ << idsInUse.size() << ' ';
    for (std::vector<ULong>::const_iterator id = idsInUse.begin();
         id != idsInUse.end(); ++id)
    {
      *result_ << *id << ' ';
    }
    *result_ << '\n';
  }
  *result_ << '\n';
}

void Saver::saveAbsUnitPoolIDs()
{
  *result_ << "Phase 4.2: save absUnitPool IDs\n";
  *result_ << absUnitPoolIDs_.size() << '\n';
  for (std::map<const absUnitPool::AbsUnitPool*, ULong>::const_iterator
       ptrAndID = absUnitPoolIDs_.begin();
       ptrAndID != absUnitPoolIDs_.end();
       ++ptrAndID)
  {
    const absUnitPool::AbsUnitPool& pool = *ptrAndID->first;
    *result_ << ptrAndID->second << ' ';
    *result_ << pool.absUnits_.size() << ' ';
    for (std::map<std::string, absUnitPool::AbsUnitPtr>::const_iterator
         strAndAbsUnitPtr = pool.absUnits_.begin();
         strAndAbsUnitPtr != pool.absUnits_.end();
         ++strAndAbsUnitPtr)
    {
      const absUnitPool::AbsUnit& unitPtr = *strAndAbsUnitPtr->second;
      *result_ << stl::checkedGet(absUnitIDs_, &unitPtr) << ' ';
    }
    *result_ << stl::checkedGet(absUnitIDs_, 
                               pool.defaultUnit_.get()) << '\n';
  }
  *result_ << '\n';
}

void Saver::saveDBMemberIDs(const StrPoolPtr& strPool,
                            const AbsUnitPoolPtr& absUnitPool)
{
  *result_ << "Phase 5: save DB members\n";
  *result_ << stl::checkedGet(strPoolIDs_, strPool.get()) << ' '
          << stl::checkedGet(absUnitPoolIDs_, absUnitPool.get()) << '\n';
}

void Saver::saveCollectionIDs(const collections::Collections& collections)
{
  *result_ << "Phase 6.1: save collection IDs\n";
  *result_ << collections.size() << '\n';
  for (collections::Collections::const_iterator collection =
       collections.begin();
       collection != collections.end();
       ++collection)
  {
    const std::string& collName = collection->first;
    *result_ << collName.size() << ' ' << collName << '\n';
    const std::list<food::Food>& foods = collection->second;
    *result_ << foods.size() << '\n';
    for (std::list<food::Food>::const_iterator food = foods.begin();
         food != foods.end(); ++food)
    {
      *result_ << stl::checkedGet(foodIDs_, &(*food)) << '\n';
    }
  }
}

void Saver::saveIngredientIDs()
{
  *result_ << "Phase 7.1: save ingredient IDs\n";
  *result_ << ingredientVecIDs_.size() << '\n';
  for (std::map<const std::vector<food::Food>*, ULong>::const_iterator
       foodVecPtrAndID = ingredientVecIDs_.begin();
       foodVecPtrAndID != ingredientVecIDs_.end();
       ++foodVecPtrAndID)
  {
    const std::vector<food::Food>& ingredients = *(foodVecPtrAndID->first);
    ULong vecID = foodVecPtrAndID->second;
    *result_ << vecID << ' ' << ingredients.size() << ' ';
    for (std::vector<food::Food>::const_iterator ingredient =
         ingredients.begin();
         ingredient != ingredients.end();
         ++ingredient)
    {
      *result_ << stl::checkedGet(foodIDs_, &(*ingredient)) << ' ';
    }
    *result_ << '\n'; 
  }
}

void Saver::saveAllFoodIDs()
{
  *result_ << "Phase 9.1: save food IDs\n";
  *result_ << foodIDs_.size() << '\n';
  for (std::map<const food::Food*, ULong>::const_iterator foodPtrAndID =
       foodIDs_.begin();
       foodPtrAndID != foodIDs_.end();
       ++foodPtrAndID)
  {
    *result_ << foodPtrAndID->second << ' ';
    saveFoodIDs(*(foodPtrAndID->first));
    *result_ << '\n';
  }
}

void Saver::saveFoodIDs(const food::Food& food)
{
  using stl::checkedGet;

  *result_
    << checkedGet(strPoolIDs_,        food.strPool_.get())      << ' '
    << checkedGet(absUnitPoolIDs_,    food.absUnitPool_.get())  << ' '
    << checkedGet(numMapIDs_,         food.nuts_.get())         << ' '
    << checkedGet(numMapIDs_,         food.arbUnits_.get())     << ' '
    << checkedGet(ingredientVecIDs_,  food.ingredients_.get())  << ' '
    << checkedGet(stringIDs_,         food.name_.get())         << ' '
    << checkedGet(stringIDs_,         food.group_.get())        << ' '
    << checkedGet(stringIDs_,         food.date_.get())         << ' '
    << checkedGet(stringIDs_,         food.meal_.get())         << ' '
    << checkedGet(stringIDs_,         food.order_.get())        << ' '
    << checkedGet(stringIDs_,         food.comment_.get())      << ' '
    << checkedGet(stringIDs_,         food.instructions_.get()) << ' '
    << checkedGet(stringIDs_,         food.refDesc_.get())      << ' '
    << checkedGet(strNumberIDs_,      food.qty_.get())          << ' '
    << checkedGet(pctIDs_,            food.pctRefuse_.get())    << ' '
    << checkedGet(numberIDs_,         food.yield_.get())        << ' ';

  saveUnitIDs(food);
}

void Saver::saveUnitIDs(const food::Food& food)
{
  const unsigned int ABSOLUTE  = 0;
  const unsigned int ARBITRARY = 1;
  if (food.unitIsAbsolute_)
  {
    *result_ << ABSOLUTE << ' ' 
            << stl::checkedGet(absUnitIDs_, food.absUnit_.get());
  }
  else
  {
    *result_ << ARBITRARY << ' '
            << stl::checkedGet(stringIDs_, food.arbUnit_->first.get());
  }
}

void Loader::tossHeader()
{
  while (in_->peek() == ' ' || in_->peek() == '\n')
    in_->ignore();
  std::string header;
  std::getline(*in_, header);
}

void Loader::loadString(std::string& string)
{
  ULong size;
  *in_ >> size;
  in_->ignore();
  string.reserve(size);
  for (ULong index=0; index < size; ++index)
  {
    char c;
    in_->get(c);
    string.append(1, c);
  }
}

db::Db Loader::load(std::istream& data)
{
  db::Db result;
  in_ = &data;
  strPool_ = &result.strPool_;
  absUnitPool_ = &result.absUnitPool_;
  collections_ = &result.collections_;
  unsaved_ = &result.changed_;
  data.exceptions(std::istream::eofbit |
                  std::istream::failbit |
                  std::istream::badbit);
  
  clear();

  loadStringData();
  loadAbsUnitData();
  loadNumMapData();
  loadStrNumberData();
  loadPctData();
  loadNumberData();
  
  loadStrPools();
  loadAbsUnitPools();
  
  loadDBMembers();
  
  loadCollectionIDs();
  
  loadIngredientIDs();
  
  createFoods();
  
  loadAllFoodIDs();
  return result;
}

void Loader::clear()
{
  ingredients_.clear();
  collectionFoodIDs_.clear();
  ingredientFoodIDs_.clear();
  allFoods_.clear();
  strPools_.clear();
  absUnitPools_.clear();
  stringData_.clear();
  absUnitData_.clear();
  numberMapData_.clear();
  strNumberData_.clear();
  pctData_.clear();
  numberData_.clear();
}

void Loader::loadStringData()
{
  tossHeader();

  ULong nStrings;
  *in_ >> nStrings;
  stringData_.resize(nStrings);

  for (ULong stringIndex = 0; stringIndex < nStrings; ++stringIndex)
  {
    ULong id;
    *in_ >> id;
    std::string string;
    loadString(string);
    stl::checkedAt(stringData_, id).reset(new std::string(string));
  }
}

void Loader::loadAbsUnitData()
{
  tossHeader();

  ULong nAbsUnits;
  *in_ >> nAbsUnits;
  absUnitData_.resize(nAbsUnits);

  for (ULong i=0; i < nAbsUnits; ++i)
  {
    ULong id;
    *in_ >> id;
    absUnitPool::AbsUnit absUnit;
    loadString(absUnit.name);
    *in_ >> absUnit.value.value_;
    stl::checkedAt(absUnitData_, id).reset(new absUnitPool::AbsUnit(absUnit));
  }
}

void Loader::loadNumMapData()
{
  tossHeader();

  ULong nMaps;
  *in_ >> nMaps;
  numberMapData_.resize(nMaps);

  for (unsigned long int i=0; i < nMaps; ++i)
  {
    ULong mapID;
    *in_ >> mapID;
    stl::checkedAt(numberMapData_, mapID).reset(new StrPtrNumMap);
    StrPtrNumMap& currMap = *stl::checkedAt(numberMapData_, mapID);

    ULong mapSize;
    *in_ >> mapSize;
    for (ULong j=0; j < mapSize; ++j)
    {
      ULong stringID;
      *in_ >> stringID;
      pointers::Shared<const std::string> strPtr = stl::checkedAt(stringData_, stringID);
      numbers::Number number;
      *in_ >> number.value_;
      currMap.insert(std::make_pair(strPtr, number));
    }
  }
}

void Loader::loadStrNumberData()
{
  tossHeader();

  ULong nNumbers;
  *in_ >> nNumbers;
  strNumberData_.resize(nNumbers);

  for (ULong i=0; i < nNumbers; ++i)
  {
    ULong id;
    *in_ >> id;

    numbers::StrNumber strNumber;
    strNumber.string_.clear();
    loadString(strNumber.string_);
    stl::checkedAt(strNumberData_, id).reset(new 
        numbers::StrNumber(strNumber));
  }
}

void Loader::loadPctData()
{
  tossHeader();

  ULong nNumbers;
  *in_ >> nNumbers;
  pctData_.resize(nNumbers);

  for (ULong i=0; i < nNumbers; ++i)
  {
    ULong id;
    *in_ >> id;

    numbers::Percent percent;
    *in_ >> percent.value_;
    stl::checkedAt(pctData_, id).reset(new numbers::Percent(percent));
  }
}

void Loader::loadNumberData()
{
  tossHeader();

  ULong nNumbers;
  *in_ >> nNumbers;
  numberData_.resize(nNumbers);

  for (ULong i=0; i < nNumbers; ++i)
  {
    ULong id;
    *in_ >> id;

    numbers::Number number;
    *in_ >> number.value_;
    stl::checkedAt(numberData_, id).reset(new numbers::Number(number));
  }
}

void Loader::loadStrPools()
{
  tossHeader();

  ULong nPools;
  *in_ >> nPools;
  strPools_.resize(nPools);

  for (ULong i=0; i < nPools; ++i)
  {
    ULong id;
    *in_ >> id;
    stl::checkedAt(strPools_, id).reset(new strPool::StrPool);
    strPool::StrPool& currPool = *stl::checkedAt(strPools_, id);
    ULong nStrings;
    *in_ >> nStrings;
    for (ULong j=0; j < nStrings; ++j)
    {
      ULong strID;
      *in_ >> strID;
      const pointers::Shared<const std::string>& strPtr =
        stl::checkedAt(stringData_, strID);
      currPool.stringPtrMap_.insert(std::make_pair(*strPtr, strPtr));
    }
  }
}

void Loader::loadAbsUnitPools()
{
  tossHeader();

  ULong nPools;
  *in_ >> nPools;
  absUnitPools_.resize(nPools);

  for (ULong i=0; i < nPools; ++i)
  {
    ULong id;
    *in_ >> id;
    stl::checkedAt(absUnitPools_, id).reset(new absUnitPool::AbsUnitPool);
    absUnitPool::AbsUnitPool& currPool = *stl::checkedAt(absUnitPools_, id);
    ULong nUnits;
    *in_ >> nUnits;

    for (ULong j=0; j < nUnits; ++j)
    {
      ULong unitID;
      *in_ >> unitID;
      const pointers::Shared<const absUnitPool::AbsUnit>& absUnit =
        stl::checkedAt(absUnitData_, unitID);
      currPool.absUnits_.insert(std::make_pair(absUnit->name, absUnit));
    }
    
    ULong defaultUnitID;
    *in_ >> defaultUnitID;
    currPool.defaultUnit_ = stl::checkedAt(absUnitData_, defaultUnitID);
  }
}

void Loader::loadDBMembers()
{
  tossHeader();

  ULong strPoolID;
  *in_ >> strPoolID;
  *strPool_ = stl::checkedAt(strPools_, strPoolID);

  ULong absUnitPoolID;
  *in_ >> absUnitPoolID;
  *absUnitPool_ = stl::checkedAt(absUnitPools_, absUnitPoolID);
}

void Loader::loadCollectionIDs()
{
  tossHeader();

  ULong nCollections;
  *in_ >> nCollections;
  for (ULong collNo = 0; collNo < nCollections; ++collNo)
  {
    std::string collName;
    loadString(collName);
    (*unsaved_)[collName] = false;
    ULong nFoods;
    *in_ >> nFoods;
    std::pair<std::map<std::string, std::vector<ULong> >::iterator, bool>
      insertResult =
      collectionFoodIDs_.insert(std::make_pair(collName,
                                               std::vector<ULong>()));
    std::vector<ULong>& foodIDVec = insertResult.first->second;
    foodIDVec.reserve(nFoods);
    for (ULong foodNo = 0; foodNo < nFoods; ++foodNo)
    {
      ULong foodID;
      *in_ >> foodID;
      foodIDVec.push_back(foodID);
    }
  }
}

void Loader::loadIngredientIDs()
{
  tossHeader();
  ULong nIngrVecs;
  *in_ >> nIngrVecs;
  ingredientFoodIDs_.resize(nIngrVecs);
  for (ULong ingrVecIndex = 0; ingrVecIndex < nIngrVecs; ++ingrVecIndex)
  {
    ULong ingrVecID;
    *in_ >> ingrVecID;
    std::vector<ULong>& ingrVec = stl::checkedAt(ingredientFoodIDs_,
                                                 ingrVecID);
    ULong nIngredients;
    *in_ >> nIngredients;
    ingrVec.reserve(nIngredients);
    for (ULong ingrIndex = 0; ingrIndex < nIngredients; ++ingrIndex)
    {
      ULong foodID;
      *in_ >> foodID;
      ingrVec.push_back(foodID);
    }
  }
}

void Loader::createFoods()
{
  for (std::map<std::string, std::vector<ULong> >::const_iterator
       stringAndVec = collectionFoodIDs_.begin();
       stringAndVec != collectionFoodIDs_.end();
       ++stringAndVec)
  {
    const std::vector<ULong>& foodIDVec = stringAndVec->second;
    const std::string& collName = stringAndVec->first;
    std::pair<collections::Collections::iterator, bool> insertCollResult = 
      collections_->insert(std::make_pair(collName,
                                         std::list<food::Food>()));
    std::list<food::Food>& collList = insertCollResult.first->second;

    for (ULong foodIndex = 0; foodIndex < foodIDVec.size(); ++foodIndex)
    {
      collList.push_back(food::Food());
      food::Food& currFood = collList.back();
      currFood.recalcIngrNuts_.reset(new bool(true));
      currFood.ingrNuts_.reset(new std::map<std::string, numbers::Number>);
      unsigned long int foodID = stl::checkedAt(foodIDVec, foodIndex);
      if (allFoods_.size() < foodID +1)
        allFoods_.resize(foodID + 1);
      stl::checkedAt(allFoods_, foodID) = &collList.back();
    }
  }

  for (ULong i=0; i < ingredientFoodIDs_.size(); ++i)
  {
    pointers::Shared<std::vector<food::Food> >ingredientsPtr
                                              (new std::vector<food::Food>);
    ingredients_.push_back(ingredientsPtr);
  }

  for (ULong ingrCollID = 0; ingrCollID < ingredientFoodIDs_.size();
       ++ingrCollID)
  {
    std::vector<food::Food>& ingrVec =
      *(stl::checkedAt(ingredients_, ingrCollID));
    std::vector<ULong>& ingrIDVec =
      stl::checkedAt(ingredientFoodIDs_, ingrCollID);
    ingrVec.reserve(ingrIDVec.size());
    for (std::vector<ULong>::const_iterator ingrID = ingrIDVec.begin();
         ingrID != ingrIDVec.end(); ++ingrID)
    {
      ingrVec.push_back(food::Food());
      food::Food& currFood = ingrVec.back();
      currFood.recalcIngrNuts_.reset(new bool(true));
      currFood.ingrNuts_.reset(new std::map<std::string, numbers::Number>);
      if (allFoods_.size() < *ingrID + 1)
        allFoods_.resize(*ingrID + 1);
      stl::checkedAt(allFoods_, *ingrID) = &ingrVec.back();
    }
  }
}

void Loader::loadAllFoodIDs()
{
  const unsigned int ABSOLUTE  = 0;
  tossHeader();

  ULong nFoods;
  *in_ >> nFoods;
  for (ULong foodIndex = 0; foodIndex < nFoods; ++foodIndex)
  {
    ULong id;
    *in_ >> id;
    food::Food& food = *stl::checkedAt(allFoods_, id);
    
    *in_ >> id;
    food.strPool_ = stl::checkedAt(strPools_, id);

    *in_ >> id;
    food.absUnitPool_ = stl::checkedAt(absUnitPools_, id);

    *in_ >> id;
    food.nuts_ = stl::checkedAt(numberMapData_, id);

    *in_ >> id;
    food.arbUnits_ = stl::checkedAt(numberMapData_, id);

    *in_ >> id;
    food.ingredients_ = stl::checkedAt(ingredients_, id);

    *in_ >> id;
    food.name_ = stl::checkedAt(stringData_, id);

    *in_ >> id;
    food.group_ = stl::checkedAt(stringData_, id);

    *in_ >> id;
    food.date_ = stl::checkedAt(stringData_, id);

    *in_ >> id;
    food.meal_ = stl::checkedAt(stringData_, id);

    *in_ >> id;
    food.order_ = stl::checkedAt(stringData_, id);

    *in_ >> id;
    food.comment_ = stl::checkedAt(stringData_, id);

    *in_ >> id;
    food.instructions_ = stl::checkedAt(stringData_, id);

    *in_ >> id;
    food.refDesc_ = stl::checkedAt(stringData_, id);

    *in_ >> id;
    food.qty_ = stl::checkedAt(strNumberData_, id);

    *in_ >> id;
    food.pctRefuse_ = stl::checkedAt(pctData_, id);

    *in_ >> id;
    food.yield_ = stl::checkedAt(numberData_, id);

    unsigned int unitType;
    *in_ >> unitType;
    *in_ >> id;
    if (unitType == ABSOLUTE)
    {
      food.unitIsAbsolute_ = true;
      food.absUnit_ = stl::checkedAt(absUnitData_, id);
    }
    else
    {
      food.unitIsAbsolute_ = false;
      food.arbUnit_ = food.arbUnits_->find(stl::checkedAt(stringData_, id));
    }
  }
}

SaveWriteData::SaveWriteData(const std::string& filename)
    : filename_(filename) {}

std::string SaveWriteData::getBriefMessage() const
{
    std::string result(filename_);
    result.append(": error while writing data to save file");
    return result;
}

UnrecognizedFile::UnrecognizedFile(const std::string& path)
    : path_(path) {}

std::string UnrecognizedFile::getBriefMessage() const
{
    std::string result = path_;
    result.append(": file format not recognized");
    return result;
}

LoadReadFile::LoadReadFile(const std::string& filename)
    : filename_(filename) {}

std::string LoadReadFile::getBriefMessage() const
{
    std::string result(filename_);
    result.append(": error opening file for reading");
    return result;
}

} // namespace serializer
