#ifndef SERIALIZER_H
#define SERIALIZER_H

#include "absUnitPool.h"
#include "collections.h"
#include "db.h"
#include "errors.h"
#include "food.h"
#include "numbers.h"
#include "pointers.h"
#include "strPool.h"

#include <istream>
#include <map>
#include <ostream>
#include <string>
#include <vector>

namespace serializer
{

/** Save a databse.
 *
 * Offers strong guarantee; however, if this throws, the
 * state of what is on disk will be unpredictable.
 */
void save(const db::Db& database, const std::string& filename);

/** Load a database.
 *
 * Offers strong guarantee.
 */
db::Db load(const std::string& filename);

typedef pointers::Shared<strPool::StrPool> StrPoolPtr;
typedef pointers::Shared<absUnitPool::AbsUnitPool> AbsUnitPoolPtr;

class Saver
{
public:

  /** Save a database to a stream.
   *
   * \param os Stream to which to send the saved data.
   *
   * This function is not exception safe. If it throws, no
   * resources will be leaked; however, it is impossible to
   * know the state of os after a throw.
   */
  void save(const db::Db& db, std::ostream& os);

private:
  typedef unsigned long int ULong;
  typedef std::map<strPool::StrPtr, numbers::Number> StrPtrNumMap;

  std::ostream* result_;

  std::map<const food::Food*, ULong> foodIDs_;
  std::map<const std::string*, ULong> stringIDs_;
  std::map<const strPool::StrPool*, ULong> strPoolIDs_;
  std::map<const absUnitPool::AbsUnitPool*, ULong> absUnitPoolIDs_;
  std::map<const StrPtrNumMap*, ULong> numMapIDs_;
  std::map<const std::vector<food::Food>*, ULong> ingredientVecIDs_;
  std::map<const numbers::StrNumber*, ULong> strNumberIDs_;
  std::map<const numbers::Percent*, ULong> pctIDs_;
  std::map<const numbers::Number*, ULong> numberIDs_;
  std::map<const absUnitPool::AbsUnit*, ULong> absUnitIDs_;

  // Phase 1: clear out old data members
  void clear();

  // Phase 2: create identifiers
  void createDBMemberIDs(const StrPoolPtr& strPool,
                        const AbsUnitPoolPtr& absUnitPool);
  void createFoodIDs(const collections::Collections& collections);
  void createFoodID(const food::Food& food);
  void createComponentIDs(const food::Food& food);
  void createStringIDs();
  void createAbsUnitIDs();

  // Phase 3: save all data
  void saveStringData();
  void saveAbsUnitData();
  void saveNumMapData();
  void saveStrNumberData();
  void savePctData();
  void saveNumberData();

  // Phase 4: save pool IDs. Saves IDs for every pool, and
  // saves IDs to all internal data for every pool.
  void saveStrPoolIDs();
  void saveAbsUnitPoolIDs();
  
  // Phase 5: save DB members (the strPool and
  // absUnitPool args to the save function)
  void saveDBMemberIDs(const StrPoolPtr& strPool,
                       const AbsUnitPoolPtr& absUnitPool);

  // Phase 6: Save collection IDs. Records ID for each food
  // that is stored in the collection itself.
  void saveCollectionIDs(const collections::Collections& collections);

  // Phase 7: save ingredient IDs. Records ID for each
  // ingredient vector and ID for every food stored in each
  // vector.
  void saveIngredientIDs();

  // (there is no Phase 8 in saving)
  
  // Phase 9: save food IDs. Saves IDs for every food, and
  // saves IDs to all internal data for every food.
  void saveAllFoodIDs();
  void saveFoodIDs(const food::Food& food);
  void saveUnitIDs(const food::Food& food);

};

class Loader
{
public:
  
  /** Load a database from a stream.
   *
   * \param strPool load will reset this shared pointer to
   * point to the loaded database's default string pool.
   *
   * \param absUnitPool load will reset this shared pointer
   * to point to the loaded database's absolute unit pool.
   *
   * \param collections load will clear this variable, and
   * then fill it with the loaded database's collections.
   *
   * \param data Stream to load data from.
   *
   * Offers strong guarantee.
   */
  db::Db load(std::istream& data);

private:
  typedef unsigned long int ULong;
  typedef std::map<strPool::StrPtr, numbers::Number> StrPtrNumMap;

  std::istream* in_;
  StrPoolPtr* strPool_;
  AbsUnitPoolPtr* absUnitPool_;
  collections::Collections* collections_;
  collections::Unsaved* unsaved_;

  std::vector<pointers::Shared<std::vector<food::Food> > > ingredients_;

  std::map<std::string, std::vector<ULong> > collectionFoodIDs_;
  std::vector<std::vector<ULong> > ingredientFoodIDs_;

  std::vector<food::Food*> allFoods_;

  std::vector<pointers::Shared<strPool::StrPool> > strPools_;
  std::vector<pointers::Shared<absUnitPool::AbsUnitPool> > absUnitPools_;

  std::vector<pointers::Shared<const std::string> > stringData_;
  std::vector<pointers::Shared<const absUnitPool::AbsUnit> > absUnitData_;
  std::vector<pointers::Shared<StrPtrNumMap> > numberMapData_;
  std::vector<pointers::Shared<numbers::StrNumber> > strNumberData_;
  std::vector<pointers::Shared<numbers::Percent> > pctData_;
  std::vector<pointers::Shared<numbers::Number> > numberData_;

  void loadString(std::string& string);
  
  /** Read in and discard a header line, discarding excess
   * whitespace
   */
  void tossHeader();

  // Phase 1: clear out old data structures
  void clear();

  // (there is no Phase 2 in loading)

  // Phase 3: load all data
  void loadStringData();
  void loadAbsUnitData();
  void loadNumMapData();
  void loadStrNumberData();
  void loadPctData();
  void loadNumberData();

  // Phase 4: load pools
  void loadStrPools();
  void loadAbsUnitPools();

  // Phase 5: load DB members
  void loadDBMembers();

  // Phase 6: load collection IDs
  void loadCollectionIDs();

  // Phase 7:load ingredient IDs
  void loadIngredientIDs();

  // Phase 8: create foods
  void createFoods();

  // Phase 9: load food IDs
  void loadAllFoodIDs();
};

/** Error while writing data to a save file.
 */
class SaveWriteData : public errors::NonFatal
{
public:
    SaveWriteData(const std::string& filename);
    std::string getBriefMessage() const;

private:
    std::string filename_;
};

/** Attempted to load a file that does not have the Pantry file
 * header.
 */
class UnrecognizedFile : public errors::NonFatal
{
public:
    UnrecognizedFile(const std::string& path);

private:
    std::string getBriefMessage() const;
    std::string path_;
};

/** Failure while reading data from a file opened for loading.
 */
class LoadReadFile : public errors::NonFatal
{
public:
    LoadReadFile(const std::string& filename);
    std::string getBriefMessage() const;

private:
    std::string filename_;
};

} // namespace serializer

#endif
