#include "comm.h"
#include "errors.h"
#include "dbholder.h"
#include "parser.h"
#include "stl.h"

#include <iostream>
#include <memory>
#include <string>
#include <vector>

#include <sys/stat.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

void doFork()
{
  pid_t pid = fork();
  if (pid == -1)
  {
    perror("pantryd: error: fork failed");
    exit(EXIT_FAILURE);
  }
  else if (pid != 0)
    exit(EXIT_SUCCESS);
}

void daemonize()
{
  // see topic 1.7 in
  // http://www.faqs.org/faqs/unix-faq/programmer/faq/
  doFork();
  setsid();
  doFork();
  
  int err = chdir("/");
  if (err == -1)
  {
    perror("pantryd: error: could not change to root directory");
    exit(EXIT_FAILURE);
  }

  umask(0); 
  freopen("/dev/null", "r", stdin);
  freopen("/dev/null", "w", stdout);
  freopen("/dev/null", "w", stderr);
}

void showHelp()
{
  std::cout <<
    "usage: pantryd [options]\n"
    "Options:\n"
    "  -h, --help Show this help and exit\n"
    "  -f, --foreground Run in foreground (useful for debugging)\n";
  exit(EXIT_SUCCESS);
}

bool processOpts(int argc, char** argv)
{
  const std::string FOREGROUND = "foreground";
  const std::string HELP = "help";
  parser::Parser parser;
  parser.addOption(FOREGROUND, 'f', 0);
  parser.addOption(HELP, 'h', 0);
  std::vector<std::string> args(argv, argv + argc);

  parser::Result result;
  try
  {
    result = parser.parse(args);
  }
  catch (errors::NonFatal& e)
  {
    std::cerr << "pantryd: error: " << e.getBriefMessage() << '\n';
    exit(EXIT_FAILURE);
  }

  if (! result.posArgs.empty())
  {
    std::cerr << "pantryd: error: unrecognized option: "
              << result.posArgs.front() << '\n';
    exit(EXIT_FAILURE);
  }
  if (parser::isSet(stl::checkedGet(result.optArgs, HELP)))
    showHelp();
  if (! parser::isSet(stl::checkedGet(result.optArgs, FOREGROUND)))
    return true;
  return false;
}

int main(int argc, char** argv)
{
  bool background = processOpts(argc, argv);
  try
  {
    comm::NamedSocket namedSocket;
    if (background)
      daemonize();
    dbHolder::DbHolder holder;

    bool done = false;
    while (! done)
    {
      comm::Bottle bottle(namedSocket);
      std::auto_ptr<comm::Response> response =
        holder.processRequest(bottle.getRequest());
      done = response->getQuit();
      bottle.setResponse(response);
    }
  } 
  catch (errors::Fatal& e)
  {
      std::cerr << "pantryd: error: "
                << e.getBriefMessage() << '\n';
      return 1;
  }
  return 0;
}
