#include "sorter.h"

#include "errors.h"
#include "food.h"
#include "stl.h"
#include "strings.h"

#include <cctype>
#include <map>
#include <string>
#include <utility>
#include <vector>

namespace sorter
{

Key::Key(const std::string& keyString)
{
  traitID_ = getTraitID(keyString);
  direction_ = getDirection(keyString);
}

Key::Key(TraitIDType trait, DirectionType direction)
  : traitID_(trait), direction_(direction) {}

DirectionType Key::getDirection() const
{
  return direction_;
}

TraitIDType Key::getTraitID() const
{
  return traitID_;
}

TraitIDType Key::getTraitID(const std::string& keyString) const
{
  std::map<std::string, TraitIDType> stringTraitMap;
  using std::make_pair;
  stringTraitMap.insert(make_pair(food::NAME, T_NAME));
  stringTraitMap.insert(make_pair(food::GROUP, T_GROUP));
  stringTraitMap.insert(make_pair(food::DATE, T_DATE));
  stringTraitMap.insert(make_pair(food::MEAL, T_MEAL));
  stringTraitMap.insert(make_pair(food::COMMENT, T_COMMENT));
  stringTraitMap.insert(make_pair(food::ORDER, T_ORDER));
  stringTraitMap.insert(make_pair(food::REFDESC, T_REFDESC));
  stringTraitMap.insert(make_pair(food::INSTRUCTIONS, T_INSTRUCTIONS));
  stringTraitMap.insert(make_pair(food::QUANTITY, T_QTY));
  stringTraitMap.insert(make_pair(food::PCTREFUSE, T_PCTREFUSE));
  stringTraitMap.insert(make_pair(food::YIELD, T_YIELD));
  stringTraitMap.insert(make_pair(food::UNIT, T_UNIT));

  std::string lowerCaseKeyString = strings::firstLetterToLower(keyString);
  std::vector<std::string> matches =
    strings::matchAbbrev(stl::getKeys(stringTraitMap), lowerCaseKeyString);
  if (matches.size() != 1)
    throw BadKeyName(keyString, matches);
  return stl::checkedGet(stringTraitMap, matches[0]);
}

DirectionType Key::getDirection(const std::string& keyString) const
{
  if (keyString.empty())
    throw errors::Logic();
  if (std::islower(static_cast<int>(keyString[0])) != 0)
    return D_ASCENDING;
  return D_DESCENDING;
}

CompareTraits::CompareTraits() {}

CompareTraits::CompareTraits(const StringPairs& sortOrders)
{
  std::map<std::string, std::map<std::string, 
                                 unsigned int>* > traitMap;
  using std::make_pair;
  traitMap.insert(make_pair(food::NAME,         &this->name));
  traitMap.insert(make_pair(food::GROUP,        &this->group));
  traitMap.insert(make_pair(food::DATE,         &this->date));
  traitMap.insert(make_pair(food::MEAL,         &this->meal));
  traitMap.insert(make_pair(food::COMMENT,      &this->comment));
  traitMap.insert(make_pair(food::ORDER,        &this->order));
  traitMap.insert(make_pair(food::REFDESC,      &this->refDesc));
  traitMap.insert(make_pair(food::INSTRUCTIONS, &this->instructions));
  traitMap.insert(make_pair(food::QUANTITY,     &this->qty));
  traitMap.insert(make_pair(food::PCTREFUSE,    &this->pctRefuse));
  traitMap.insert(make_pair(food::YIELD,        &this->yield));
  traitMap.insert(make_pair(food::UNIT,         &this->unit));

  std::vector<std::string> traitNames = stl::getKeys(traitMap);
  for (StringPairs::const_iterator stringPair = sortOrders.begin();
          stringPair != sortOrders.end(); ++stringPair)
  {
    std::vector<std::string> matches =
        strings::matchAbbrev(traitNames, stringPair->first);
    if (matches.size() != 1)
        throw SorterBadTraitAbbreviation(matches.size(),
                                         stringPair->first);
    const std::string& traitName = matches[0];
    std::map<std::string, unsigned int>* destMap =
      stl::checkedGet(traitMap, traitName);
    unsigned int newIndex = destMap->size();
    destMap->insert(std::make_pair(stringPair->second, newIndex));
  }
}

std::string Keys::getTraitValue(TraitIDType trait,
                                const food::Food& food) const
{
  if (trait == T_NAME)
    return food.getName();
  else if (trait == T_GROUP)
    return food.getGroup();
  else if (trait == T_DATE)
    return food.getDate();
  else if (trait == T_MEAL)
    return food.getMeal();
  else if (trait == T_ORDER)
    return food.getOrder();
  else if (trait == T_COMMENT)
    return food.getComment();
  else if (trait == T_INSTRUCTIONS)
    return food.getInstructions();
  else if (trait == T_REFDESC)
    return food.getRefDesc();
  else if (trait == T_QTY)
    return food.getQty().getString();
  else if (trait == T_PCTREFUSE)
    return food.getPctRefuse().getString();
  else if (trait == T_YIELD)
    return food.getYield().getString();
  else
    return food.getCurrentUnitName();
}

const CompareTraits::SortMapType&
CompareTraits::getSortMap(TraitIDType traitID) const
{
  if (traitID == T_NAME)
    return this->name;
  else if (traitID == T_GROUP)
    return this->group;
  else if (traitID == T_DATE)
    return this->date;
  else if (traitID == T_MEAL)
    return this->meal;
  else if (traitID == T_COMMENT)
    return this->comment;
  else if (traitID == T_ORDER)
    return this->order;
  else if (traitID == T_REFDESC)
    return this->refDesc;
  else if (traitID == T_INSTRUCTIONS)
    return this->instructions;
  else if (traitID == T_QTY)
    return this->qty;
  else if (traitID == T_PCTREFUSE)
    return this->pctRefuse;
  else if (traitID == T_YIELD)
    return this->yield;
  else
    return this->unit;
}

bool CompareTraits::comp(const Key& key,
                         const std::string& xStr,
                         const std::string& yStr) const
{
  const SortMapType& sortMap = getSortMap(key.getTraitID());
  SortMapType::const_iterator xIndex = sortMap.find(xStr);
  SortMapType::const_iterator yIndex = sortMap.find(yStr);
  bool xInSortMap = (xIndex != sortMap.end());
  bool yInSortMap = (yIndex != sortMap.end());

  bool xIsLessThanY;
  if (xInSortMap && yInSortMap)
    xIsLessThanY = (xIndex->second < yIndex->second);

  else if (xInSortMap && (! yInSortMap))
    xIsLessThanY = true;

  else if ((! xInSortMap) && yInSortMap)
    xIsLessThanY = false;

  else
    xIsLessThanY = (xStr < yStr);

  if (key.getDirection() == D_DESCENDING)
    xIsLessThanY = (! xIsLessThanY);
  return xIsLessThanY;
}

Keys::Keys(const std::vector<std::string>& keyStrings)
{
  for (std::vector<std::string>::const_iterator string = keyStrings.begin();
       string != keyStrings.end(); ++string)
  {
    keys_.push_back(Key(*string));
  }
}

Keys::Keys(const std::vector<Key>& keys)
  : keys_(keys) {}

bool Keys::comp(const food::Food& x, const food::Food& y,
                const CompareTraits& compareTraits) const
{
  for (std::vector<Key>::const_iterator key = keys_.begin();
       key != keys_.end(); ++key)
  {
    std::string xStr = getTraitValue(key->getTraitID(), x);
    std::string yStr = getTraitValue(key->getTraitID(), y);
    if (xStr != yStr)
      return compareTraits.comp(*key, xStr, yStr);
  }
  return false;
}

Sorter::Sorter(const Keys& keys, const CompareTraits& compareTraits)
  : keys_(keys), compareTraits_(compareTraits) {}

bool Sorter::operator() (const food::Food* x, const food::Food* y) const
{
  return keys_.comp(*x, *y, compareTraits_);
}

SorterBadTraitAbbreviation::SorterBadTraitAbbreviation(unsigned int nMatches,
                                                       const std::string& abbrev)
    : nMatches_(nMatches), abbrev_(abbrev) {}

std::string SorterBadTraitAbbreviation::getBriefMessage() const
{
  std::string result = "invalid trait abbreviation: ";
  result.append(abbrev_);
  if (nMatches_ == 0)
    result.append(": no");
  else
    result.append(": multiple");
  result.append(" matches found");
  return result;
}

BadKeyName::BadKeyName(const std::string& keyName,
           const std::vector<std::string>& matches)
  : keyName_(keyName), matches_(matches) {}

std::string BadKeyName::getBriefMessage() const
{
  std::string result = "bad abbreviation for sort key: ";
  result.append(keyName_);
  result.append(" has ");
  if (matches_.empty())
    result.append("no matches.");
  else
    result.append("more than one match.");
  return result;
}

} // namespace sorter
