#ifndef SORTER_H
#define SORTER_H

#include "errors.h"
#include "food.h"

#include <map>
#include <string>
#include <utility>
#include <vector>

namespace sorter {

typedef std::vector<std::pair<std::string, std::string> > StringPairs;

enum TraitIDType
{
  T_NAME,
  T_GROUP,
  T_DATE,
  T_MEAL,
  T_COMMENT,
  T_ORDER,
  T_REFDESC,
  T_INSTRUCTIONS,
  T_QTY,
  T_PCTREFUSE,
  T_YIELD,
  T_UNIT
};

enum DirectionType
{
  D_ASCENDING,
  D_DESCENDING
};

class Key
{
public:
  Key(const std::string& keyString);
  Key(TraitIDType trait, DirectionType direction);
  DirectionType getDirection() const;
  TraitIDType getTraitID() const;

private:
  TraitIDType getTraitID(const std::string& keyString) const;
  DirectionType getDirection(const std::string& keyString) const;

  TraitIDType           traitID_;
  DirectionType         direction_;
};

class CompareTraits
{
public:
  CompareTraits();
  CompareTraits(const StringPairs& sortOrders);
  bool comp(const Key& key,
            const std::string& xStr,
            const std::string& yStr) const;

private:
  typedef std::map<std::string, unsigned int> SortMapType;
  const SortMapType& getSortMap(TraitIDType traitID) const;

  SortMapType     name;
  SortMapType     group;
  SortMapType     date;
  SortMapType     meal;
  SortMapType     comment;
  SortMapType     order;
  SortMapType     refDesc;
  SortMapType     instructions;
  SortMapType     qty;
  SortMapType     pctRefuse;
  SortMapType     yield;
  SortMapType     unit;
};

class Keys
{
public:
  Keys(const std::vector<std::string>& keyStrings);
  Keys(const std::vector<Key>& keys);
  bool comp(const food::Food& x, const food::Food& y,
            const CompareTraits& compareTraits) const;

private:
  std::string getTraitValue(TraitIDType trait,
                            const food::Food& food) const;
  std::vector<Key> keys_;
};

class Sorter
{
public:
  Sorter(const Keys& keys,
         const CompareTraits& sortOrders);
  bool operator() (const food::Food* x, const food::Food* y) const;

private:
  Keys keys_;
  CompareTraits compareTraits_;
};

/** Invalid specifier for a sorter key.
 */
class BadKeyName : public errors::NonFatal
{
public:
  BadKeyName(const std::string& keyName,
             const std::vector<std::string>& matches);
  std::string getBriefMessage() const;

private:
  std::string keyName_;
  std::vector<std::string> matches_;
};

/** An incorrect trait abbreviation specified for a sort order.
 */
class SorterBadTraitAbbreviation : public errors::NonFatal
{
public:
  SorterBadTraitAbbreviation(unsigned int nMatches,
                                const std::string& abbrev);
  std::string getBriefMessage() const;

private:
  unsigned int nMatches_;
  std::string abbrev_;
};

} // namespace sorter

#endif
