#ifndef STL_H
#define STL_H

#include <algorithm>
#include <utility>
#include <vector>

#include "errors.h"

namespace stl
{

/** Takes a associative container and a sequence of keys; returns a
 * associative container containing the keys that are in the
 * key sequence.
 */

template <typename T, typename U>
T selectKeys(const T& container, const U& keys)
{
    T result;
    for (typename U::const_iterator it = keys.begin(); it < keys.end(); ++it)
    {
        std::pair<typename T::const_iterator, 
                  typename T::const_iterator> rangePair =
            container.equal_range(*it);
        result.insert(rangePair.first, rangePair.second);
    }
    return result;
}

/** Takes a associative container and a sequence of keys; returns a 
 * associative container containing the keys that are NOT in the
 * key sequence.
 */
template <typename T, typename U>
T deleteKeys(const T& container, const U& keys)
{
    T result(container);
    for (typename U::const_iterator key = keys.begin();
            key != keys.end(); ++key)
    {
        std::pair<typename T::iterator,
                  typename T::iterator> rangePair =
            result.equal_range(*key);
        result.erase(rangePair.first, rangePair.second);
    }
    return result;
}

/** Gets a vector of keys from a pair associative container.
 */
template <typename T>
std::vector<typename T::key_type> getKeys(const T& container)
{
  std::vector<typename T::key_type> result;
  result.reserve(container.size());
  for (typename T::const_iterator it = container.begin();
       it != container.end(); ++it)
  {
    result.push_back(it->first);
  }
  return result;
}

template <typename T>
const typename T::mapped_type& checkedGet(const T& map,
                                          const typename T::key_type& key)
{
  typename T::const_iterator findResult = map.find(key);
  if (findResult == map.end())
    throw errors::Logic();
  return findResult->second;
}

template <typename T>
typename T::mapped_type& checkedGet(T& map,
                                    const typename T::key_type& key)
{
  typename T::iterator findResult = map.find(key);
  if (findResult == map.end())
    throw errors::Logic();
  return findResult->second;
}

template <typename T>
const typename T::value_type& checkedAt(const T& sequence,
                                        typename T::size_type index)
{
  if (index >= sequence.size())
    throw errors::Logic();
  return sequence[index];
}

template <typename T>
typename T::value_type& checkedAt(T& sequence,
                                  const typename T::size_type index)
{
  if (index >= sequence.size())
    throw errors::Logic();
  return sequence[index];
}

} // namespace stl

#endif
