#ifndef STRPOOL_H
#define STRPOOL_H

#include "pointers.h"
#include "serializer_fwd.h"

#include <map>
#include <string>

namespace strPool
{

typedef pointers::Shared<const std::string> StrPtr;

/** A shared pool of strings.
 *
 * C++ does not have built-in string interning. Therefore,
 * if you use only the built-in string class, multiple copies of
 * the same string will each occupy their own space in
 * memory. (Though some string classes have copy-on-write,
 * not all do; besides, COW does not always solve this
 * problem.) When you have seven thousand copies of the same
 * string, this can add up a lot of memory usage.
 *
 * Thus the Food class uses this class to solve this
 * problem. The pool holds a bunch of shared pointers to
 * strings. Instead of the Food class holding C++ strings by
 * value, it instead holds shared pointers to strings that
 * are in the pool. Use of this class cut Pantry's memory
 * usage by almost half.
 */
class StrPool
{
public:

  /** Get a pointer to a string.
   *
   * \return A shared pointer to a string that is in the
   * pool. This pointer is empty if str is not already in
   * the pool.
   */

  StrPtr getStrPtr(const std::string& str);

  /** Add a string to the pool.
   *
   * If string was already in the pool, has no effect.
   *
   * \return a shared pointer to the new string in the pool.
   */
  StrPtr addToPool(const std::string& str);

  /** Removes a string from the pool, but only if that
   * string is not being used by any other StrPtr instances.
   *
   * Has no effect if the string was not in the pool.
   */
  void removeUnusedStr(const std::string& str);

  /** Determine whether a string is already in the pool.
   */
  bool inPool(const std::string& str) const;

  /** Removes all strings from the pool that are not in use.
   *
   * Has same affect as running removeUnusedStr for every
   * string in the pool.
   */
  void cleanup();

  friend class serializer::Loader;
  friend class serializer::Saver;

  /** Swap this StrPool with another.
   */
  void swap(StrPool& rhs);

private:
  std::map<std::string, StrPtr> stringPtrMap_;
};

} // namespace pool

#endif
