#include "strings.h"

#include "errors.h"

#include <algorithm>
#include <cctype>
#include <functional>
#include <string>
#include <vector>

namespace strings {

std::vector<std::string> explode (const std::string &inString,
        const std::string &separator)
{
  std::vector<std::string> returnVector;
  std::string::size_type start = 0;
  
  std::string::size_type end = inString.find(separator, start);
  while (end != std::string::npos)
  {
    returnVector.push_back(inString.substr(start, end-start));
    start = end + separator.size();
    end = inString.find(separator, start);
  }

  returnVector.push_back(inString.substr(start));
  return returnVector;
}

std::vector<std::string> explode (const std::string &inString,
        const std::string& separator, int maxSplits)
{
  std::vector<std::string> returnVector;
  std::string::size_type start = 0;

  int splits = 0;
  std::string::size_type end = inString.find(separator, start);
  while ((end != std::string::npos) && (splits <= maxSplits))
  {
    returnVector.push_back(inString.substr(start, end - start));
    start = end + separator.size();
    end = inString.find(separator, start);
    ++splits;
  }

  returnVector.push_back(inString.substr(start));
  return returnVector;
}

std::string ltrim(const std::string& x, const std::string &toTrim)
{
  size_t firstToKeep = x.find_first_not_of(toTrim);
  if (firstToKeep == std::string::npos)
    return std::string();
  return std::string(x, firstToKeep);
}

std::string rtrim(const std::string& x, const std::string &toTrim)
{
  size_t lastToKeep = x.find_last_not_of(toTrim);
  if (lastToKeep == std::string::npos)
    return std::string();
  return std::string(x, 0, lastToKeep + 1);
}

std::string trim(const std::string& x, const std::string &toTrim)
{
  std::string result = rtrim(x, toTrim);
  result = ltrim(result, toTrim);
  return result;
}

bool startswith(const std::string& subject, const std::string& test)
{
  return (subject.substr(0, test.length()) == test);
}

bool endswith(const std::string& subject, const std::string& test)
{
  return (subject.substr(subject.length() - test.length()) == test);
}

std::vector<std::string> matchAbbrev(const std::vector<std::string>& strings,
        const std::string& abbrev)
{
  std::vector<std::string> result;
  for (std::vector<std::string>::const_iterator string = strings.begin();
          string != strings.end(); ++string)
  {
    if (startswith(*string, abbrev))
      result.push_back(*string);
  }
  return result;
}

std::string replace(const std::string& subject,
                    const char toReplace,
                    const std::string& replacement)
{
  std::string result;
  for(std::string::const_iterator currChar = subject.begin();
      currChar != subject.end(); ++currChar)
  {
    if (*currChar == toReplace)
      result.append(replacement);
    else
      result.append(1, *currChar);
  }
  return result;
}

std::string capitalizeFirstLetter(const std::string& x)
{
  if (x.empty())
      return x;
  std::string result(x);
  result[0] = std::toupper(result[0]);
  return result;
}

std::string firstLetterToLower(const std::string& x)
{
  if (x.empty())
      return x;
  std::string result(x);
  result[0] = std::tolower(result[0]);
  return result;
}

std::string toLower(const std::string& x)
{
  std::string result;
  result.reserve(x.size());
  for (std::string::const_iterator currChar = x.begin();
          currChar != x.end(); ++currChar)
  {
    result.append(1, std::tolower(*currChar));
  }
  return result;
}

std::string pathJoin(const std::vector<std::string>& parts)
{
  std::string result;
  if (parts.empty())
      return result;
  result = parts[0];

  for (std::vector<std::string>::size_type i=1; i < parts.size(); ++i)
  {
    result.append("/");
    result.append(parts[i]);
  }
  return result;
}

std::string pathJoin(const std::string& part1, const std::string& part2)
{
  std::vector<std::string> parts;
  parts.push_back(part1);
  parts.push_back(part2);
  return pathJoin(parts);
}

std::string getAbsoluteFilename(const std::string& relativeTo,
                                const std::string& filename)
{
  if (relativeTo.empty())
    throw errors::Logic();
  if (filename[0] == '/')
    return filename;
  return pathJoin(relativeTo, filename);
}


} // namespace strings
