#include "totaler.h"

#include "food.h"
#include "numbers.h"

#include <algorithm>
#include <map>
#include <string>
#include <utility>
#include <vector>

namespace totaler
{

void Totaler::initTotals(const std::vector<const food::Food*>& foods)
{
  for (std::vector<const food::Food*>::const_iterator food =
       foods.begin(); food != foods.end(); ++food)
  {
    StringVec nutNames = (*food)->getNutNames();
    for (StringVec::const_iterator nutName = nutNames.begin();
         nutName != nutNames.end(); ++nutName)
    {
      numbers::Number& sum = totals_[*nutName];
      numbers::Number nutValue = (*food)->getNut(*nutName);
      double newSum = sum.getValue() + nutValue.getValue();
      sum = numbers::Number(newSum);
    }
  }
}

Goals getGoals(const StringPairs& goalStrings)
{
  Goals result;
  for (StringPairs::const_iterator stringPair = goalStrings.begin();
       stringPair != goalStrings.end(); ++stringPair)
  {
    result.push_back(std::make_pair(stringPair->first,
                                    numbers::Number(stringPair->second)));
  }
  return result;
}

Totaler::Totaler(const std::vector<const food::Food*>& foods,
                 const Goals& goals)
    : goals_(goals), 
      goalsMap_(goals_.begin(), goals_.end())
{
  initTotals(foods);
}

FoodGoal::FoodGoal(const NutMap::value_type& goal,
                   bool paramHasNut,
                   const numbers::Number& argNutAmt,
                   const NutMap& totals)
{
  nutName = goal.first;
  if (goal.second.isZero())
    goalStatus = G_ZEROGOAL;
  else
    goalStatus = G_GOAL;

  if (! paramHasNut)
    hasNut = false;
  else
  {
    hasNut = true;
    nutAmt = argNutAmt;
    pctGoal = nutAmt.getValue() / goal.second.getValue() * 100;

    NutMap::const_iterator totalNameAndNumber = totals.find(nutName);
    if ((totalNameAndNumber == totals.end())
            || (totalNameAndNumber->second.isZero()))
    {
      hasPctTotal = false;
    }
    else
    {
      hasPctTotal = true;
      pctTotal = nutAmt.getValue() /
        totalNameAndNumber->second.getValue() * 100;
    }
  }
}

std::vector<FoodGoal> Totaler::getFoodGoals(const food::Food& food) const
{
  std::vector<FoodGoal> result;
  std::vector<std::string> foodNutNames = food.getNutNames();
  std::sort(foodNutNames.begin(), foodNutNames.end());
  for (Goals::const_iterator goal = goals_.begin();
       goal != goals_.end(); ++goal)
  {
    numbers::Number nutAmt;
    bool hasNut = 
      std::binary_search(foodNutNames.begin(), foodNutNames.end(),
                         goal->first);
    if (hasNut)
      nutAmt = food.getNut(goal->first);
    result.push_back(FoodGoal(*goal, hasNut, nutAmt, totals_));
  }
  return result;
}

TotalGoal::TotalGoal(const NutMap::value_type& goal,
                     const NutMap& totals)
{
  nutName = goal.first;
  NutMap::const_iterator totalNameAndAmt = totals.find(goal.first);
  if (totalNameAndAmt == totals.end())
  {
    hasNut = false;
    return;
  }
  else
  {
    hasNut = true;
    nutAmt = totalNameAndAmt->second;
  }

  if (goal.second.isZero())
  {
    goalStatus = G_ZEROGOAL;
  }
  else
  {
    goalStatus = G_GOAL;
    pctGoal = totalNameAndAmt->second.getValue()
        / goal.second.getValue() * 100;
  }
}

std::vector<TotalGoal> Totaler::getTotalGoals() const
{
  std::vector<TotalGoal> result;
  for (Goals::const_iterator goal = goals_.begin();
       goal != goals_.end(); ++goal)
  {
    result.push_back(TotalGoal(*goal, totals_));
  }
  return result;
}

FoodAll::FoodAll(const std::string& argNutName,
                 const numbers::Number& argNutAmt,
                 const NutMap& goals,
                 const NutMap& totals)
{
  nutName = argNutName;
  nutAmt = argNutAmt;
  NutMap::const_iterator goal = goals.find(nutName);
  if (goal == goals.end())
    goalStatus = G_NOGOAL;
  else if (goal->second.isZero())
    goalStatus = G_ZEROGOAL;
  else
  {
    goalStatus = G_GOAL;
    pctGoal = nutAmt.getValue() / goal->second.getValue() * 100;
  }
  NutMap::const_iterator totalNameAndNumber = totals.find(nutName);
  if ((totalNameAndNumber == totals.end())
          || (totalNameAndNumber->second.isZero()))
  {
    hasPctTotal = false;
    return;
  }
  hasPctTotal = true;
  pctTotal = nutAmt.getValue()
      / totalNameAndNumber->second.getValue() * 100;
}

std::map<std::string, FoodAll> 
Totaler::getFoodAll(const food::Food& food) const
{
  std::vector<std::string> nutNames = food.getNutNames();
  std::map<std::string, FoodAll> result;
  for (StringVec::const_iterator nutName = nutNames.begin();
       nutName != nutNames.end(); ++nutName)
  {
    numbers::Number nutAmt = food.getNut(*nutName);
    result.insert(std::make_pair(*nutName, FoodAll(*nutName, nutAmt,
                                                   goalsMap_, totals_)));
  }
  return result;
}

TotalAll::TotalAll(const std::string& argNutName,
                   const numbers::Number& argTotalNutAmt,
                   const NutMap& goals)
{
  nutName = argNutName;
  totalNutAmt = argTotalNutAmt;
  NutMap::const_iterator goal = goals.find(nutName);
  if (goal == goals.end())
    goalStatus = G_NOGOAL;
  else if (goal->second.isZero())
    goalStatus = G_ZEROGOAL;
  else
  {
    goalStatus = G_GOAL;
    pctGoal = totalNutAmt.getValue() / goal->second.getValue() * 100;
  }
}

std::map<std::string, TotalAll> Totaler::getTotalAll() const
{
  std::map<std::string, TotalAll> result;
  for (NutMap::const_iterator nameAndAmt = totals_.begin();
          nameAndAmt != totals_.end(); ++nameAndAmt)
  {
    const std::string& nutName = nameAndAmt->first;
    const numbers::Number& totalNutAmt = nameAndAmt->second;
    result.insert(std::make_pair(nutName,
                                 TotalAll(nutName, totalNutAmt,
                                          goalsMap_)));
  }
  return result;
}

} // namespace totaler
