#ifndef TOTALER_H
#define TOTALER_H

#include "food.h"
#include "numbers.h"

#include <map>
#include <string>
#include <utility>
#include <vector>

namespace totaler {

typedef std::map<std::string, numbers::Number> NutMap;
typedef std::vector<std::pair<std::string, std::string> > StringPairs;
typedef std::vector<std::pair<std::string, numbers::Number > > Goals;
typedef std::vector<std::string> StringVec;

enum GoalStatus
{
  G_NOGOAL,
  G_ZEROGOAL,
  G_GOAL
};

/** A vector of FoodGoal instances is returned when getFoodGoals is
 * called. There is one FoodGoal instance in the vector for each
 * nutrient that has a goal, even if this food has no data for that
 * nutrient. Thus, hasNut is False if this food has none of this
 * particular nutrient.
 *
 * goalStatus may be G_ZEROGOAL or G_GOAL. If goalStatus is
 * G_ZEROGOAL, then pctGoal is garbage; otherwise, if goalStatus is
 * G_GOAL, then pctGoal is this nutrient's percentage of the goal.
 * This is necessary to avoid divide by zero errors.
 */
struct FoodGoal
{
  FoodGoal(const NutMap::value_type& goal,
           bool hasNut,
           const numbers::Number& nutAmt,
           const NutMap& totals);
  std::string nutName;
  bool hasNut;
  numbers::Number nutAmt;
  GoalStatus goalStatus;
  float pctGoal;
  bool hasPctTotal;
  float pctTotal;
};

/** A vector of TotalGoal instances is returned when getTotalGoals
 * is called.
 */
struct TotalGoal
{
  TotalGoal(const NutMap::value_type& goal, const NutMap& totals);
  std::string nutName;
  bool hasNut;
  numbers::Number nutAmt;
  GoalStatus goalStatus;
  float pctGoal;
};

/** A map of FoodAll instances is returned when getFoodAll is
 * called. 
 */
struct FoodAll
{
  FoodAll(const std::string& nutName,
          const numbers::Number& nutAmt,
          const NutMap& goals,
          const NutMap& totals);
  std::string nutName;
  numbers::Number nutAmt;
  GoalStatus goalStatus;
  float pctGoal;
  bool hasPctTotal;
  float pctTotal;
};

/** A map of TotalAll instances is returned when getTotalAll is
 * called.
 */
struct TotalAll
{
  TotalAll(const std::string& nutName,
           const numbers::Number& nutAmt,
           const NutMap& goals);
  std::string nutName;
  numbers::Number totalNutAmt;
  GoalStatus goalStatus;
  float pctGoal;
};

/** Transform vector of pairs of strings to pairs of strings and
 * numbers::Number instances. The string is the nutrient name; the
 * numbers::Number is the goal. This returns a vector that may be used
 * in the ctor for Totaler.
 *
 * Separating this function from the Totaler ctor allows the user
 * input to be parsed early to make sure there are no errors. This
 * function will throw exceptions if the input is erroneous.
 */
Goals getGoals(const StringPairs& goalStrings);

/** Using a vector of foods and some goals (if any), compute the
 * total nutrients of the foods and the goals.
 *
 * This task is actually somewhat complicated, so there is an entire
 * class dedicated solely to this purpose. When printing a nutrients
 * report or a goals report, the columns on the right describe this
 * nutrient's percent of the total in the buffer and this nutrient's
 * percent of the total goal. In addition, Nutrient and Goal reports
 * are avaiable to give totals for the entire buffer. This class
 * performs the necessary calculations.
 *
 * Doing these calculations can take some time (especially for huge
 * buffers) and the calculations take place when the class is
 * instantiated, so only instantiate this class if if will be
 * needed. However, calculating the totals is the hard part--copying
 * this class is relatively cheap because the totals themselves
 * aren't all that big.
 */
class Totaler
{
public:
  /** Constructor.
   *
   * \param foods Totals will be computed for these foods.
   * Typically this will be a buffer, so we will say "buffer" in
   * the documentation.
   *
   * \param goals Vector of pairs. First value in the pair is the
   * name of a nutrient (e.g. "Calories"). The second value
   * is a numbers::Number instance containing the goal.
   */
  Totaler(const std::vector<const food::Food*>& foods,
          const Goals& goals);

  /** Obtain data on this food's percentage of the goals, as well
   * as its percentage of the total nutrients. The vector contains
   * data in the same order as the goals given in the ctor.
   */
  std::vector<FoodGoal>          getFoodGoals(const food::Food&) const;

  /** Returns data on the total amount of nutrients for the foods
   * in the buffer. Nutrients are given in the same order as the
   * goals.
   */
  std::vector<TotalGoal>         getTotalGoals()                  const;

  /** Returns this food's percentage of all the nutrients in the
   * buffer. Goals are irrelevant; all nutrients are reported.
   */
  std::map<std::string, FoodAll>  getFoodAll(const food::Food&)   const;

  /** Returns total nutrients in the buffer. Goals are irrelevant;
   * all nutrients are reported.
   */
  std::map<std::string, TotalAll> getTotalAll()                    const;

private:

  /** Goals from the ctor.
   */
  Goals goals_;

  /** Same as above, but in map form instead.
   */
  NutMap goalsMap_;

  /** Calculated totals for the buffer.
   */
  NutMap totals_;

  void initTotals(const std::vector<const food::Food*>& foods);
};

} // namespace totaler

#endif
